#include <stdlib.h>
#include "config.h"
#include "classifier.h"
#include "packet.h"
#include "ip.h"
#include "SourceRtg.h"

class QoS_Classifier : public Classifier {
public:
	QoS_Classifier();
	~QoS_Classifier();

	//	void recv(Packet* p, Handler* h);

	int maxslot(int CoS) const { 
		if(CoS == 0) return maxslot_; 
		else if (CoS == 1) return maxslot1_;
		else if (CoS == 2) return maxslot2_;
		else { printf("Erro, CoS invalido\n");
		       return -1;}
	}

	inline NsObject* slot(int slot, int CoS) {                      //luish
		if (CoS == 0){
			if ((slot >= 0) || (slot < nslot_))
				return slot_[slot];
			return 0;
		} else if (CoS == 1){
			if ((slot >= 0) || (slot < nslot1_))
				return slot1_[slot];
			return 0;
		} else if (CoS == 2){
			if ((slot >= 0) || (slot < nslot2_))
				return slot2_[slot];
			return 0;
		} else return 0;         //rever!!!
	}

	//	int mshift(int val) { return ((val >> shift_) & mask_); }
	// a principio, isto e' herdado de classifier luish
	//      NsObject* find(Packet*);
	//	int classify(Packet *const);     //tirei o virtual luish


protected:
	int classify(Packet *const p);
	void QoS_Classifier::alloc(int slot, int CoS);
	void QoS_Classifier::install(int slot, NsObject* p, int CoS);
	void QoS_Classifier::clear(int slot, int CoS);
	int QoS_Classifier::getnxt(NsObject *nullagent, int CoS);
	void QoS_Classifier::recv(Packet* p, Handler*h);
	NsObject* QoS_Classifier::find(Packet* p);
	int QoS_Classifier::command(int argc, const char*const* argv);

	NsObject** slot_;	/* table that maps slot number to a NsObject */
	int nslot_;
	int maxslot_;
	NsObject** slot1_;	/* luish: now we have two more tables, or    */
	int nslot1_;                                                    //luish
	int maxslot1_;                                                  //luish
	NsObject** slot2_;	/* three entries per destin in the table     */
	int nslot2_;                                                    //luish
	int maxslot2_;                                                  //luish
	int offset_;		// offset for Packet::access()
	int shift_;
	int mask_;
};

static class QoS_ClassifierClass : public TclClass {
public:
	QoS_ClassifierClass() : TclClass("Classifier/QoS") {}
	TclObject* create(int, const char*const*) {
		return (new QoS_Classifier());
	}
} class_qos_classifier;

//luish: Now we have three entries per destin in the routing table, one for 
//       each Class of Service
QoS_Classifier::QoS_Classifier() : slot_(0), nslot_(0), maxslot_(-1)
	, slot1_(0), nslot1_(0), maxslot1_(-1)                          //luish
	, slot2_(0), nslot2_(0), maxslot2_(-1)                          //luish
	, shift_(0), mask_(0xffffffff)
{
	bind("offset_", &offset_);
	bind("shift_", &shift_);
	bind("mask_", &mask_);
//	bind("off_ip_", &off_ip_); // XXX to be removed
}

//int QoS_Classifier::classify(Packet *const p)
//{
//	return (mshift(*((int*) p->access(offset_))));
//}

int QoS_Classifier::classify(Packet *const p) {

  //  int i;

  hdr_ip* iph = hdr_ip::access(p);

  hdr_src_rtg *srt = hdr_src_rtg::access(p);
  
  //hdr_cmn *ch = (hdr_cmn*)p->access(off_cmn_);

  if (srt->srcflag_ == 1) 
    { 
      //printf("Source ROUTE: ");
      //for ( i=0; i < srt->length(); i++ )
      //{
      //  printf(" %d ",srt->list_k(i));
      //}
      //printf("\n");
      
      int pt = srt->pointer();
      srt->pointer() = srt->pointer() + 1;
      
      //printf("nao %d\n",mshift(iph->dst()));
      //printf("name: %s pt: %d sim list %d uid: %d  \n",name(),pt,srt->list_k(pt),ch->uid_);

      if ( pt >= srt->length() ) return mshift(iph->daddr());
      return srt->list_k(pt);
    }
  else
    {
      return mshift(iph->daddr());
    }
}

//luish: This method must destroy the other two entries too!
QoS_Classifier::~QoS_Classifier()
{
	delete [] slot_;
	delete [] slot1_;                                               //luish
	delete [] slot2_;                                               //luish
}

//luish: This method receives another argument to alloc slots for each CoS
void QoS_Classifier::alloc(int slot, int CoS)
{
	if (CoS == 0){
		NsObject** old = slot_;
		int n = nslot_;
		if (old == 0)
			nslot_ = 32;
		while (nslot_ <= slot)
			nslot_ <<= 1;
		slot_ = new NsObject*[nslot_];
		memset(slot_, 0, nslot_ * sizeof(NsObject*));
		for (int i = 0; i < n; ++i)
			slot_[i] = old[i];
		delete old;
	} else   if (CoS == 1){
		NsObject** old = slot1_;
		int n = nslot1_;
		if (old == 0)
			nslot1_ = 32;
		while (nslot1_ <= slot)
			nslot1_ <<= 1;
		slot1_ = new NsObject*[nslot1_];
		memset(slot1_, 0, nslot1_ * sizeof(NsObject*));
		for (int i = 0; i < n; ++i)
			slot1_[i] = old[i];
		delete old;
	} else   if (CoS == 2){
		NsObject** old = slot2_;
		int n = nslot2_;
		if (old == 0)
			nslot2_ = 32;
		while (nslot2_ <= slot)
			nslot2_ <<= 1;
		slot2_ = new NsObject*[nslot2_];
		memset(slot2_, 0, nslot2_ * sizeof(NsObject*));
		for (int i = 0; i < n; ++i)
			slot2_[i] = old[i];
		delete old;
	} else printf ("Erro em alloc slot\n");
}

//luish: This method receives another argument to install slots for each CoS
void QoS_Classifier::install(int slot, NsObject* p, int CoS)
{
	if (CoS == 0){                                                  //luish
		if (slot >= nslot_)
			alloc(slot,0);
		slot_[slot] = p;
		if (slot >= maxslot_)
			maxslot_ = slot;
	} else if (CoS == 1){
		if (slot >= nslot1_)
			alloc(slot,1);
		slot1_[slot] = p;
		if (slot >= maxslot1_)
			maxslot1_ = slot;
	}else if (CoS == 2){
		if (slot >= nslot2_)
			alloc(slot,2);
		slot2_[slot] = p;
		if (slot >= maxslot2_)
			maxslot2_ = slot;
	} else printf("Erro, invalido CoS: %d em install \n", CoS);
}

//luish: changes to consider CoS... 
void QoS_Classifier::clear(int slot, int CoS)
{
	if (CoS == 0){
		slot_[slot] = 0;
		if (slot == maxslot_) {
			while (--maxslot_ >= 0 && slot_[maxslot_] == 0)
				;
		}
	}else if (CoS == 1){
		slot1_[slot] = 0;
		if (slot == maxslot1_) {
			while (--maxslot1_ >= 0 && slot1_[maxslot1_] == 0)
				;
		}
	}else if (CoS == 2){
		slot2_[slot] = 0;
		if (slot == maxslot2_) {
			while (--maxslot2_ >= 0 && slot2_[maxslot2_] == 0)
				;
		}	
	}else printf("Erro, invalido CoS:%d em clear\n", CoS);
}

//luish: Added CoS to this method
int QoS_Classifier::getnxt(NsObject *nullagent, int CoS)
{
	int i;
	if (CoS == 0){
		for (i=0; i < nslot_; i++)
			if (slot_[i]==0 || slot_[i]==nullagent)
				return i;
		i=nslot_;
		alloc(nslot_,CoS);
		return i;
	}else if (CoS == 1){
		for (i=0; i < nslot1_; i++)
			if (slot1_[i]==0 || slot1_[i]==nullagent)
				return i;
		i=nslot1_;
		alloc(nslot1_,CoS);
		return i;
	}else if (CoS == 2){
		for (i=0; i < nslot2_; i++)
			if (slot2_[i]==0 || slot2_[i]==nullagent)
				return i;
		i=nslot2_;
		alloc(nslot2_,CoS);
		return i;
	}else {
		printf("Erro, invalido CoS:%d em Classifier::getnxt\n", CoS);
		return -1;
	}
}

/*
 * objects only ever see "packet" events, which come either
 * from an incoming link or a local agent (i.e., packet source).
 */
//luish: no modifications to this method
void QoS_Classifier::recv(Packet* p, Handler*h)
{
	NsObject* node = find(p);
	if (node == NULL) {
		/*
		 * XXX this should be "dropped" somehow.  Right now,
		 * these events aren't traced.
		 */
		Packet::free(p);
		return;
	}
	node->recv(p,h);
}

/*
 * perform the mapping from packet to object
 * perform upcall if no mapping
 */

NsObject* QoS_Classifier::find(Packet* p)
{
	NsObject* node = NULL;
	int cl = classify(p);

	if (cl < 0 || cl >= nslot_ || (node = slot_[cl]) == 0) { 
		/*
		 * Sigh.  Can't pass the pkt out to tcl because it's
		 * not an object.
		 */
		printf("esta dando merda aqui em find luish");
	       
		Tcl::instance().evalf("%s no-slot %d", name(), cl);
		/*
		 * Try again.  Maybe callback patched up the table.
		 */
		cl = classify(p);
		if (cl < 0 || cl >= nslot_ || (node = slot_[cl]) == 0)
			return (NULL);
	}
	return (node);
}

//luish: some modifications must be done to consider CoS 
int QoS_Classifier::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 4) {             //luish: now, we have CoS also as argument
		/*
		 * $classifier alloc-port nullagent $CoS                //luish
		 */
		if (strcmp(argv[1],"alloc-port") == 0) {
			int slot;
			int CoS = atoi(argv[4]);                        //luish
			NsObject* nullagent =(NsObject*)TclObject::lookup(argv[2]);
			slot=getnxt(nullagent,CoS);                     //luish
			tcl.resultf("%u",slot);
			return(TCL_OK);
		}

		/*
		 * $classifier clear $slot $CoS
		 */
		if (strcmp(argv[1], "clear") == 0) {
			int slot = atoi(argv[2]);
			int CoS = atoi(argv[4]);                        //luish
			clear(slot,CoS);
			return (TCL_OK);
		}
		/*
		 * $classifier installNext $node $CoS
		 */
		if (strcmp(argv[1], "installNext") == 0) {
			int CoS = atoi(argv[4]);                        //luish
			int slot;
			if(CoS == 0) slot = maxslot_ + 1;
			else if (CoS == 1) slot = maxslot1_ + 1;
			else if (CoS == 2) slot = maxslot2_ + 1;
			else printf("Erro, invalido CoS:%d em Classifier::command\n", CoS);
                        //int slot = maxslot_ + 1;                     original
			NsObject* node = (NsObject*)TclObject::lookup(argv[2]);
			if (node == NULL) {
				tcl.resultf("Classifier::installNext attempt to install non-object %s into classifier", argv[2]);
				return TCL_ERROR;
			};
				
			install(slot, node, CoS);                       //luish
			tcl.resultf("%u", slot);
			return TCL_OK;
		}
		/*
		 * $classifier slot snum $CoS
		 * returns the name of the object in slot # snum, for $CoS
		 */
		if (strcmp(argv[1], "slot") == 0) {
			int slot = atoi(argv[2]);
                        int CoS = atoi(argv[4]);
                        if (CoS == 0){
                           if (slot >= 0 && slot < nslot_ && slot_[slot] != NULL) {
				tcl.resultf("%s", slot_[slot]->name());
				return TCL_OK;
                           }
                        }
                        if (CoS == 1){
                           if (slot >= 0 && slot < nslot1_ && slot1_[slot] != NULL) {
				tcl.resultf("%s", slot1_[slot]->name());
				return TCL_OK;
                           }
                        }
                        if (CoS == 2){
                           if (slot >= 0 && slot < nslot2_ && slot2_[slot] != NULL) {
				tcl.resultf("%s", slot2_[slot]->name());
				return TCL_OK;
                           }
                        }
			tcl.resultf("Classifier: no object at slot %d", slot);
			return (TCL_ERROR);
		}
		/*
		 * $classifier findslot $node $CoS
		 * finds the slot containing $node at $CoS
		 */
		if (strcmp(argv[1], "findslot") == 0) {
			int slot = 0;
                        int CoS = atoi(argv[4]);
			NsObject* node = (NsObject*)TclObject::lookup(argv[2]);
			if (node == NULL) {
				return (TCL_ERROR);
			}
                        if (CoS==0){
			  while (slot < nslot_) {
			    if (strcmp(slot_[slot]->name(), argv[2]) == 0) {
                                tcl.resultf("%u", slot);
			        return (TCL_OK);
			    }
			    slot++;
			  }
                        } 
                        if (CoS==1){
			  while (slot < nslot1_) {
			    if (strcmp(slot1_[slot]->name(), argv[2]) == 0) {
                                tcl.resultf("%u", slot);
			        return (TCL_OK);
			    }
			    slot++;
			  }
                        } 
                        if (CoS==2){
			  while (slot < nslot2_) {
			    if (strcmp(slot2_[slot]->name(), argv[2]) == 0) {
                                tcl.resultf("%u", slot);
			        return (TCL_OK);
			    }
			    slot++;
			  }
                        } 
			tcl.result("-1");
			return (TCL_OK);
		}
	} else if (argc == 5) {      //luish: now, we have CoS also as argument
		/*
		 * $classifier install $slot $node $CoS
		 */
		if (strcmp(argv[1], "install") == 0) {
			int slot = atoi(argv[2]);
			int CoS = atoi(argv[4]);
			NsObject* node = (NsObject*)TclObject::lookup(argv[3]);
			install(slot, node,CoS);
			return (TCL_OK);
		}
	}
	return (NsObject::command(argc, argv));
}

