#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /usr/src/mash/repository/vint/ns-2/classifier-mcast.cc,v 1.29 1999/09/20 01:54:59 heideman Exp $";
#endif

#include <stdlib.h>
#include "config.h"
#include "packet.h"
#include "ip.h"
#include "classifier.h"
#include "classifier-addr.h"

// Modified Unicast Address Classifier  - Classifier/Addr/REU

class REU_AddressClassifier : public AddressClassifier {
protected:
  void recv(Packet* p, Handler* h);
  int classify(Packet * p);
};

static class REU_AddressClassifierClass : public TclClass {
public:
  REU_AddressClassifierClass() : TclClass("Classifier/Addr/REU") {}
  TclObject* create(int, const char*const*) {
    return (new REU_AddressClassifier());
  }
} class_reu_address_classifier;

void REU_AddressClassifier::recv(Packet* p, Handler* h)
{
  NsObject* node = find(p);
  if (node == NULL) {
    /*
     * XXX this should be "dropped" somehow.  Right now,
     * these events aren't traced.
     */
    Packet::free(p);
    return;
  }
  node->recv(p,h);
}

int REU_AddressClassifier::classify(Packet *p) {
  hdr_cmn *th = hdr_cmn::access(p);
  hdr_ip* iph = hdr_ip::access(p);
  
  // who am I?
  Tcl& tcl = Tcl::instance();
  tcl.evalf("[%s set node_] id", name());
  char *tcl_my_addr = tcl.result();
  int my_addr = atoi(tcl_my_addr);
  packet_t t = th->ptype();
  const char* type = packet_info.name(t);
  //printf("REU addr classifier, pkt type: %s, end: %d\n",type,my_addr);
  //printf("source: %d dst-port: %d\n",mshift(iph->saddr()),iph->dport());
  if ( strcmp(type,"rtProtoREU")==0 && (mshift(iph->saddr())!=my_addr) ) {
    //printf("pacote reu controle\n");
    return(my_addr);
  }
  
  if ( mshift(iph->daddr()) == my_addr && iph->dport() == 570 ) {
    //printf("Pacote pra mim2(%d)! \n",mshift(iph->daddr()));
    //hashnode *p = lookup(mshift(iph->saddr()),iph->dport());
  }
  return mshift(iph->daddr());
};

/***************************************************************************/
/* Classifier/REU
/* REU Classifier - decides if the packet is a REU packet and forwards     */
/* it to the address classifier (modified version - Classifier/Addr/REU    */
/***************************************************************************/

class REU_Classifier : public AddressClassifier {
public:
  int REU_Classifier::command(int, const char *const *);
protected:
  enum {HASHSIZE = 256};
  struct hashnode {
    int slot;
    nsaddr_t src;
    nsaddr_t dst;
    hashnode* next;
  };
  int hash(nsaddr_t src, nsaddr_t dst) const {
    u_int32_t s = src ^ dst;
    s ^= s >> 16;
    s ^= s >> 8;
    return (s & 0xff);
  }
  hashnode* ht_[HASHSIZE];
  
  int classify(Packet * p);
  void recv(Packet* p, Handler* h);
  NsObject* find(Packet* p);
  void set_hash(hashnode* ht[], nsaddr_t src, nsaddr_t dst,
		      int slot);
  void clearAll();
  void clearHash(hashnode* h[], int size);
  hashnode* lookup(nsaddr_t src, nsaddr_t dst) const;

  int mft_dst; 
};

static class REU_ClassifierClass : public TclClass {
public:
  REU_ClassifierClass() : TclClass("Classifier/REU") {}
  TclObject* create(int, const char*const*) {
    return (new REU_Classifier());
  }
} class_reu_classifier;

int REU_Classifier::classify(Packet *p) {
  hdr_cmn *th = hdr_cmn::access(p);
  hdr_ip* iph = hdr_ip::access(p);
  
  // who am I?
  Tcl& tcl = Tcl::instance();
  tcl.evalf("[%s set node_] id", name());
  char *tcl_my_addr = tcl.result();
  int my_addr = atoi(tcl_my_addr);
  
  packet_t t = th->ptype();
  const char* type = packet_info.name(t);

  // send the packet to the reu replicator, if it is destined to MFT<S>.dst
  if ((iph->dport() == 570) && (iph->daddr() == mft_dst)) {
    hashnode *hn = lookup(mshift(iph->saddr()),iph->dport());
    if ( hn != 0 && slot_[hn->slot] != 0 ) {
      return hn->slot;
    }
  }
  
  // classify me retorna o cara do slot 0, um Classifier/Addr/REU...
  return 0;
};

void REU_Classifier::recv(Packet* p, Handler* h)
{
  NsObject* node = find(p);
  if (node == NULL) {
    /*
     * XXX this should be "dropped" somehow.  Right now,
     * these events aren't traced.
     */
    printf("REU_Classifier::recv packet dropped, agent is gone!\n");
    Packet::free(p);
    return;
  }
  //printf("reuclassifier recv, node: %s\n",node->name());
  node->recv(p,h);
}

NsObject* REU_Classifier::find(Packet* p)
{
  NsObject* node = NULL;
  int cl = classify(p);
  if (cl < 0 || (node = slot_[cl]) == 0) { 
    if (default_target_) return default_target_;
    /*
     * Sigh.  Can't pass the pkt out to tcl because it's
     * not an object.
     */
    //Tcl::instance().evalf("%s no-slot %ld", name(), cl);
    printf("REU_Classifier::find packet dropped, agent is gone!\n");
    return NULL;
    //if (cl == TWICE) {
    //	/*
    //	 * Try again.  Maybe callback patched up the table.
    //	 */
    //	cl = classify(p);
    //	if (cl < 0 || cl >= nslot_ || (node = slot_[cl]) == 0)
    //		return (NULL);
    //}
  }
  return (node);
}

void REU_Classifier::set_hash(hashnode* ht[], nsaddr_t src, nsaddr_t dst, int slot)
{
  int h = hash(src, dst);
  //printf("hashed: %d\n",h);
  hashnode* p = new hashnode;
  p->src = src;
  p->dst = dst;
  p->slot = slot;
  p->next = ht[h];
  ht[h] = p;
}

REU_Classifier::hashnode* 
REU_Classifier::lookup(nsaddr_t src, nsaddr_t dst) const
{
  int h = hash(src, dst);
  //printf("src: %d dst: %d hash looked: %d\n",src,dst,h);
  hashnode* p = 0;
  for (p = ht_[h]; p != 0; p = p->next) {
    if (p->src == src && p->dst == dst)
      break;
  }
  return (p);
}

void REU_Classifier::clearHash(hashnode* h[], int size) 
{
  for (int i = 0; i < size; ++i) {
    hashnode* p = h[i];
    while (p != 0) {
      hashnode* n = p->next;
      delete p;
      p = n;
    }
  }
  memset(h, 0, size * sizeof(hashnode*));
}

void REU_Classifier::clearAll()
{
  clearHash(ht_, HASHSIZE);
}

int REU_Classifier::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();

  if (argc == 5) {
    if (strcmp(argv[1], "set-hash") == 0) {
      // $classifier set-hash $src $group $slot
      printf("REU Classifier set-hash\n");
      nsaddr_t src = strtol(argv[2], (char**)0, 0);
      nsaddr_t dst = strtol(argv[3], (char**)0, 0);
      int slot = atoi(argv[4]);
      set_hash(ht_, src, dst, slot);
      return (TCL_OK);
    }
  } else if (argc == 4) {    
    if (strcmp(argv[1], "lookup") == 0) {
      // $classifier lookup $src $group
      // returns name of the object (replicator)
      Tcl &tcl = Tcl::instance();
      nsaddr_t src = strtol(argv[2], (char**)0, 0);
      nsaddr_t dst = strtol(argv[3], (char**)0, 0);
      int iface = atoi(argv[4]);
      
      hashnode* p = lookup(src, dst);
      if ((p == 0) || (slot_[p->slot] == 0))
	tcl.resultf("");
      else 
	tcl.resultf("%s", slot_[p->slot]->name());
      return (TCL_OK);
    }
  } else if (argc == 3) {
    /*
     * $classifier delete-node $node
     */
    if (strcmp(argv[1], "delete-node") == 0) {
      printf("delete-node, maxslot: %d\n",maxslot_);
      
      NsObject* node = (NsObject*)TclObject::lookup(argv[2]);
      if (node == NULL) {
	tcl.resultf("Classifier::delete-node attempt to delete non-object %s from classifier", argv[2]);
	return TCL_ERROR;
      };
      
      if (maxslot_ < 0) {
	tcl.result("Sinistro em Classifier/REU delete-node");
	return (TCL_OK);
      };
      for (int i = 0; i <= maxslot_; i++) {
	printf("node : %s ",node->name());
	if ( slot_[i] == 0 ) 
	  continue;
	printf("slot[i] : %s ",slot_[i]->name());
	if ( strcmp(slot_[i]->name(),node->name()) == 0 ) {
	  clear(i);
	  printf("slotf : %d",i);
	  tcl.resultf("%u", i);
	  return TCL_OK;
	};
	tcl.resultf("Classifier::delete-node attempted to delete non-existing node %s from classifier", argv[2]);
      };
    }
    /*
     * $classifier installDst $dst_addr
     */
    if (strcmp(argv[1], "installDst") == 0) {
      mft_dst = atoi(argv[2]);
      printf("installDst: %d\n",mft_dst);
      return (TCL_OK);
    }
    
  } else if (argc == 2) {
    if (strcmp(argv[1], "clearAll") == 0) {
      clearAll();
      return (TCL_OK);
    }
  }
  return (AddressClassifier::command(argc, argv));
}

