/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) Xerox Corporation 1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative
 * works for research and evaluation purposes, provided that Xerox is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Xerox grants no other licenses expressed or
 * implied. The Xerox trade name should not be used in any advertising
 * without its written permission. 
 *
 * XEROX CORPORATION MAKES NO REPRESENTATIONS CONCERNING EITHER THE
 * MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS SOFTWARE
 * FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" without
 * express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this
 * software. 
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /archive/cvs/cvsroot/diffserv/ns-2/param-lh-adc.cc,v 1.2 2000/02/23 10:42:08 costa Exp $";
#endif


/* Parameter-based admission control.  Admission decisions are
 * based on the sum of reserved rates.
 */

#include "adc.h"
#include <stdlib.h>

class Param_lh_ADC : public ADC {
public:
	Param_lh_ADC();
	void teardown_action(int,double,int);
	void rej_action(int,double,int);
	void trace(TracedVar* v);
protected:
	int admit_flow(int,double,int);
	double utilization_;
	TracedDouble resv_rate_;
	double oresv_rate_;
};

Param_lh_ADC::Param_lh_ADC() : resv_rate_(0), oresv_rate_(0)
{
	bind("utilization_",&utilization_);
	type_ = new char[5];
	strcpy(type_, "PBAC");

	resv_rate_.tracer(this);
	resv_rate_.name("\"Reserved Rate\"");
}

void Param_lh_ADC::rej_action(int /*cl*/,double p,int /*r*/)
{
	resv_rate_-=p;
	//luish
	double av_bw;           
	av_bw = (bandwidth_ - double(resv_rate_)) / 1e6;
	//printf("rej av_bw: %f src: %d dst: %d \n",av_bw,src_,dst_);
	Tcl& tcl = Tcl::instance();
	tcl.evalf("[Simulator instance] cost1 [[Simulator instance] \
		           get-node-by-id %d] [[Simulator instance] get-node-by-id \
			   %d] %f",src_,dst_,av_bw);
	// end luish
}


void Param_lh_ADC::teardown_action(int /*cl*/,double p,int /*r*/)
{
	resv_rate_-=p;
	//luish
	//printf("tear resv_rate: %f\n",double(resv_rate_));
	double av_bw;           
	av_bw = (bandwidth_ - double(resv_rate_)) / 1e6;
	//printf("tear av_bw: %f src: %d dst: %d \n",av_bw,src_,dst_);
	Tcl& tcl = Tcl::instance();
	tcl.evalf("[Simulator instance] cost1 [[Simulator instance] \
                   get-node-by-id %d] [[Simulator instance] get-node-by-id \
                   %d] %f",src_,dst_,av_bw);
	// end luish
}

int Param_lh_ADC::admit_flow(int /*cl*/,double r,int /*b*/)
{
	double av_bw;           //luish

	if (resv_rate_ + r <= utilization_ * bandwidth_) {
	  resv_rate_ +=r;
	  //luish - bandwidth in Mbps within routing
	  av_bw = (bandwidth_ - double(resv_rate_)) / 1e6;
	  //printf("admit av_bw: %f src: %d dst: %d \n",av_bw,src_,dst_);
	  //printf("resv_rate_: %f, src: %d dst: %d av-bw: %f\n",double(resv_rate_),src_,dst_,av_bw);
	  Tcl& tcl = Tcl::instance();
	  tcl.evalf("[Simulator instance] cost1 [[Simulator instance] \
                     get-node-by-id %d] [[Simulator instance] get-node-by-id \
                     %d] %f",src_,dst_,av_bw);
	  // end luish
	  return 1;
	}
	return 0;
}

static class Param_lh_ADCClass : public TclClass {
public:
	Param_lh_ADCClass() : TclClass("ADC/Param_lh") {}
	TclObject* create(int,const char*const*) {
		return (new Param_lh_ADC());
	}
}class_Param_lh_adc;

void Param_lh_ADC::trace(TracedVar* v)
{
	char wrk[500];
	double *p, newval;

	/* check for right variable */
	if (strcmp(v->name(), "\"Reserved Rate\"") == 0) {
		p = &oresv_rate_;
	}
	else {
		fprintf(stderr, "PBAC: unknown trace var %s\n", v->name());
		return;
	}

	newval = double(*((TracedDouble*)v));

	if (tchan_) {
		int n;
		double t = Scheduler::instance().clock();
		/* f -t 0.0 -s 1 -a SA -T v -n Num -v 0 -o 0 */
		sprintf(wrk, "f -t %g -s %d -a %s:%d-%d -T v -n %s -v %g -o %g",
			t, src_, type_, src_, dst_, v->name(), newval, *p);
		n = strlen(wrk);
		wrk[n] = '\n';
		wrk[n+1] = 0;
		(void)Tcl_Write(tchan_, wrk, n+1);
		
	}

	*p = newval;

	return;



}

