#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /usr/src/mash/repository/vint/ns-2/replicator.cc,v 1.20 1999/09/20 01:55:03 heideman Exp $";
#endif

#include "classifier.h"
#include "packet.h"
#include "ip.h"

/*
 * A replicator is not really a packet classifier but
 * we simply find convenience in leveraging its slot table.
 * (this object used to implement fan-out on a multicast
 * router as well as broadcast LANs)
 */
class REUReplicator : public Classifier {
public:
  REUReplicator();
  void recv(Packet*, Handler* h = 0);
  int installNext(NsObject* p);
  virtual int classify(Packet*) {/*NOTREACHED*/ return -1;};
protected:
  virtual int command(int argc, const char*const* argv);
  int ignore_;
};

static class REUReplicatorClass : public TclClass {
public:
  REUReplicatorClass() : TclClass("Classifier/Replicator/REU") {}
  TclObject* create(int, const char*const*) {
    return (new REUReplicator());
  }
} class_reu_replicator;

REUReplicator::REUReplicator() : ignore_(0)
{
  //printf("created REUReplicator\n");
  bind("ignore_", &ignore_);
}

void REUReplicator::recv(Packet* p, Handler*)
{
  // who am I?
  Tcl& tcl = Tcl::instance();
  tcl.evalf("[%s set node_] id", name());
  char *tcl_my_addr = tcl.result();
  int my_addr = atoi(tcl_my_addr);

  //printf("REUReplicator recv node: %d\n",my_addr);
  
  hdr_ip* iph = hdr_ip::access(p);
  hdr_cmn* ch = hdr_cmn::access(p);
  if (maxslot_ < 0) {
    printf("problems in REUReplicator::recv maxslot: %d, node: %d\n",maxslot_,my_addr);
    if (!ignore_) 
      Tcl::instance().evalf("%s drop %ld %ld %d", name(), 
			    iph->saddr(), iph->daddr(), ch->iface());
    Packet::free(p);
    return;
  }
  
  //printf("cheguei aqui, maxslot: %d %ld %ld, port: %ld %ld\n",maxslot_,iph->saddr(),iph->daddr(),iph->sport(),iph->dport());

  NsObject* q = 0;
  for (int i = 0; i <= maxslot_; ++i) {
    NsObject* o = slot_[i];
    if (o != 0) {
      q = o;
      Packet* pc = p->copy();
      hdr_ip* iph_pc = hdr_ip::access(pc);
      iph_pc->dst_.addr_ = i;
      o->recv(pc);
    }
  }
  /* finally, send a copy of the original packet with the same destination address */
  if (q != 0) {
    Packet* pc = p->copy();
    q->recv(pc);
  }
  Packet::free(p);
}

int REUReplicator::installNext(NsObject* p)
{
  printf("installNext, maxslot: %d, no instalado: %s\n",maxslot_,p->name());
  int slot = maxslot_+1;
  if (slot >= nslot_)
    alloc(slot);
  slot_[slot] = p;
  printf("slotiN: %d\n",slot);
  return maxslot_++;
} 

int REUReplicator::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();
  
  if (argc == 2) {
    /*
     * $replicator slots
     */
    if (strcmp(argv[1], "slots") == 0) {
      if (maxslot_ < 0) {
	tcl.result("");
	return (TCL_OK);
      }
      for (int i = 0; i <= maxslot_; i++) {
	if (slot_[i] == 0) 
	  continue;
	//tcl.resultf("%s %s", tcl.result(),
	//    slot_[i]->name());
	tcl.resultf("%s %d",tcl.result(),i);
      }
      return (TCL_OK);
    }
    /*
     * $replicator slot-pairs
     */
    if (strcmp(argv[1], "slot-pairs") == 0) {
      //printf("slot-pairs\n");
      if (maxslot_ < 0) {
	tcl.result("");
	return (TCL_OK);
      }
      for (int i = 0; i <= maxslot_; i++) {
	if (slot_[i] == 0) 
	  continue;
	tcl.resultf("%s %u %s", tcl.result(),
		    i,slot_[i]->name());
      }
      return (TCL_OK);
    }
  }
  if (argc == 3) {
    /*
     * $classifier installNext $node
     */
    if (strcmp(argv[1], "installNext") == 0) {
      NsObject* node = (NsObject*)TclObject::lookup(argv[2]);
      if (node == NULL) {
	tcl.resultf("Classifier::installNext attempt to install non-object %s into classifier", argv[2]);
	return TCL_ERROR;
      };
      
      int slot = installNext(node);
      tcl.resultf("%u", slot);
      return TCL_OK;
    }
    /*
     * $classifier delete-node $node
     */
    if (strcmp(argv[1], "delete-node") == 0) {     
      NsObject* node = (NsObject*)TclObject::lookup(argv[2]);
      if (node == NULL) {
	tcl.resultf("Classifier::delete-node attempt to install non-object %s into classifier", argv[2]);
	return TCL_ERROR;
      };
      
      if (maxslot_ < 0) {
	tcl.result("");
	return (TCL_OK);
      };
      for (int i = 0; i <= maxslot_; i++) {
	if ( slot_[i] == 0 ) 
	  continue;
	if ( strcmp(slot_[i]->name(),node->name()) == 0 ) {
	  clear(i);
	  printf("slotf : %d",i);
	  tcl.resultf("%u", i);
	  return TCL_OK;
	};
	tcl.resultf("Classifier::delete-node attempted to delete non-existing node %s from classifier", argv[2]);
      };
    }
  }
  
  if (argc == 4) {
    /*
     * $classifier install $slot $node
     */
    if (strcmp(argv[1], "install") == 0) {
      int slot = atoi(argv[2]);
      NsObject* node = (NsObject*)TclObject::lookup(argv[3]);
      if (node == NULL) {
	tcl.resultf("REUReplicator::install attempt to install non-object %s into classifier", argv[3]);
	return TCL_ERROR;
      };
      install(slot, node);
      return (TCL_OK);
    }
    
  }
  return Classifier::command(argc, argv);
}
