/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
* Copyright (c) Xerox Corporation 1997. All rights reserved.
*
* License is granted to copy, to use, and to make and to use derivative
* works for research and evaluation purposes, provided that Xerox is
* acknowledged in all documentation pertaining to any such copy or
* derivative work. Xerox grants no other licenses expressed or
* implied. The Xerox trade name should not be used in any advertising
* without its written permission. 
*
* XEROX CORPORATION MAKES NO REPRESENTATIONS CONCERNING EITHER THE
* MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS SOFTWARE
* FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" without
* express or implied warranty of any kind.
*
* These notices must be retained in any copies of any part of this
* software. 
*/

//SignalAckClass
//Null receiver agent which merely acknowledges a request

#include "agent.h"
#include "ip.h"
#include "resv.h"
#include "SourceRtg.h"

class SAack_lh_Agent : public Agent {
public:
    SAack_lh_Agent();

protected:
    void recv(Packet *, Handler *);
    int command(int,const char* const*);
    int off_resv_;
};

SAack_lh_Agent::SAack_lh_Agent(): Agent(PT_TCP)
{
    bind("off_resv_",&off_resv_);
}

static class SAack_lhClass : public TclClass {
public:
    SAack_lhClass() : TclClass("Agent/SAack_lh") {}
    TclObject* create(int, const char*const*) {
	    return (new SAack_lh_Agent());
    }
} class_saack_lh;


void SAack_lh_Agent::recv(Packet *p, Handler *h)
{
    int i;

    hdr_src_rtg *srt = hdr_src_rtg::access(p);            //luish
    int route_sz = srt->length();
    //printf("pkt recv: %d length: %d \n",srt->srcflag(),route_sz);
    //printf("endereco: %d \n",addr_>>23);	
    //printf("enderecos addr: %d saddr: %d daddr: %d\n",addr(),0,daddr());	

    hdr_cmn *ch = (hdr_cmn*)p->access(off_cmn_);

    if (ch->ptype() == PT_REQUEST) {
	    //printf("packet request \n");
       Packet *newp =allocpkt();
       hdr_cmn *newch=(hdr_cmn*)newp->access(off_cmn_);
       newch->size()=ch->size();
       // turn the packet around by swapping src and dst
       hdr_ip * iph = (hdr_ip*)p->access(off_ip_);
       hdr_ip * newiph = (hdr_ip*)newp->access(off_ip_);
       newiph->dst() = iph->src();
       //newiph->src_ = iph->dst();
       //newiph->daddr()=iph->saddr();
       //newiph->saddr()=iph->daddr();
       //newiph->sport() =iph->sport();
       //newiph->dport() =iph->dport();
       newiph->flowid()=iph->flowid();
       newiph->prio()=iph->prio();

       // the response should follow the same route...  luish
       if (srt->srcflag()) {
	  hdr_src_rtg *newsrt = hdr_src_rtg::access(newp);
	  newsrt->srcflag() = 1; 
	  newsrt->pointer() = 1;
	  newsrt->length() = route_sz;
	  newsrt->list_ = (u_int32_t *) malloc(route_sz*sizeof(u_int32_t));
	  for (i=0;i<route_sz;i++) {
	      newsrt->list_[i] = srt->list_k(route_sz - 1 - i);
	  }
       }

       hdr_resv* rv=(hdr_resv*)p->access(off_resv_);

       // luish rever a linha abaixo!!!
       // if we didn't find a QoS route, reject the connection...
       if (srt->srcflag() == 0 ) rv->decision() = 0;

       hdr_resv* newrv=(hdr_resv*)newp->access(off_resv_);
       newrv->decision()=rv->decision();
       newrv->rate()=rv->rate();
       newrv->bucket()=rv->bucket();
       if (rv->decision())  
	  newch->ptype() = PT_ACCEPT;    
       else {                                 
	  newch->ptype() = PT_REJECT;
       }
       target_->recv(newp);
       Packet::free(p);
       srt->srcflag() = 0;   // luish lixo rever - hdr_src_rtg tem 
       free(srt->list_);
       return;               // que ser jogada fora
    }
    srt->srcflag() = 0;   // luish lixo rever
    Agent::recv(p,h);
}


int SAack_lh_Agent::command(int argc, const char*const* argv)
{
    return (Agent::command(argc,argv));
}
  
  

