#
# Copyright (c) 1999 The University of Paris VI
# All rights reserved.
# 
# Permission to use and copy this software in source and binary forms
# is hereby granted, provided that the above copyright notice, this
# paragraph and the following disclaimer are retained in any copies
# of any part of this software and that the University of Paris VI is
# acknowledged in all documentation pertaining to any such copy
# or derivative work. The name of the University of Paris VI may not
# be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
# THE UNIVERSITY OF PARIS VI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#
# Contributed by the RP Research Group, http://www.lip6.fr/rp
#

# Contain the common methods to lead performance studies
#

source ../airs/post_process.awk

#====================================================
# MonitorQueue  Class
#====================================================
Class Monitor

Monitor instproc init {ns link bandwidth} {
    $self instvar link_ bandwidth_
    $self instvar ns_
    
    set ns_ $ns   
    set link_ $link
    set bandwidth_ $bandwidth
}

#
# trace bytes departure variable of Monitor
Monitor instproc trace_bdep {interval nametracefile} {
        $self instvar ns_ interval_ dumpfilefd_ nametracefile_
        $self instvar attribut_
	
        set interval_ $interval
	set nametracefile_ $nametracefile
	
        set dumpfilefd_ [open $nametracefile a]
	puts "Monitor trace byte departure: opening file $nametracefile"
        
        set attribut_    bdepartures_
        $ns_ at $interval_ "$self dump $attribut_"
}


#====================================================
# MonitorQueue  Class
#====================================================
Class MonitorQueue -superclass Monitor

#
# Add Monitor to the link
#
MonitorQueue instproc monitor-link {} {
        $self instvar ns_  link_
        $self instvar qMonitor_
        
	set qMonitor_ [new QueueMonitor]

	$link_ attach-monitors [new SnoopQueue/In] \
		[new SnoopQueue/Out] [new SnoopQueue/Drop] $qMonitor_

	set bytesInt_ [new Integrator]
	$qMonitor_ set-bytes-integrator $bytesInt_
	set pktsInt_ [new Integrator]
	$qMonitor_ set-pkts-integrator $pktsInt_                
}



#
# prints "fromNode_ to toNode_; time: $time class: $class bytes: $bytes" 
#
MonitorQueue instproc dump {attribut} {
	$self instvar ns_ link_ qMonitor_
        $link_ instvar fromNode_ toNode_
        $self instvar interval_ dumpfilefd_ 
        $self instvar lastSample_
        
        set now [$ns_ now]
        
        if ![info exists $lastSample_] {
        	set lastSample_ 0
        }
        
        $ns_ at [expr $now  + $interval_] "$self dump"
        set bytes [eval "\$qMonitor_ set $attribut"]        	
        puts $dumpfilefd_ \
              "time: $now  $attribut: $bytes $interval_ [expr ($bytes-$lastSample_)/$interval]"     
	set lastSample_ $bytes
}

# A FINIR
# plot time vs. per-flow bytes
MonitorQueue instproc finish { bandwidth tmp graphfile sumTP} {
        $self instvar nametracefile_ namegraphfile_
	global opt
        
	 flush $dumpfilefd_
	 close $dumpfilefd_ 
	  
        # print: fid time bytes interval
	# initial format: "time: $now class: fid bytes: $bytes $interval"
        set awkCode {
                { printf "%4d %8d %16d %4d\n", $4, $2, $6, $7; }
        }

        puts "writing flow data to $namegraphfile_ ..."
        # prepare file
        exec awk $awkCode $nametracefile_ | sort > $tmp
        # plot for each flow
        exec awk $awk_byte bandwidth=$bandwidth $tmp >@ $graphfile
        # plot by class
        exec awk $awk_total_tp rho1=$opt(rho1) rho2=$opt(rho2) bandwidth=$bandwidth $tmp >@ $sumTP 
}


#====================================================
# MonitorFlowFid  Class
#====================================================
Class MonitorFlow -superclass Monitor

# Dump an attribut of flow  monitor 
# format (with field number)
#  1     time: $time 
#  3     fid:  $fid
#  5     class: $prio
#  7     $attribut: $bytes
#  9	 $interval 
MonitorFlow instproc dump {attribut} {
	$self instvar ns_ link_ dumpfilefd_
        $link_ instvar fromNode_ toNode_
        $self instvar interval_ flowmon_
        
        set now [$ns_ now]
        
        $ns_ at [expr $now  + $interval_] "$self dump $attribut"
        foreach flow [$flowmon_ flows] {
        	set bytes [eval "\$flow set $attribut"]
        	        	
                puts $dumpfilefd_ \
                   "time: $now fid: [$flow set flowid_] class: [$flow set prio_] $attribut $bytes $interval_" 
                eval "\$flow set $attribut 0"
                }
               
}


# plot time vs. per-flow bytes
MonitorFlow instproc finish {tmp namegraphfile sumTPfd} {
        $self instvar nametracefile_ bandwidth_ dumpfilefd_
        $self instvar attribut_
	global opt awk_rate awk_rate_avg
        
 
        puts "Process flowmonitor file $nametracefile_  ..."

	flush $dumpfilefd_
	close $dumpfilefd_ 
	 
        # prepare file
        exec awk [$self preparefile] attribut=$attribut_ $nametracefile_ | sort > $tmp
        # plot for each flow
        exec awk $awk_rate bandwidth=$bandwidth_ $tmp > $namegraphfile
        # compute average rate for each class
        exec awk $awk_rate_avg bandwidth=$bandwidth_ startt=$opt(begin-compute) $tmp >@ $sumTPfd 
}


#====================================================
# MonitorFlowFid  Class
#====================================================
Class MonitorFlowFid -superclass MonitorFlow

#
# Add Monitor to the link
#
MonitorFlowFid instproc monitor-link {} {
        $self instvar ns_  link_
        $self instvar flowmon_
       
        set flowmon_ [new QueueMonitor/ED/Flowmon]
        set cl [new Classifier/Hash/Fid 2]
        
        $flowmon_ classifier $cl
        
        $cl proc unknown-flow { src dst fid hashbucket } {
                set nflow [new QueueMonitor/ED/Flow]
                set slot [$self installNext $nflow]
                $self set-hash $hashbucket $src $dst $fid $slot
        }
        
        $cl proc no-slot slotnum {
                puts stderr "classifier $self, no-slot for slotnum $slotnum"
        }
        
        $link_ attach-monitors [new SnoopQueue/In] [new SnoopQueue/Out] \
                [new SnoopQueue/Drop] $flowmon_
                
}


# plot time vs. per-flow bytes
MonitorFlowFid instproc preparefile {} {        

        # print: prio time bytes interval
        set awkCode {
                $7 == attribut { printf "%4d %8d %16d %4d\n", $4, $2, $8, $9; }
        }
        return $awkCode
}



#====================================================
# MonitorFlowPrio  Class
#====================================================
Class MonitorFlowPrio -superclass MonitorFlow

#
# Add Monitor to the link
#
MonitorFlowPrio instproc monitor-link {numclass} {
        $self instvar ns_  link_
        $self instvar flowmon_
       
        set flowmon_ [new QueueMonitor/ED/Flowmon]
        set cl [new Classifier]
        
        $flowmon_ classifier $cl
        
	for {set i 1} { $i <= $numclass } {incr i} {
                set nflow [new QueueMonitor/ED/Flow]
		$cl install $i $nflow
        }
        
        $cl set shift_ 24
        $cl set mask_  0x07
        $cl set offset_ [PktHdr_offset PacketHeader/IP  prio_]
        
        $cl proc no-slot slotnum {
                puts stderr "ERROR classifier $self, no-slot for slotnum $slotnum"
                exit 1
        }
        
        $link_ attach-monitors [new SnoopQueue/In] [new SnoopQueue/Out] \
                [new SnoopQueue/Drop] $flowmon_
                
}



# plot time vs. per-flow bytes
MonitorFlowPrio instproc preparefile {} {
        
        # print: prio time bytes interval
        set awkCode {
                $7 == attribut { printf "%4d %8d %16d %4d\n", $6, $2, $8, $9; }
        }
	return $awkCode
}





#====================================================
# MonitorAgent  Class
#====================================================
Class MonitorAgent

# trace traffic to determine the throughtput sending or receiving by an agent
MonitorAgent instproc init {ns agent filefd interval} {
    $self instvar agent_ fid_
    $self instvar ns_
    $self instvar interval_ filefd_
    
    set ns_ $ns   
    set agent_ $agent
    set interval_ $interval
    set filefd_ $filefd
 
}

MonitorAgent instproc dump_d_agent {} {
#     format in file trace:
#     <time> recv <fid> <rate> in byte/sec

    $self instvar agent_
    $self instvar ns_
    $self instvar interval_ filefd_
             	
       	#Get the current time
        set now [$ns_ now]
	
#      How many bytes have been received by the traffic sinks?
	set byte [$agent_ set bytes_]

	set fid [$agent_ set fid_]
	
	puts $filefd_ "$now\trecv\t$fid\t[expr $byte/$interval_]"
	$agent_ set bytes_ 0
        
	$ns_ at [expr $now+$interval_] "$self dump_d_agent"
}


# A FINIR
proc dump_s_agent {} {

        set ns [Simulator instance]
             	
       	#Get the current time
        set now [$ns now]
	
#      How many bytes have been send by agent source?
       set seq [$agent set t_seqno_]

#     format in opt(graph) file trace:
#      send: <# fid> <time> <# of bytes>
	puts $filefd "$now\tsend\t$fid\t$seq"
        
	$ns at [expr $now+$interval] "dump_s_agent"
}


MonitorAgent instproc finish {tracefilename resultfilename bw maxfid} {

	set maxbytes [expr $bw / (8.0*100)]
	# maxbytes = bandwidth in byte/sec, nee to compute the % of bandwidth 
	#            used by each flow

	set awkCode  {
		BEGIN { OFS = "\t"}
		$3 == fid { print $1, $4/maxbytes }
	}
	set awkCodeAll {
		BEGIN { OFS = "\t"
			sum=0;
		      }
		{
			sum += $4/maxbytes;
			time = $1;
		}
		$3 == fid { print time, sum; sum = 0 }
	}
 
		
	if [file exists $resultfilename] {
		catch "exec rm -f $resultfilename"
	}

	set fd [open $resultfilename w]	

	for {set dst 0} {$dst <= $maxfid} {incr dst} {
		exec awk $awkCode fid=$dst maxbytes=$maxbytes $tracefilename >@ $fd
		puts $fd "\n\n "
		
	}

	exec awk $awkCodeAll fid=1 maxbytes=$maxbytes $tracefilename >@ $fd
	close $fd
}
