#
# Copyright (c) 1999 The University of Paris VI
# All rights reserved.
# 
# Permission to use and copy this software in source and binary forms
# is hereby granted, provided that the above copyright notice, this
# paragraph and the following disclaimer are retained in any copies
# of any part of this software and that the University of Paris VI is
# acknowledged in all documentation pertaining to any such copy
# or derivative work. The name of the University of Paris VI may not
# be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
# THE UNIVERSITY OF PARIS VI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#
# Contributed by the RP Research Group, http://www.lip6.fr/rp
#

#
# routines to perform post processing, mostly for plotting
#

#====================================================
# PostProcess Base Class
#====================================================

Class PostProcess

PostProcess instproc init { test nametr namegra } {
        $self instvar nametracefile_ namegraphfile_
        $self instvar label_ format_

        set label_ $test
        set nametracefile_ $nametr
        set namegraphfile_ $namegra
        
        set name [$self info class]
        set lname [split $name /]
        set format_ [lindex $lname 1]
        
        puts "format according $format_ "
}


#
# Plot per-flow bytes vs. time.
#
PostProcess instproc plot_bytes { bandwidth } {
        $self instvar label_ nametracefile_ namegraphfile_
	global opt

        puts "\nStart post process"
        puts "Process file $nametracefile_"
        
	# clean working directory 
        if [file exists $namegraphfile_] {
        	catch "exec rm -f $namegraphfile_"
        }

        set tmp tmp[pid]
        catch "exec rm -f [glob tmp*]"

	# open graph files        
        set sumTPfd [open $opt(summary-file) a]

 	$self rate-class $bandwidth $tmp $sumTPfd
        
        close $sumTPfd
        
        catch "exec rm -f $tmp"

	$self trace
	
}

PostProcess instproc trace {} {
        puts stderr "output format not defined"
}


#====================================================
# Specific graphic Class
#====================================================

Class PostProcess/gnuplot -superclass PostProcess

PostProcess/gnuplot instproc trace {}  {
	$self instvar namegraphfile_

#	exec gnuplot insimul.gp  2>/dev/null &
}


# plot rho vs. per-flow rates
# used from file generated by script-lib
PostProcess/gnuplot instproc rate-class {bandwidth tmp sumTPfd} {
        $self instvar label_ 
        $self instvar nametracefile_ namegraphfile_
	global opt
        
        exec awk [$self avg-rate] bandwidth=$bandwidth rho1=$opt(rho1) rho2=$opt(rho2)  startt=$opt(begin-compute) $nametracefile_ >@ $sumTPfd 
        
        for {set i 1} {$i <= 2} {incr i} {
         	exec awk [$self class-rate] bandwidth=$bandwidth class=$i $nametracefile_ >> $namegraphfile_
	}
}



#--------------------------------------------------------------
#
#                        AWK procedure
#
#--------------------------------------------------------------


# input line format:
# t: 13.305 finished_client 1 transfert_id 1 class 1  992848 bytes inter_delay 0.000 duration 10.919 When_active_client 0
# produce file for gnuplot
#   x axis: time
#   y axis: throughput for each class

PostProcess/gnuplot instproc avg-rate { } {
	set awkCodeAll {
		# <var name>b  before server
		# <var name>a  after server (normal case)
		BEGIN { OFS = "\t"; 
			for (i=1; i<=2 ; i++) {
			 nbclient[i] = 0; 
			 samplesize[i] = 0;
			 Lengthb[i] = 0;
			 Length[i] = 0;
			 inter_delay[i] = 0;
			 inter_delayb[i] = 0;
			 duration[i] = 0;
			 Maxbyte[i] = 0;
			 minbyte[i] = 1000000;
			}
		}

		$2 >= startt {
			if ($3 == "finished_client") {
		       		Length[$8] += $9/(1024*1024);
				duration[$8] += $14*0.01;
				inter_delay[$8] += $12;
				samplesize[$8] += 1;
				if ($14 != 0)
				{
				 rate =  $9/$14
				 if (rate > Maxbyte[$8])
					Maxbyte[$8] = rate;
				 if (rate < minbyte[$8])
					minbyte[$8] = rate;
				}
			} 
		}

		$3 == "starting_client" {
			    nbclient[$8]+=1
			    Lengthb[$8] += $9/(1024*1024);
 			    inter_delayb[$8] += $12;
 			    qlimit = $13;
		        }

		END {
		
# format according column number:
#  1: rho1
#  2: rho2
#  3: rho1 measured before server
#  4: rho1 measured after server
#  5: throughput for a micro flow in class 1
#  6: min throughput for a micro flow in class 1
#  7: Max throughput for a micro flow in class 1
#  8: number of client generated from beginning for class 1
#  9: size of sample from startt for class 1
#  10: rho2 measured before server
#  11: rho2 measured after server
#  12: min throughput for a micro flow in class 2
#  13: Max throughput for a micro flow in class 2
#  14: rate for a micro flow in class 2
#  15: number of client generated from beginning for class 2
#  16: size of sample from startt for class 2
#  17: rho measured before server
#  18: rho measured after server
#  19: rho theorical
#  20: Queue limit for each class
#  added by in_simul
#  21: average rate on bottlencek for class 1
#  22: average rate on bottlencek for class 2

		     rhob = 0;
		     rhoa = 0;
		     printf "%f %f ", rho1, rho2
		     for (x in Lengthb) {
			if (inter_delayb[x] !=0) 
		     	   	 rhoib = (Lengthb[x])/(inter_delayb[x]*bandwidth/(1024*1024*8));
		     	else
		     	   	 rhoib = 0;
		     	printf " %f ", rhoib;
		     	rhob += rhoib;
		     	if (inter_delay[x] !=0) 
		     	   	rhoia = (Length[x])/(inter_delay[x]*bandwidth/(1024*1024*8));
		     	else
		     	   	rhoia = 0;
		        printf " %f ", rhoia;
		     	rhoa += rhoia;
		     	# micro flow throughput normalized by total bandwidth
			if (duration[x] != 0)
				printf " %f ", Length[x]/(duration[x]*bandwidth/(1024*1024*8));
     			else 
		        	printf " 0 "
		     printf " %f ", minbyte[x]*8*100/bandwidth;
		     printf " %f ", Maxbyte[x]*8*100/bandwidth;
		     printf " %d ", nbclient[x];
		     printf " %d ", samplesize[x]; 
		     }
		     printf " %f ", rhob;
		     printf " %f ", rhoa;
		     printf " %f ", rho1+rho2;
		     printf " %d \n", qlimit
	     }
	}
}	


PostProcess/gnuplot instproc class-rate { } {
	set awkCodeAll {
		BEGIN { OFS = "\t"; 
		}
		$8 == class {
			if ($3 == "finished_client")  {
				if ($14 == 0)
					print "ERROR :", $2, $9, $14
				else
		   			print $2, ($9)/($14*0.01*bandwidth/8);
			}
		}
		END {
			print "\n\n"
		}
	}
}	



