#!/usr/local/bin/tclsh
#
# Copyright (c) 1999 The University of Paris VI
# All rights reserved.
# 
# Permission to use and copy this software in source and binary forms
# is hereby granted, provided that the above copyright notice, this
# paragraph and the following disclaimer are retained in any copies
# of any part of this software and that the University of Paris VI is
# acknowledged in all documentation pertaining to any such copy
# or derivative work. The name of the University of Paris VI may not
# be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
# THE UNIVERSITY OF PARIS VI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#
# Contributed by the RP Research Group, http://www.lip6.fr/rp
#


#====================================================
# Procedures Definition
#====================================================

# Procedure Usage ()
proc usage {} {
	global argv0
	puts stderr "usage: ns $argv0 <params> <test> resultfile"
	puts stderr "       params are: \n\t\tr,  \n\t\tphi1,  \n\t\trho"
	puts stderr "       valid tests are: \n\t\tWFQ,  \n\t\tDropTail"

	exit 1
}


proc finish {} {
	
#	exec "csh tracegr $phi1 $rho &
# 	catch "exec plot.sh $phi1 $rho  &"
}


#=======================================
##                 MAIN               ##
#=======================================

	global argc argv
		
	if {$argc < 5} {
		usage
	}
	
	set r    [lindex $argv 0]; # ratio Bw/peak rate
	set phi1 [lindex $argv 1]; # share for IN packets (class 1)
	set rho  [lindex $argv 2]; # Maximum link load
	set test [lindex $argv 3]
	set file [lindex $argv 4]

#====================================================
# Others Simulation Parameters
#====================================================
set M 1e6              ; # in bps
set max_client 2047     ; # Max nb of sources inside a class
set L [expr 2*1024*1024]  ; # Transfert Length in bytes
set pkt_size  1024    ; # Payload of 1 packet     
set sumTP sumTP.data ; # Working file
set directory test33r$r
set begin_compute	50. ;# duration of the initialisation period
set duration		150 ;# duration of the simulation
# Directory: 
#	testrx      Rho at level packet
#	test1rx     L is static and it is equal to 10 MBytes
#	test2rx     L is a random number, and its average is equal to 5 Mbytes
#	test3rx     L is a random number, and its average is equal to 1 Mbytes, pkt size = 1024
#		    stop when 1000 samples has been gotten
#=====================================================

	set under _



	if ![file exists $directory] {
		catch "exec mkdir $directory"
	}
	
        if [file exists $sumTP] {
	        catch "exec rm -f $sumTP"
	}        
        #10
	set step [expr $rho/10]
#	lappend lval [expr $step/10]
	set max [expr $rho-$step]
	for { set i $step} { $i<$max} {set i [expr $i+$step]} {
		lappend lval $i
	}
#	lappend lval [expr $rho - ($step/10)]
	exec clear
	puts "\n========================= ============== ==================="
	puts "Launch ns with rho1 according $lval"
	
foreach rho1 $lval  {
        # set rho1 0.8
	set rho2 [expr $rho - $rho1]
	set phi2 [expr 1- $phi1]
	set nphi1 [expr $phi1/$phi2]; # share of phi1 when phi2 =1

	set bottle_bw [expr $M*$r]; # Bottleneck bandwidth
	set lambda1 [expr $rho1*$bottle_bw/($L*8)]
	set lambda2 [expr $rho2*$bottle_bw/($L*8)]

	puts "========================= New Simulation ==================="
	puts "at [exec date]"	
	puts "r        = $r"
	puts "phi 1    = $phi1"
	puts "rho      = $rho"
	puts "rho 1    = $rho1"
	puts "rho 2    = $rho2"
	puts "lambda 1 = $lambda1"
	puts "lambda 2 = $lambda2"
	puts "L        = $L"
	puts "C        = $bottle_bw"

	puts "executing: ns script-lib.tcl $nphi1 $lambda1 $lambda2 $max_client $pkt_size $bottle_bw $M $begin_compute $L $rho1 $rho2  $sumTP $duration $test"
 	set errno [catch "exec ns script-lib.tcl $nphi1 $lambda1 $lambda2 $max_client $pkt_size $bottle_bw $M $begin_compute $L $rho1 $rho2  $sumTP $duration $test" msg] 
 	if {$errno == 1} {
		puts "ERROR:\n$msg\n"
		continue
	} 
	set suffix [lindex [split $rho1 .] 1]

	puts "move $test in $directory/$file$under$suffix"
	catch "exec mv $test $directory/$file$under$suffix"

}

finish
puts "copy $sumTP in $directory/$file.data"
catch "exec cp $sumTP $directory/$file.data"
#set fnam "*.ps"
#catch "exec mv $fnam  $directory"

