# load source Tcl for WFQ
set dir [pwd]
source ns-airs.tcl
source in_process.tcl
source post_process.tcl
catch "cd $dir"


#==============
# options table
set opt(stoptime)   10
set opt(trace)      out
set opt(interval)   1
set opt(bdep)	bdep.tr
set opt(bw)     1e6

#=================
set ns [new Simulator]


#Define different colors for data flows
$ns color 0 Blue
$ns color 1 Red

catch "exec rm -f [glob $opt(trace).*]"

#Open the trace file
set f [open $opt(trace).tr w]
$ns trace-all $f

#Open the nam trace file
set nf [open $opt(trace).nam w]
$ns namtrace-all $nf

#Open the record trace file
set rdf [open $opt(trace).rd w]


#  n0
#    \
#     n2 --- n3
#    /
#  n1
#Create four nodes
set n0 [$ns node]
set n1 [$ns node]
set n2 [$ns node]
set n3 [$ns node]

#Create links between the nodes
$ns duplex-link $n0 $n2 2Mb 10ms DropTail
$ns duplex-link $n1 $n2 2Mb 10ms DropTail

# install a duplex link with WFQ queues at both ends
$ns duplex-link $n2 $n3 $opt(bw) 10ms PQCM

# install queue  on the PQCM server at node $n2 and $n3
$ns install-phb $n2 $n3 0  Queue/DropTail 0.5
$ns install-phb $n2 $n3 1  Queue/DropTail 1

set bottleneck_link [$ns link $n2 $n3]

$ns duplex-link-op $n0 $n2 orient right-down
$ns duplex-link-op $n1 $n2 orient right-up
$ns duplex-link-op $n2 $n3 orient right

$ns duplex-link-op $n2 $n3 queuePos 0.5

set link [$ns link $n2 $n3]


set udp0 [new Agent/UDP]
$ns attach-agent $n0 $udp0
$udp0 set prio_ 0; # this is the info used by WFQ classifier
$udp0 set fid_ 0

set cbr0 [new Application/Traffic/CBR]
$cbr0 set packet_size_ 1000
$cbr0 set rate_ 1mb
$cbr0 attach-agent $udp0

set udp1 [new Agent/UDP]
$ns attach-agent $n1 $udp1
$udp1 set prio_ [expr 1*pow(16,6)]; # prio 1 for PQCM classifier
$udp1 set fid_ 1

set cbr1 [new Application/Traffic/CBR]
$cbr1 set packet_size_ 1000
$cbr1 set rate_ 1Mb
$cbr1 attach-agent $udp1

set lossm0 [new Agent/LossMonitor]
$ns attach-agent $n3 $lossm0
$lossm0 set fid_ 0

set lossm1 [new Agent/LossMonitor]
$ns attach-agent $n3 $lossm1
$lossm1 set fid_ 1

$ns connect $udp0 $lossm0
$ns connect $udp1 $lossm1

$ns at  0 "$cbr0 start"
$ns at  0 "$cbr1 start"


# trace byte departure for bottleneck link
        set monitor_ [new Monitor $ns $bottleneck_link]
        $monitor_ monitor-link
        
        $monitor_ trace_bdep $opt(stoptime) $opt(interval) $opt(bdep)


# trace byte received by agent

        set magent1 [new MonitorAgent $ns $lossm0 $rdf 1]
        set magent2 [new MonitorAgent $ns $lossm1 $rdf 1]


proc finish {} {
	global ns f nf rdf opt magent1
	
	$ns flush-trace
	close $f

	#Close the trace file
        close $nf
        close $rdf
	#Execute nam on the trace file
        #exec nam out.nam &

	# Post process the file created by record instproc	
	set graphfile $opt(trace).data
	
	$magent1 finish  $opt(trace).rd $graphfile $opt(bw) 1
	
	exit 0
}


$ns at 0.0 "$magent1 dump_d_agent"
$ns at 0.0 "$magent2 dump_d_agent"
$ns at $opt(stoptime) "finish"

puts "RUN"
$ns run

