#
# Copyright (c) 1998 The University of Paris VI
# Authors:  P. ANELLI,
#	    B. NASSEREDDINE
#
# All rights reserved.
# 
# Permission to use and copy this software in source and binary forms
# is hereby granted, provided that the above copyright notice, this
# paragraph and the following disclaimer are retained in any copies
# of any part of this software and that the University of Paris VI is
# acknowledged in all documentation pertaining to any such copy
# or derivative work. The name of the University of Paris VI may not
# be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
# THE UNIVERSITY OF PARIS VI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#
# Contributed by the RP Research Group, http://www.lip6.fr/rp
#

source ../lan/vlan.tcl

#defaults Parameters

Est set period_ 1 
Est set dst_ -1

Est/TimeWindow set T_ 3
Est/ExpAvg set w_ 0.8

#default setting
Clep set macSA_ 0
Clep set delay_ 0
Clep set bandwidth_ 1e6
Clep set rtot_     1e6
Clep set rbe_      1000
Clep set rfree_    1000
Clep set rfree_be_ 1000


Clep set fid_             0
Clep set prio_            1
Clep set checktime_       1
Clep set variance_        1
Clep set bcasttime_       30 
Clep set decrtime_        10
Clep set incrtime_        3
Clep set missing_message_ 3 
# % of error admitted before to take a decision
Clep set deviation_       5 
Clep set rmin_   1000
 
TBF set rate_   1000
TBF set bucket_ 1000
TBF set qlen_   100
TBF set delay_  1

# Global variables
 set Maxuses_    0.7

# LanClep-----------------------------------------
#
#	Use LanNode as base class. Add a module to control 
# bandwith share.
#
#--------------------------------------------------
Class LanClep -superclass  LanNode
LanClep set estType_  TimeWindow

LanClep instproc init {ns args} {

$self instvar tbfpriv_

	eval $self next $ns $args
	set tbfpriv_(-1) -1
}

# $lan addnode "$n1 $n2" $bw $delay $est $llType $ifqType $macType
LanClep instproc addNode {nodes bw delay {estType ""} \
                                {llType ""} \
		                {ifqType ""} \
		                {macType ""} } {

$self instvar lanIface_ estType_ ns_

        # call addNode of LanNode (Parent class)
	$self next $nodes $bw $delay $llType $ifqType $macType

	if {$estType == ""} { set estType $estType_ }

	foreach src $nodes {
		set nif $lanIface_($src)
		$nif add-clep $src $bw $delay $estType	
		
   	}
}


LanClep instproc make-resv { src rate fid } {

$self instvar lanIface_
$self instvar tbfpriv_
 	if { [info exist tbfpriv_($fid)] }  {
			return
	}
 	set nif $lanIface_($src)
 	set clep [$nif set clep_]

 	if { [$clep make-resv $rate ] } {
		set mclfid [$nif set measclassifier_]
		set measmodpriv [$nif set  measmodpriv_]
		set tbfpriv_($fid) [new TBF]
 		set slot [$mclfid installNext $tbfpriv_($fid)]
		$mclfid set-hash "auto" [$src id] 0 $fid $slot
		$tbfpriv_($fid) target $measmodpriv
		$tbfpriv_($fid) set rate_ $rate 
		$tbfpriv_($fid) set bucket_  $rate
	}
}

	
LanClep instproc  change-resv { src prate fid } {

$self instvar lanIface_ 
$self instvar tbfpriv_

	if { [ info exist tbfpriv_($fid)] } {
		set rate [$tbfpriv_($fid) set rate_]
		
		set nif $lanIface_($src)
 		set clep [$nif set clep_]
 		if { [$clep change-resv $rate $prate]  } {
			$tbfpriv_($fid) set rate_ [expr rate + prate]
			$tbfpriv_($fid) set bucket_ [expr rate + prate]
		}
		return  
			
	}
 	return
}



LanClep instproc delete-resv { src fid } {

$self instvar lanIface_ tbfpriv_

	if { [info exist tbfpriv_($fid)] } {
		set rate [$tbfpriv_($fid) set rate_]
		set nif $lanIface_($src)
 		set clep [$nif set clep_]
		$clep delete-resv $rate 
		
		set mclfid [$nif set measclassifier_]
		$mclfid del-hash [$src id] 0 $fid 		
		$tbfpriv_($fid) target $tbfpriv_($fid)
		delete $tbfpriv_($fid)
	}
		

}

# LanIface-----------------------------------------
#
#	add clep module at node's interface
#--------------------------------------------------
# FlowId 0 -> Best Effort traffic
# FlowId non-zero -> Int Serv traffic 
# measClassifier is an instance of Classifier/Hash/Flow

LanIface instproc add-clep { src bw delay est } {

$self instvar measclassifier_ estbe_ estpriv_ measmod_ measmodpriv_
$self instvar ifq_ tbf_ entry_
$self instvar clep_ mac_ ll_
$self instvar filter_ 
global ptype Maxuses_

	set ns [Simulator instance]
	
	set clep_  [new Clep]
	$clep_ set bandwidth_ $bw
	$clep_ set macSA_ [$mac_ id]
 	$clep_ set delay_ $delay
 	
 	$clep_ set rtot_ [expr  $bw*$Maxuses_] 
	set rbe [$clep_ set rbe_ ]
	$clep_ set rfree_  [ expr ($bw - $rbe ) ]
	$clep_ set rfree_be_ [ expr ($bw - $rbe ) ]  

	#Create modules for BE
	for {set i 0 } {$i<=1} {incr i} {	
	    #Create an est unit for class 0  (BE)
		set estbe_($i) [new Est/$est]
		$estbe_($i) set src_ [$src id]
		# 0 = class; $i= Input or output
		$clep_ attach-est $estbe_($i) 0 $i
    
		#Create a Measurement Module 
	    set measmod_($i) [new MeasureMod]
	    $clep_ attach-measmod $measmod_($i) 0 $i
	}
		
	#Create an est unit for class 1 (Priv)
	set estpriv_ [new Est/$est]
	$estpriv_ set src_ [$src id]
	$clep_ attach-est $estpriv_ 1 1
	set measmodpriv_ [new MeasureMod]
	$clep_ attach-measmod $measmodpriv_ 1 1
	$measmodpriv_ target $ll_			
	
	set tbf_ [new TBF]
	$tbf_ target $measmod_(1)
	$measmod_(0) target $tbf_
	$clep_ attach-tbf $tbf_
	$measmod_(1) target $ll_			

	#Create a fid classifier to decide which packets to measure 
	# with 5 buckets
        set measclassifier_ [new Classifier/Hash/Fid 5 ]
	#set slots for measclassifier
	# fid = 0 -> BE packets
	set slot [$measclassifier_ installNext $measmod_(0)]
	$measclassifier_ set-hash 0 0 0 0 $slot 
	$measclassifier_ set default_ 0
	
	# tell entry point for vlink
	$self set entry_ $measclassifier_

	#now attach clep module to send et receive its own message
	$clep_ target $measclassifier_

	#add a filter to catch clep packet
	set filter_ [new Filter/Field]
	$self add-clep-filter $filter_
	$filter_ filter-target $clep_
	$filter_ set offset_ [PktHdr_offset PacketHeader/Common ptype_]
	$filter_ set match_ [set ptype(clep)]

	#Schedule to start the clep object
	$ns at 0.0 "$clep_ start"
}

LanIface instproc add-clep-filter { filter } {
	$self instvar ll_
        $filter target [$ll_ up-target]
        $ll_ up-target $filter
}

LanIface instproc attach-trace { filefd } {
	$self instvar estbe_ estpriv_ clep_

#	$estpriv_ attach $filefd
#	$estbe_(0) attach $filefd
#	$estbe_(1) attach $filefd
	$clep_ attach $filefd
}



#    Helper function to output utilization and bw estimate
LanIface instproc trace-util { node interval {filefd ""}} {
	$self instvar estbe_ estpriv_

	set ns [Simulator instance]
	
	if { $filefd != "" } {
		set time [$ns now]
		puts $filefd "$time\t$node BE in: [$estbe_(0) load-est] [$estbe_(0) link-utlzn] BE out: [$estbe_(1) load-est] [$estbe_(1) link-utlzn]" 
		puts $filefd "$time\t$node PR out: [$estpriv_ load-est] [$estpriv_ link-utlzn]" 
	}
	$ns at [expr $time+$interval] "$self trace-util $node $interval $filefd" 
}




#
# newLan:  create a LAN from a set of nodes
#
Simulator instproc newLan {nodelist bw delay args} {
	set lan [eval new LanClep $self -bw $bw \
					-delay $delay \
					 $args]
	$lan addNode $nodelist $bw $delay
	$lan addVlink $nodelist $bw
	return $lan
}

# For convenience, use make-lan.  For more fine-grained control,
# use newLan instead of make-lan.
Simulator instproc make-lan { nodelist bw delay \
	                     {estType TimeWindow} \
                             {llType LL} \
			     {ifqType Queue/DropTail} \
			     {macType Mac} \
			     {chanType Channel}} {

 if {[string compare $macType "Mac/Csma/Cd"] == 0} {
            puts "Warning: Mac/Csma/Cd is out of date"
            puts "Warning: Please use Mac/802_3 to replace Mac/Csma/Cd"
            set macType "Mac/802_3"
        }



	set lan [new LanClep $self \
			-bw      $bw \
			-delay   $delay \
			-llType  $llType \
			-ifqType $ifqType \
			-macType $macType \
			-chanType $chanType]
	$lan addNode $nodelist $bw $delay $estType $llType $ifqType $macType
	$lan addVlink $nodelist $bw

	return $lan
}

