#
# Copyright (c) 1998 The University of Paris VI
# All rights reserved.
# 
# Permission to use and copy this software in source and binary forms
# is hereby granted, provided that the above copyright notice, this
# paragraph and the following disclaimer are retained in any copies
# of any part of this software and that the University of Paris VI is
# acknowledged in all documentation pertaining to any such copy
# or derivative work. The name of the University of Paris VI may not
# be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
# THE UNIVERSITY OF PARIS VI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#
# Contributed by the RP Research Group, http://www.lip6.fr/rp
#

# TO DO
#  use bw_parse for opt(bw) 


# To run all tests: test-all-clep
# to run individual test:
# ns test-suite-clep.tcl share
# ns test-suite-clep.tcl resv
# ns test-suite-clep.tcl mix
# ....
#
# To view a list of available tests to run with this script:
# ns test-suite-clep.tcl
#

# load source Tcl for Clep LanNode
set dir [pwd]
catch "cd ../clep"
source ns-clep.tcl
catch "cd $dir"


# Set options table
set opt(trace)	"" ; # name of trace file 
set opt(namtr)	"" ; # name of namtrace file
set opt(graph)  xfile; #name of trace file used by record procedure (record throughput)
set opt(debug)	debug; #name of debug file, record specific traces
set opt(startinterval) 10; # interval between starting a new source
set opt(udpnode)	6;        # number of udp source, (node n0 is the destination)
set opt(tcpnode)	1;        # number of tcp source, (node n0 is the destination)
set opt(bw)	1e6; # bandwith in bit/s
set opt(delay)	0.01ms; # Time taken by LL process 
set opt(stop)	2000 ; # duration of simulation
set opt(period) 5.0 ;   # duration for a sample of throughput
set opt(ll)	LL
set opt(ifq)	Queue/DropTail; # queue management for interface buffer
set opt(mac)	Mac/802_3; # Mac protocol
set opt(chan)	Channel
set opt(est)	ExpAvg; # algorithm to estimate throughput for Clep
set opt(app)	Traffic/CBR

#  Source rate  < 180 kbit/s
set opt(packetSize) 256 ;# constant size of the packets generated
set opt(interval)  0.01; #
set opt(random)  1; # 1=on, 0=off; flag indicating whether or not to introduce "noise" 
		    #              in the departure times
# For TCP
set opt(tcptype) Reno
set opt(windowsize) 20 ;# max bound on window size
set opt(verbosetcp) 0 
 
#Options for Reservation
set opt(resvbw) 400000
set opt(resvstart1)  160 
set opt(resvstart2)  500 
set opt(resvduration) 1400


# init global variable
set test ""


# -------------------------------------------
Class Topology

Topology instproc init ns {
	$self instvar node_ lan_
	global opt test

	set num [expr $opt(udpnode)+$opt(tcpnode)]
	for {set i 0} {$i <= $num} {incr i} {
		set node_($i) [$ns node]
		lappend nodelist $node_($i)
	}

	#Make a lan with clep structure 
	if { $test != "classic"} {
		set lan_ [$ns make-lan $nodelist  $opt(bw) $opt(delay) $opt(est) \
			$opt(ll)   $opt(ifq) \
			$opt(mac)   $opt(chan)]
	} else { 
		set lan_ [$ns make-lan $nodelist  $opt(bw) $opt(delay)  \
			$opt(ll)   $opt(ifq) \
			$opt(mac)   $opt(chan)]
	
		}
	
	
}


Topology instproc node? n {
    $self instvar node_
    if [info exists node_($n)] {
	set ret $node_($n)
    } else {
	set ret ""
    }
    set ret
}



# -------------------------------------------
Class TestSuite

TestSuite instproc init {} {
	$self instvar ns_  node_ topo_
	global opt
	
	set ns_ [new Simulator]
	
	$self create-trace
	$self info class
	#split classname /
	
	set topo_ [new Topology $ns_]
	foreach i [$topo_ array names node_] {
		set node_($i) [$topo_ node? $i]
	}
	
	$self create-source
}

TestSuite instproc run {} {
	$self instvar ns_ 
	
	$ns_ at 0.0 "$self record"
	$ns_ run
}

TestSuite instproc create-trace {} {
	$self instvar ns_ 
	$self instvar trfd_ nmfd_ xfilefd_
	global opt 

	if [file exists $opt(trace).tr] {
		catch "exec rm -f [glob *.tr]"
	}
	if { $opt(trace)!=""} {
		set trfd_ [open $opt(trace).tr w]
		$ns_ trace-all $trfd_
	}	

	if {$opt(namtr) != ""} {
		set nmfd_ [open $opt(namtr).nam w]
		$ns_ namtrace-all $nmfd_
	}

	if {$opt(graph) != ""} {
		set xfilefd_ [open $opt(graph).tr w]
	}
}

TestSuite instproc create-source {} {
	$self instvar node_ dstag_
	global opt  

	set starttime 0
	set pktSize $opt(packetSize)	
	set interval $opt(interval)
	set num $opt(udpnode)
	
	for {set i 1} {$i <= $num} {incr i} {
		puts "source udp $i"
		set dstag_($i) [$self new_udp $starttime $node_($i) $node_(0) $pktSize $interval $i]
		incr starttime $opt(startinterval)
	}
	incr num $opt(tcpnode)
	for {} {$i <= $num} {incr i} {
		puts "source tcp $i"
		set dstag_($i) [$self new_tcp $starttime $node_($i) $node_(0) $pktSize $interval \
					$i $opt(tcptype) $opt(windowsize) $opt(verbosetcp) ]
		incr starttime $opt(startinterval)
	}
}



TestSuite instproc new_udp { startTime source dest pktSize interval fid } {
	$self instvar ns_
	global opt
	
	set agents [$ns_ create-connection-list UDP $source LossMonitor $dest $fid ]
 	set s_agent [lindex $agents 0]
 	set d_agent [lindex $agents 1]
   
    	set app [$s_agent attach-app $opt(app)]
    	$app set random_ $opt(random)
    	$app set rate_ [expr 8 * $pktSize / $interval]
    	if {$pktSize > 0} {
		$app set packet_size_ $pktSize
    	}

   	$ns_ at $startTime  "$app start"
    	return $d_agent
}


TestSuite instproc new_tcp { startTime source dest pktSize interval fid type window verbose} {
	$self instvar ns_
	global opt
	
	set agents [$ns_ create-connection-list TCP/$type $source TCPSink $dest $fid]
 	set s_agent [lindex $agents 0]
 	set d_agent [lindex $agents 1]
	
	$s_agent set window_ $window
	if {$verbose == "1"} {
	  $self tcpDumpAll $s_agent 20.0 $fid 
	}

	set app [$s_agent attach-app $opt(app)]
    	$app set random_ $opt(random)
    	$app set rate_ [expr 8 * $pktSize / $interval]
	if {$pktSize > 0}  {
		$app set packet_size_ $pktSize
	}
	
	$ns_ at $startTime "$app start"
    	return $d_agent
}

TestSuite instproc tcpDumpAll { tcpSrc interval label } {
    $self instvar dump_inst_ ns_
    
    if ![info exists dump_inst_($tcpSrc)] {
	set dump_inst_($tcpSrc) 1
	set report $label/window=[$tcpSrc set window_]/packetSize=[$tcpSrc set packetSize_]
	puts $report
	$ns_ at 0.0 "$self tcpDumpAll $tcpSrc $interval $label"
	return
    }
    
    $ns_ at [expr [$ns_ now] + $interval] "$self tcpDumpAll $tcpSrc $interval $label"
    set report time=[$ns_ now]/class=$label/ack=[$tcpSrc set ack_]/packets_resent=[$tcpSrc set nrexmitpack_]
    puts $report
}       

TestSuite instproc makeresv {nodenum resvbw fid} {
	$self instvar node_ topo_
	
	set lan [$topo_ set lan_ ]
	$lan make-resv $node_($nodenum)  $resvbw $fid
}

TestSuite instproc delresv { nodenum  } {
	$self instvar node_ topo_
	
	set lan [$topo_ set lan_ ]
	$lan delete-resv $node_($nodenum) $nodenum
}

TestSuite instproc record {} {
        $self instvar ns_ xfilefd_  dstag_
        global opt  
             	
       	#Get the current time
        set now [$ns_ now]
	set num [expr $opt(udpnode) + $opt(tcpnode)]
	
	for {set dst 1} {$dst <= $num} {incr dst} {
#               How many bytes have been received by the traffic sinks?
       	        set bw [$dstag_($dst)  set bytes_]

		# format in opt(graph) file trace:
		#  <# node> <time> <# of bytes>
		puts $xfilefd_ "$dst\t$now\t[expr $bw/$opt(period)]"
		#puts "bw de dstag_(dst): [$dstag_($dst)  set bytes_]"
		#puts "variables de dstag $dst b6 : [$dstag_($dst) info vars]"
		$dstag_($dst) set bytes_ 0
	}
        
	$ns_ at [expr $now+$opt(period)] "$self record"
}


TestSuite instproc finish testname {
        global opt
        $self instvar ns_ trfd_ nmfd_  xfilefd_  topo_
        $topo_ instvar lan_   

	$ns_ flush-trace

        #Close the output files
	if { [info exists trfd_] } {
		close $trfd_
	}
	if { [info exists nmfd_] } {
		close $nmfd_
	}

	close $xfilefd_  

	# Post process the file created by record instproc	
	set graphfile $opt(graph).xg
	
	set maxbytes [expr $opt(bw) / (8.0*100)]
	set num [expr $opt(udpnode) + $opt(tcpnode)]

	set awkCode  {
		BEGIN { OFS = "\t"}
		$1 == fid { print $2, $3/maxbytes }
	}
	set awkCodeAll {
		BEGIN { OFS = "\t";
			sum =0; 
		      }
		{
			sum += $3/maxbytes;
			time = $2;
		}
		$1 == fid { print time, sum; sum = 0 }
	}
 
		
	if [file exists $graphfile] {
		catch "exec rm -f $graphfile"
	}
	set f [open $graphfile w]
	puts $f "TitleText: $testname"
	puts $f "Device: Postscript"
	
	catch "exec rm -f temp.p"
	exec touch temp.p
	for {set dst 1} {$dst <= $num} {incr dst} {
		exec echo "\n\"flow $dst" >> temp.p
		exec awk $awkCode fid=$dst maxbytes=$maxbytes $opt(graph).tr > temp.$dst
		exec cat temp.$dst >> temp.p
		exec echo " " >> temp.p
		
	}

	exec awk $awkCodeAll fid=$num maxbytes=$maxbytes $opt(graph).tr > temp.all
	exec echo "\n\"all " >> temp.p  
	exec cat temp.all >> temp.p 

	exec cat temp.p >@ $f
	close $f

	if [file exists $graphfile] {
		catch "exec rm -f [glob temp.*]"
	}
	
	exec xgraph -bb -tk -ly 0,100 -x time(s) -y bandwidth(%) $graphfile &
	exit 0
}
		

TestSuite instproc debug { period {filetr ""}} {
	$self instvar topo_ node_ ns_ debugfd_
	
	set name $filetr.tr
	
	if [file exists $name] {
		catch "exec rm -f $name"
	}

	if {$filetr != ""} {
		set debugfd_ [open $name w]
	}

	set lan [$topo_ set lan_]

	foreach i [array names node_] {
		set nif [$lan set lanIface_($node_($i)) ]
		$nif attach-trace $debugfd_
		
		set clep [$nif set clep_]
		$clep trace "rbe_" 
		$clep trace "rfree_" 
		$clep trace "rfree_be_" 
		
	#	$ns_ at 0.0 "$nif trace-util $i $period $debugfd_"
	}

}

TestSuite instproc debugraph { vtrace } {
        global opt
        $self instvar ns_  debugfd_   topo_
        $topo_ instvar lan_   

	$ns_ flush-trace

        #Close the output files
	if { [info exists debugfd_] } {
		close $debugfd_
	}

	
	set graphfile $opt(debug).$vtrace

	set num [expr $opt(udpnode) + $opt(tcpnode)]

	set awkCoderbe {
		($2==fid && $3==vartrace) { print $1, $4 }
	}
 
 
	
	set f [open $graphfile w] 
	puts $f "TitleText: debug $vtrace "
	puts $f "Device: Postscript"
	
	

	catch "exec rm -f temp.d"
	exec touch temp.d
	
	for {set dst 0} {$dst <= $num} {incr dst} {
		exec echo "\n\"rbe node: $dst" >> temp.d
		set dst1 [ expr $dst +1 ]
		exec awk $awkCoderbe fid=$dst1 vartrace=$vtrace  $opt(debug).tr > tempd.$dst
		exec cat tempd.$dst >> temp.d
		exec echo " " >> temp.d	
	}
	
	exec cat temp.d >@ $f
	close $f
	if [file exists $graphfile] {
		catch "exec rm -f [glob tempd.*]"
		
	}
	exec xgraph -bb -tk -x time(s) -y token(t/s) $graphfile &
	
}



#====================================================
# Specific Tests
#====================================================

Class Test/classic -superclass TestSuite

Test/classic instproc init {} {
	$self instvar test_
	set test_ "Classical_Ethernet"
	
	# load source Tcl for LanNode
	set dir [pwd]
	catch "cd ../lan"
	source vlan.tcl
	catch "cd $dir"


	$self next
}


Test/classic instproc run {} {
	$self instvar test_  ns_
	global opt
	$ns_ at $opt(stop) "$self finish $test_"
	$self next
}


Class Test/share -superclass TestSuite

Test/share instproc init {} {
	$self instvar test_
	set test_ "Clep_share"
	
	$self next
}


Test/share instproc run {} {
	$self instvar test_  ns_
	global opt
			
	#$self debug $opt(period) $opt(debug)
	#$ns_ at $opt(stop) "$self debugraph rbe_"
	#$ns_ at $opt(stop) "$self debugraph rfree_be_"
	#$ns_ at $opt(stop) "$self debugraph rfree_"
	$ns_ at $opt(stop) "$self finish $test_"
	$self next
}


Class Test/resvudp -superclass TestSuite

Test/resvudp instproc init {} {
	$self instvar test_
	set test_ "Clep_reservation"
	
	$self next
}


Test/resvudp instproc run {} {
	$self instvar test_ ns_ 
	global opt

	$self debug $opt(period) $opt(debug)

	$ns_ at $opt(resvstart1) "$self makeresv 1 $opt(resvbw) 1"
	#$ns_ at $opt(stop) "$self debugraph rbe_"
	#$ns_ at $opt(stop) "$self debugraph rfree_be_"
	$ns_ at $opt(stop) "$self debugraph rfree_"
	
	$ns_ at $opt(stop) "$self finish $test_"
	$self next
}  


Class Test/resvtcp -superclass TestSuite

Test/resvtcp instproc init {} {
	$self instvar test_
	
	$self next
	
	set test_ "Clep_resvtcp"

}


Test/resvtcp instproc run {} {
	$self instvar test_ ns_ 
	global opt
	
	$self debug $opt(period) $opt(debug)
	set num [expr $opt(udpnode) + 1 ]
 	$ns_ at $opt(resvstart1) "$self makeresv $num $opt(resvbw) $num"
 	set resvback [expr $opt(resvbw)/5]
 	$ns_ at $opt(resvstart1) "$self makeresv 0 $resvback $num"
	$ns_ at [expr $opt(resvstart1) + $opt(resvduration) ] "$self delresv $num"
	$ns_ at $opt(stop) "$self finish $test_"

	$self next
}  


Class Test/shareresv -superclass TestSuite

Test/shareresv instproc init {} {
	$self instvar test_
	set test_ "Clep_shareresv"
	
	$self next
}


Test/shareresv instproc run {} {
	$self instvar test_ ns_ 
	global opt

	$self debug $opt(period) $opt(debug)

	$ns_ at $opt(resvstart1) "$self makeresv 1 $opt(resvbw) 1"
	$ns_ at [expr $opt(resvstart1) + $opt(resvduration) ] "$self delresv 1 "
	$ns_ at $opt(stop) "$self debugraph rfree_be_"
	$ns_ at $opt(stop) "$self finish $test_"
	$self next
}  


Class Test/mix -superclass Test/shareresv

Test/mix instproc init {} {
	$self instvar test_
	
	$self next
	
	set test_ "Clep_mix"

}


Test/mix instproc run {} {
	$self instvar test_ ns_ 
	global opt

	set num [expr $opt(udpnode) + 1 ]
 	$ns_ at $opt(resvstart2) "$self makeresv $num $opt(resvbw) $num"
	$ns_ at [expr $opt(resvstart2) + $opt(resvduration) ] "$self delresv $num"

	$self next
}  



#====================================================
# Procedures Definition
#====================================================

# Procedure Usage ()
proc usage {} {
	global argv0
	puts stderr "usage: ns $argv0 <tests> "
	puts "Valid Tests: classic share resvudp resvtcp shareresv mix"
	exit 1
}


proc runtest {arg} {
 global test
	set b [llength $arg]
	if {$b == 1} {
		set test $arg
	} elseif {$b == 2} {
		set test [lindex $arg 0]
	} else {
		usage
	}
	switch $test {
		classic -
		share -
		resvudp -
		resvtcp -
		shareresv -
		mix {
			set t [new Test/$test ]
		}
		default {
			puts stderr "Unknown test $test"
			exit 1
		}
	}
	$t run
}

#=======================================
##                 MAIN               ##
#=======================================


global argv arg0
runtest $argv

