#
# tcl/lip6/HBH.tcl
#

set timeout1_  2
set timeout2_  3

Class HBH -superclass McastProtocol

HBH set advertInterval 0.8

set rtglibRNG [new RNG]
$rtglibRNG seed 1

HBH set mid_ 0

HBH instproc init { sim node } {
    $self instvar mctrl_ oiflist_
    $self instvar hbh_agent

    set hbh_agent [new Agent/HBH $self]
    $node attach $hbh_agent

    set mctrl_ [new Agent/Mcast/Control $self]
    $node attach $mctrl_
    $self next $sim $node
}

HBH instproc join-group  { grp src } {
    $self instvar node_ ns_ joined_

    $self dbg "sending HBH join-group < $src ([$src id]), $grp, [$node_ id] > "

    # We create a MFT with one entry (itself) at the endnodes.
    # This entry forwards the packets to the agent that joined at this node.

    $node_ hbh-create-mft $src $grp [$node_ id]

    # At the end-nodes we don't need timers to destroy the MFT entry 
    # (which is unique). We destroy this entry when the member leaves 
    # the group.

    [$node_ set mft_entry_timer_($src:$grp:[$node_ id])] cancel
    [$node_ set mft_entry_destroy_timer_($src:$grp:[$node_ id])] cancel
    
    set joined_($src:$grp) 1
    #
    # Send a periodic join concerning the agents joined locally
    #
    set join_updateTime [$ns_ now]
    $ns_ at $join_updateTime "$self send-periodic-join $src $grp"
}

HBH instproc leave-group { grp src } {
    $self instvar joined_ node_ ns_
    $self dbg "testando leave-group"
    set joined_($src:$grp) 0

    $node_ hbh-destroy-mft-entry $src $grp [$node_ id]
    $node_ hbh-destroy-mft $src $grp
}

HBH instproc join { src grp join_node } {
    $self instvar node_

    #send internal join msg

    $self send-ctrl "join" $src $grp $join_node [$src id]
    $self dbg "sent join: < $src $grp $join_node [$src id]>"
}

HBH instproc recv-join { src grp from_node to_node } {
    global rtglibRNG

    $self instvar node_ ns_
    
    #join msg processing
    #called like this: $proto recv-$type $src $grp $from_node
    
    $self dbg "received join <$src $grp $from_node $to_node>"

    # If I am the source...
    ###########################################################################

    if { $node_ == $src } {
	$self dbg "recv-join: I am the source"
	# Create the MFT if it does not exist
	#######################################################################
	if { [$node_ hbh-exist-mft $src $grp] == 0 } {
	    $node_ hbh-create-mft $src $grp $from_node
	    #
	    # Send a periodic tree to live branches (in MFT)
	    # Start sending tree msgs
	    ###################################################################
	    set tree_updateTime [$ns_ now]
	    $ns_ at $tree_updateTime "$self send-periodic-tree $src $grp"
	} else {
	    # Add the node to MFT if it is not there, refresh it otherwise
	    ###################################################################
	    if { [$node_ hbh-exist-mft-entry $src $grp $from_node] == 0 } {
		$node_ hbh-add-mft-entry $src $grp $from_node
	    } else {
		$node_ hbh-refresh-mft-entry $src $grp $from_node
	    }
	}
    } else {
	# If I am not the source, forward the message...
	# Does MFT<S> exist?
	#######################################################################
	if [$node_ hbh-exist-mft $src $grp] {
	    # Yes, MFT<S> exists
	    # Is MFT<S> stale?
	    ###################################################################
	    if { ![$node_ hbh-mft-fresh? $src $grp] } {
		# Yes, MFT<S> stale
		$self send-ctrl "join" $src $grp $from_node [$src id]
	    } else {
		# No, MFT<S> is not stale
		#
		# Is R in MFT<S>?
		###############################################################
		if [$node_ hbh-exist-mft-entry $src $grp $from_node] {
		    # Yes, MFT<S>.R exists
		    # Is MFT<S>==R?
		    set tam [$node_ hbh-mft-size? $src $grp]
		    if { $tam == 1 } {
			# Yes
			$node_ hbh-destroy-mft-entry $src $grp $from_node
			# destroy mft... ?
			$self dbg "bizarro"
			$node_ hbh-destroy-mft $src $grp
			$node_ hbh-create-mct $src $grp $from_node
			$self send-ctrl "join" $src $grp $from_node [$src id]
		    } else {
			# No
			#
			# Refresh MFT<S>.R
			# Discard join(S,R) forward join(S,B)
			#######################################################
			$node_ hbh-refresh-mft-entry $src $grp $from_node
			$self send-ctrl "join" $src $grp [$node_ id] [$src id]
		    }
		} else {
		    # No MFT<S>.R
		    # Forward join(src,from_node) unchanged
		    ###########################################################
		    $self send-ctrl "join" $src $grp $from_node [$src id]
		}	
	    }
	} else {	    
	    # No MFT<S>
	    # Forward join(src,from_node) unchanged
	    ###################################################################
	    $self send-ctrl "join" $src $grp $from_node [$src id]
	}
    }
}

HBH instproc tree { src grp to_node } {
    $self instvar node_ hbh_agent
    # send tree msg
    $self dbg "sent tree <$src $grp [$node_ id] $to_node>"
    $self send-ctrl "tree" $src $grp [$node_ id] $to_node
}

HBH instproc send-all-mft-tree { src grp } {
    $self instvar node_

    array set hbhreplicator_ [$node_ array get hbhreplicator_]
    array set mft_entry_fg_ [$node_ array get mft_entry_fg_]

    $self dbg "send-all-mft-tree $src $grp"
    
    foreach recv [$hbhreplicator_($src:$grp) slots] {
	set fresh  $mft_entry_fg_($src:$grp:$recv)
	if { ([$node_ id] != $recv) && ($fresh) } {
	    $self send-ctrl "tree" $src $grp [$node_ id] $recv
	}    
    }
}

HBH instproc recv-tree { src grp from_node to_node } {
    $self instvar node_
    $self dbg "received tree <$src $grp $from_node $to_node>"

    # Am I R?
    ###########################################################################
    if { [$node_ id] == $to_node } {
	# Yes, I'm R
	# Discard tree (S,R) Forward tree(S,R[n])
	#######################################################################
	$self dbg "I am R in recv-tree"
	if [$node_ hbh-exist-mft $src $grp] {
	    $node_ hbh-refresh-mft $src $grp          ;# COMENTER LUIS
	    $self send-all-mft-tree $src $grp
	} else {
	    # if I don't have a MFT, I should not receive a tree message
	    # addressed to me - discard it...
	    $self err "Oops, I have no MFT but I received a tree msg..."
	}
    } else {
	# No, I'm not R
	# Does MFT<S> exist?
	#######################################################################
	if [$node_ hbh-exist-mft $src $grp] {	    
	    # Yes, MFT<S> exists
	    # Is R in MFT<S>?
	    ###################################################################
	    $node_ hbh-refresh-mft $src $grp                   ;# COMENTER LUIS
	    if [$node_ hbh-exist-mft-entry $src $grp $to_node] {
		# Yes
		# Refresh MFT<S>.R
		###############################################################
		$node_ hbh-refresh-mft-entry $src $grp $to_node
		# Is R marked?
		###############################################################
		if [$node_ hbh-mft-entry-marked? $src $grp $to_node] {
		    # Yes, R is marked
		    # Forward the tree message re-signed
		    ###########################################################
		    $node_ hbh-refresh-mft-entry $src $grp $to_node
		    $self fusion $src $grp [$node_ id] $from_node
		    $self send-ctrl "tree" $src $grp [$node_ id] $to_node ;#4
		} else {
		    # No, R is not marked
		    # send fusion R + Rn to Bp
		    # Forward the tree message re-signed
		    ##########################################################
		    $node_ hbh-refresh-mft-entry $src $grp $to_node
		    $self fusion $src $grp [$node_ id] $from_node
		    $self send-ctrl "tree" $src $grp [$node_ id] $to_node
		}
	    } else {
		# No, MFT<S>.R does not exist
		# Send fusion(R+R[n])
		# Insert R in MFT<S>
		# Discard tree (S,R) Forward tree(S,R[n]) (signed tree)
		###############################################################
		$node_ hbh-add-mft-entry $src $grp $to_node
		$self fusion $src $grp [$node_ id] $from_node
		$self send-ctrl "tree" $src $grp [$node_ id] $to_node
	    }
	} else {
	    # No MFT<S>
	    # Does MCT<S> exist?
	    ###################################################################
	    if [$node_ hbh-exist-mct $src $grp] {
		# Yes
		# Forward tree(src,from_node) unchanged
		# Does MCT<S> = R?
		###############################################################
		if [$node_ hbh-exist-mct-entry $src $grp $to_node] {
		    
		    # Yes
		    # Refresh MCT<S>.R
		    # Forward tree(S,R)
		    ###########################################################
		    $node_ hbh-refresh-mct $src $grp
		    $self send-ctrl "tree" $src $grp $from_node $to_node ;#4
		} else {
		    # No, MCT<S> = R'
		    # Is MCT<S> stale?
		    ###########################################################
		    if [$node_ hbh-mct-fresh? $src $grp] {
			# Not stale	
			# Send fusion(R+R')
			# Remove R' from MCT<S>
			# Insert R, R' in MFT<S>
			# Discard tree (S,R) Forward tree(S,R[n])
			#######################################################
			set node1 [$node_ hbh-read-mct $src $grp]
			# read MCT<S> and construct array nodes
			$node_ hbh-destroy-mct $src $grp
			$node_ hbh-create-mft $src $grp $to_node
			$node_ hbh-add-mft-entry $src $grp $node1
			$self fusion $src $grp [$node_ id] $from_node
			# Discard tree (S,R) Forward tree(S,R[n])
			#$self send-all-mft-tree $src $grp   ;# ???
			$self send-ctrl "tree" $src $grp [$node_ id] $node1 ;#4
			$self send-ctrl "tree" $src $grp [$node_ id] $to_node ;#4
		    } else {
			# MCT<S> stale
			# Remove R' from MCT<S>
			# Insert R in MCT<S>
			# Forward tree(S,R)
			#######################################################
			$node_ hbh-destroy-mct $src $grp
			$node_ hbh-create-mct $src $grp $to_node
			$self send-ctrl "tree" $src $grp $from_node $to_node ;#4
		    }
		}
	    } else {
		# No MCT<S>
		# Create MCT<S>
		# Insert R in MCT<S>
		# Forward tree(S,R)
		###############################################################
		$node_ hbh-create-mct $src $grp $to_node
		$self send-ctrl "tree" $src $grp $from_node $to_node ;#4
	    }
	}
    }
}

HBH instproc fusion { src grp from_node to_node } {
    $self instvar node_ hbh_agent
    # send fusion msg

    if [$node_ hbh-exist-mft $src $grp] {
	array set rep [$node_ hbh-read-mft $src $grp]
	set fid [$class get-next-mid]
	$self dbg "sent fusion($fid - [array names rep]) <$src $grp [$node_ id] $to_node>"
	set i 0
	$class set msg_($fid) ""
	foreach n [array names rep] {
	    #$self dbg "fusion rep: [array names rep]"
	    if [$node_ hbh-mft-entry-fresh? $src $grp $n ] {
		#if { ![$node_ hbh-mft-entry-marked? $src $grp $n] } {
		    set a [$class set msg_($fid)]
		    $class set msg_($fid) [concat $a $i $n ]
		    incr i
		#}
	    }
	}
	$self send-ctrl "fusion" $src $grp [$node_ id] $to_node $fid
    } else {
	$self err "problems in fusion"
    }
    
}

HBH instproc recv-fusion { src grp from_node to_node fid } {
    $self instvar node_
    # fusion msg processing
    
    set fsn [HBH read-msg $fid]
    array set fusioned $fsn
    $self dbg "received fusion fd: [array get fusioned] <$src $grp $from_node $to_node>"
    
    # Is this fusion addressed to me?
    #
    if { [$node_ id] == $to_node } {
    	#
	# Does MFT<S> exist?
	#
	if [$node_ hbh-exist-mft $src $grp] {
	    # Yes, MFT<S> exists
	    #
	    # Mark R,R[n]
	    # Insert B in MFT<S>
	    # (Forward tree(S,R[n]) Forward data do B)
	    #
	    set R_here 0
	    foreach R [array names fusioned] {
		if [$node_ hbh-exist-mft-entry $src $grp $fusioned($R)] {
		    $node_ hbh-mark-mft-entry $src $grp $fusioned($R) $from_node
		    set R_here 1
		} else { set R_there 1 }
	    }
	    if { $R_here == 1 } {
		#
		# Does MFT<S>.B exist?
		#
		if [$node_ hbh-exist-mft-entry $src $grp $from_node] {
		    # Yes, MFT<S>.B exists
		    #
		    # Refresh MFT<S>.B
		    #
		    $node_ hbh-refresh-mft-entry $src $grp $from_node
		    set t1 [$node_ set mft_entry_timer_($src:$grp:$from_node)]
		    $t1 expire
		} else {
		    # No, MFT<S>.B does not exist
		    #
		    # Insert B in MFT<S>
		    #
		    $node_ hbh-add-mft-entry $src $grp $from_node
		    set t1 [$node_ set mft_entry_timer_($src:$grp:$from_node)]
		    $t1 expire
		}
		#HBH retrieve-msg $fid   ;# delete the fusion message...
	    }
	    #else 
	    # there was no receiver in the MFT, forward the message
	    #	if { $node_ != $src } 
	    #$self send-ctrl "fusion" $src $grp $from_node $to_node $fid
	    #if [info exists R_there] 
	    #    puts stderr "PQP"
	    #    if { $node_ != $src } 
	    #    $self send-ctrl "fusion" $src $grp $from_node $to_node $fid
	} else {
	    # No MFT<S>
	    # Forward fusion
	    #
	    $self err "this should be impossible"
	    #$self send-ctrl "fusion" $src $grp $from_node $to_node $fid
	}
    } else {
	$self send-ctrl "fusion" $src $grp $from_node $to_node $fid
    }
}

HBH instproc drop { replicator src dst iface} {
    $self instvar node_ ns_
    puts stderr "testando drop"
}

HBH instproc send-ctrl { which src grp from_node to_node {fid ""} } {
    $self instvar mctrl_ ns_ node_
    $self instvar hbh_agent
    
    set message(0) $which
    set message(1) $src
    set message(2) $grp
    set message(3) $from_node
    set message(4) $to_node
    if { $which == "fusion" } {
	set message(5) $fid
    } else { set message(5) 0 }
    set id [$class get-next-mid]
    $class set msg_($id) [array get message]
    $hbh_agent send-control $to_node 0 $id [array size message]]
}

################ Helpers

HBH instproc dbg arg {
    $self instvar ns_ node_
    puts [format "At %.4f : Node([$node_ id]) $arg" [$ns_ now]]
}

HBH instproc err arg {
    $self instvar ns_ node_
    puts [format "Error at %.4f : Node([$node_ id]) $arg" [$ns_ now]]
}

Node instproc dbg arg {
    $self instvar ns_
    puts [format "At %.4f : Node([$self id]) $arg" [$ns_ now]]
}

Node instproc err arg {
    $self instvar ns_
    puts [format "Error at %.4f : Node([$self id]) $arg" [$ns_ now]]
}

Class SkelTimer -superclass Timer

SkelTimer instproc init { ns node src grp } {
    $self instvar node_ src_ grp_
    $self next $ns

    set ns_ $ns
    set node_ $node
    set src_ $src
    set grp_ $grp
}

SkelTimer instproc dbg arg {
    $self instvar ns_ node_
    puts [format "At %.4f : Node([$node_ id]) $arg" [$ns_ now]]
}


########
# Table manipulation
########

Classifier/HBH instproc init args {
    $self next
    $self instvar nrep_
    set nrep_ 1  
    # luish: 1 because slot 0 is reserved to the unicast classifer
}

Classifier/HBH instproc add-rep { rep src group } {
    $self instvar nrep_
    $self set-hash [$src id] $group $nrep_
    $self install $nrep_ $rep
    puts stderr "hbhreplicator installed $nrep_ $rep"
    incr nrep_
}

Node instproc hbh-create-mft { src grp from_node } {

    # multiclassifier_ is a reference to the Multicast/HBH Classifier
    # hbhreplicator_ is an array indexed by ($src:$grp)

    $self instvar ns_ hbhclassifier_ hbhreplicator_ Agents_ 
    $self instvar mft_fg_ mft_entry_fg_ mft_entry_mk_ 
    $self instvar  mft_entry_timer_ mft_entry_destroy_timer_

    $self dbg "hbh-create-mft: $src $grp $from_node"

    if [info exists hbhreplicator_($src:$grp)] {
	set r $hbhreplicator_($src:$grp)
    } else {
	set r [new Classifier/Replicator/HBH $ns_ $self $src $grp]
	$r set srcID_ $src
	$r set grp_ $grp
	set hbhreplicator_($src:$grp) $r
	$r set node_ $self

	#
	# Install the replicator.  
	#
	# luish - instala-lo em hbhclassifier_ 
	$hbhclassifier_ add-rep $r $src $grp
    }

    $self hbh-add-mft-entry $src $grp $from_node

    set mft_fg_($src:$grp) 1
}

Node instproc hbh-exist-mft { src grp } {
    # multiclassifier_ is a reference to the Multicast/HBH Classifier
    # hbhreplicator_ is an array indexed by ($src:$grp)

    $self instvar multiclassifier_ hbhreplicator_

    $self dbg "hbh-exist-mft: $src $grp"
    return [info exists hbhreplicator_($src:$grp)]
}

Node instproc hbh-destroy-mft { src grp } {
    $self instvar hbhreplicator_ hbhclassifier_

    $self dbg "hbh-destroy-mft $src:$grp"

    #luish:  falta tirar a referencia a rep de hbhclassifier...
    $hbhclassifier_ delete-node $hbhreplicator_($src:$grp)

    delete $hbhreplicator_($src:$grp)
    unset hbhreplicator_($src:$grp)
}

Node instproc hbh-read-mft { src grp } {
    $self instvar hbhreplicator_
    array set resp [$hbhreplicator_($src:$grp) slot-pairs]
    return [array get resp]
}

Node instproc hbh-refresh-mft { src grp } {
    $self instvar mft_fg_
    set mft_fg_($src:$grp) 1
}

Node instproc hbh-mft-fresh? { src grp } {
    $self instvar mft_fg_
    $self dbg "hbh-mft-fresh?"
    return $mft_fg_($src:$grp)
}

Node instproc hbh-mft-entry-fresh? { src grp node } {
    $self instvar mft_entry_fg_
    $self dbg "hbh-mft-entry-fresh?"
    return $mft_entry_fg_($src:$grp:$node)
}

Node instproc hbh-mft-size? { src grp } {
    $self instvar hbhreplicator_
    array set lixo [$hbhreplicator_($src:$grp) slot-pairs]
    return [array size lixo]
}

Node instproc hbh-add-mft-entry {src grp node} {
    $self instvar ns_ classifier_ hbhreplicator_ mft_entry_mk_ mft_entry_fg_ 
    $self instvar mft_entry_timer_ mft_entry_destroy_timer_
    global timeout1_ timeout2_

    if [info exists hbhreplicator_($src:$grp)] {
	$hbhreplicator_($src:$grp) install $node $classifier_
	set mft_entry_mk_($src:$grp:$node) -1
	set mft_entry_fg_($src:$grp:$node)  1
	#
	# Stale Timer creation
	#
	set mft_entry_timer_($src:$grp:$node) \
		[new SkelTimer/MFTentry $ns_ $self $src $grp $node]
	$mft_entry_timer_($src:$grp:$node) sched $timeout1_
	#
	# Destroy Timer creation
	#
	set mft_entry_destroy_timer_($src:$grp:$node) \
		[new SkelTimer/MFTentry/Destroy $ns_ $self $src $grp $node]
	$mft_entry_destroy_timer_($src:$grp:$node) sched $timeout2_
    } else {
	$self err "hbh-add-mft-entry error: no replicator for $src:$grp"
    }
    $self dbg "hbh-add-mft-entry: $src $grp $node"
}

Node instproc hbh-exist-mft-entry {src grp node} {
    $self instvar hbhreplicator_ ns_

    $self dbg "hbh-exist-mft-entry: $src $grp $node"

    if [info exists hbhreplicator_($src:$grp)] {
	array set x [$hbhreplicator_($src:$grp) slot-pairs]
	foreach i [array names x] {
	    if { $i == $node } { return 1 }
	}
	return 0
    } else {
	$self err "hbh-exist-mft-entry error: no replicator for $src:$grp"
    }
}

Node instproc hbh-mark-mft-entry {src grp node from_node} {
    $self instvar ns_ hbhreplicator_ mft_entry_mk_ mft_entry_mk_timer_
    global timeout1_

    $self dbg "hbh-mark-mft-entry $src:$grp:$node"

    $hbhreplicator_($src:$grp) deactivate $node
    set mft_entry_mk_($src:$grp:$node) $from_node
    if [info exists mft_entry_mk_timer_($src:$grp:$node)] {
	$mft_entry_mk_timer_($src:$grp:$node) cancel
	$mft_entry_mk_timer_($src:$grp:$node) sched $timeout1_
    } else {	
	set mft_entry_mk_timer_($src:$grp:$node) \
		[new SkelTimer/Mark $ns_ $self $src $grp $node]
	$mft_entry_mk_timer_($src:$grp:$node) sched $timeout1_
    }
}

Node instproc hbh-unmark-mft-entry {src grp node} {
    $self instvar ns_ hbhreplicator_ classifier_ 
    $self instvar mft_entry_mk_ mft_entry_mk_timer_

    $self dbg "hbh-unmark-mft-entry $src:$grp:$node"

    if [$self hbh-exist-mft-entry $src $grp $node] {
	set mft_entry_mk_($src:$grp:$node) -1
	$hbhreplicator_($src:$grp) install $node $classifier_
    }
    $mft_entry_mk_timer_($src:$grp:$node) destroy
    delete $mft_entry_mk_timer_($src:$grp:$node)
    unset mft_entry_mk_timer_($src:$grp:$node)
}

Node instproc hbh-mft-entry-marked? {src grp node} {
    $self instvar mft_entry_mk_
    if { $mft_entry_mk_($src:$grp:$node) == -1 } {
	return 0
    } else { return 1 }
    #return $mft_entry_mk_($src:$grp:$node)
}

Node instproc hbh-read-marked {src grp node} {
    $self instvar mft_entry_mk_
    return $mft_entry_mk_($src:$grp:$node)
}

Node instproc hbh-refresh-mft-entry {src grp node} {
    $self instvar mft_entry_fg_ mft_ mft_entry_timer_ mft_entry_destroy_timer_
    global timeout1_ timeout2_

    $self dbg "hbh-refresh-mft-entry $src $grp $node"

    $mft_entry_timer_($src:$grp:$node) cancel
    $mft_entry_timer_($src:$grp:$node) sched $timeout1_
    set mft_entry_fg_($src:$grp:$node) 1
    $mft_entry_destroy_timer_($src:$grp:$node) cancel
    $mft_entry_destroy_timer_($src:$grp:$node) sched $timeout2_
}

Node instproc hbh-destroy-mft-entry {src grp node} {
    $self instvar ns_ hbhreplicator_ mft_entry_mk_ mft_entry_fg_
    $self instvar mft_entry_timer_ mft_entry_destroy_timer_
    $self instvar mft_entry_mk_timer_
    
    $self dbg "hbh-destroy-mft-entry $src:$grp:$node" 

    if [info exists hbhreplicator_($src:$grp)] {
	$hbhreplicator_($src:$grp) clear $node
	$mft_entry_timer_($src:$grp:$node) cancel
	delete $mft_entry_timer_($src:$grp:$node)
	unset mft_entry_timer_($src:$grp:$node)
	$mft_entry_destroy_timer_($src:$grp:$node) cancel
	delete $mft_entry_destroy_timer_($src:$grp:$node)
	unset mft_entry_destroy_timer_($src:$grp:$node)
	if { [$self hbh-mft-entry-marked? $src $grp $node] } {
	    $mft_entry_mk_timer_($src:$grp:$node) cancel
	    delete $mft_entry_mk_timer_($src:$grp:$node)
	    unset mft_entry_mk_timer_($src:$grp:$node)
	}
	unset mft_entry_mk_($src:$grp:$node)
	unset mft_entry_fg_($src:$grp:$node)
	array set lixo [$hbhreplicator_($src:$grp) slot-pairs]
	$self dbg "mft size: [array size lixo]"
	if { $self != $src } {
	    set tam [$self hbh-mft-size? $src $grp]
	    if { $tam == 1 } {
		array set resp [$self hbh-read-mft $src $grp]
		set dst_node [array names resp]
		$self dbg "dkos: $dst_node"
		$self hbh-destroy-mft-entry $src $grp $dst_node
		$self hbh-destroy-mft $src $grp
		$self hbh-create-mct $src $grp $dst_node
	    }
	}

    } else {
	$self err "hbh-destroy-mft-entry error: no replicator for $src:$grp"
    }
}    

Node instproc hbh-create-mct { src grp node } {
    $self instvar ns_ mct_ mct_fg_ mct_timer_ mct_destroy_timer_

    global timeout1_ timeout2_

    $self dbg "hbh-create-mct $src $grp $node"

    set mct_($src:$grp) $node
    set mct_fg_($src:$grp) 1
    
    #
    # Stale Timer creation
    #
    set mct_timer_($src:$grp) [new SkelTimer/MCT $ns_ $self $src $grp]
    $mct_timer_($src:$grp) sched $timeout1_
    #
    # Destroy Timer creation
    #
    set mct_destroy_timer_($src:$grp) \
	    [new SkelTimer/MCT/Destroy $ns_ $self $src $grp]
    $mct_destroy_timer_($src:$grp) sched $timeout2_
}

Node instproc hbh-exist-mct { src grp } {
    $self instvar mct_
    if [info exists mct_($src:$grp)] {
	return 1
    } else { return 0 }
}

Node instproc hbh-read-mct { src grp } {
    $self instvar mct_
    if [info exists mct_($src:$grp)] {
	return $mct_($src:$grp)
    } else { 
	$self err "hbh-read-mct error: no mct_($src:$grp)" 
    }
}

Node instproc hbh-exist-mct-entry {src grp node} {
    $self instvar mct_ ns_

    if [info exists mct_($src:$grp)] {
	if { $mct_($src:$grp) == $node } {
	    return 1
	} else { return 0 }
    } else {
	$self err "hbh-exist-mct-entry error: no replicator for $src:$grp"
    }
}

Node instproc hbh-destroy-mct { src grp } {
    $self instvar mct_ mct_fg_ mct_timer_ mct_destroy_timer_

    unset mct_($src:$grp)
    unset mct_fg_($src:$grp)
    $mct_timer_($src:$grp) destroy
    delete $mct_timer_($src:$grp)
    unset mct_timer_($src:$grp)
    $mct_destroy_timer_($src:$grp) destroy
    delete $mct_destroy_timer_($src:$grp)
    unset mct_destroy_timer_($src:$grp)
}

Node instproc hbh-refresh-mct {src grp} {
    $self instvar mct_ mct_fg_ mct_timer_ mct_destroy_timer_

    global timeout1_ timeout2_

    $self dbg "hbh-mct-refresh"

    set mct_fg_($src:$grp) 1
    $mct_timer_($src:$grp) cancel
    $mct_timer_($src:$grp) sched $timeout1_
    $mct_destroy_timer_($src:$grp) cancel
    $mct_destroy_timer_($src:$grp) sched $timeout2_
}

Node instproc hbh-mct-fresh? { src grp } {
    $self instvar mct_fg_
    $self dbg "hbh-mct-fresh?"
    return $mct_fg_($src:$grp)
}

Node instproc hbh-dump-tables { file } {
    $self instvar ns_ node_ mct_ hbhreplicator_
    $self instvar mft_fg_ mft_entry_fg_ mft_entry_mk_

    if {$ns_ != ""} {
	set time [$ns_ now]
    } else {
	set time 0.0
    }

    puts $file [concat "Node: ([$self id])\tat t =" [format "%4.2f" $time]]
    puts $file "---MCT---"
    puts $file "<source,group>\trecv"
    foreach mcti [array names mct_] {
	puts $file "<$mcti>\t$mct_($mcti)"
    }

    puts $file "---MFT---"
    puts $file "<source,group>\tFlag\trecv(F-M)"
    foreach mfti [array names hbhreplicator_] {
	puts -nonewline $file "<$mfti>\t $mft_fg_($mfti)\t"
	foreach recv [$hbhreplicator_($mfti) slots] {
	    puts -nonewline $file "  $recv\
		    ($mft_entry_fg_($mfti:$recv)-$mft_entry_mk_($mfti:$recv))"
	}
    }
    
    puts $file ""
    puts $file "--------------------------------------------------------------"
}

# Classifier/Replicator/HBH

Classifier/Replicator/HBH instproc init { ns node src grp } {
    $self instvar table_timer_
    $self next

    puts stderr "Classifier/Replicator/HBH init"
}    

Agent/HBH instproc init { proto } {
    global rtglibRNG
    $self instvar proto_
    $self set proto_ $proto
    puts stderr "Agent HBH init: $proto"
    $self next $proto
}

Agent/HBH instproc recv-control { from id } {
    $self instvar node_ proto_

    #puts stderr "Node [$node_ id] testando recv-control - msgid: $id"

    set msg [HBH retrieve-msg $id]
    array set message $msg
    #puts stderr "mensagem: $msg"
    set type      $message(0)
    set src       $message(1)
    set grp       $message(2)
    set from_node $message(3)
    set to_node   $message(4)
    #puts stderr "message: [array names message] type: $type src: $src grp: $grp fid: $message(5)"
    if { $type == "fusion" } {
	$proto_ recv-$type $src $grp $from_node $to_node $message(5)
    } else {
	$proto_ recv-$type $src $grp $from_node $to_node
    }
}

HBH proc get-next-mid {} {
    set ret [HBH set mid_]
    HBH set mid_ [expr $ret + 1]
    set ret
}

HBH proc retrieve-msg id {
    set ret [HBH set msg_($id)]
    HBH unset msg_($id)
    set ret
}

HBH proc read-msg id {
    set ret [HBH set msg_($id)]
#    HBH unset msg_($id)
    set ret
}

HBH instproc send-periodic-join { src grp } {
    global rtglibRNG
    $self instvar node_ joined_
    set ns_ [Simulator instance]

    $self dbg "send-periodic-join entrada"

    if { $joined_($src:$grp) } {
	$self join $src $grp [$node_ id]
	set join_updateTime [expr [$ns_ now] + \
		([$class set advertInterval] * [$rtglibRNG uniform 0.9 1.1])]
	$ns_ at $join_updateTime "$self send-periodic-join $src $grp"
    }
}

HBH instproc send-periodic-tree { src grp } {
    global rtglibRNG
    $self instvar proto_

    set ns_ [Simulator instance]
    $self dbg "send-periodic-tree entrada"

    #
    # Send a tree message to all non-stale receivers in MFT
    #
    $self send-all-mft-tree $src $grp

    set tree_updateTime [expr [$ns_ now] + \
	    ([$class set advertInterval] * [$rtglibRNG uniform 0.9 1.1])]
    $ns_ at $tree_updateTime "$self send-periodic-tree $src $grp"
}

Class SkelTimer/MFT -superclass SkelTimer

SkelTimer/MFT instproc init { ns node src grp } {
    $self instvar node_ src_ grp_
    $self next $ns $node $src $grp

    $self dbg "MFT Table Timer created"
}

SkelTimer/MFT instproc timeout {} {
    $self instvar node_ src_ grp_

    $self dbg "MFT Timeout!"
    $node_ set mft_fg_($src_:$grp_) 0
}

Class SkelTimer/MFTentry -superclass SkelTimer

SkelTimer/MFTentry instproc init { ns node src grp recv } {
    $self instvar node_ src_ grp_ recv_
    $self next $ns $node $src $grp

    $self dbg "MFT Entry Timer created"
    set recv_ $recv
}

SkelTimer/MFTentry instproc timeout {} {
    $self instvar node_ src_ grp_ recv_

    $node_ set mft_entry_fg_($src_:$grp_:$recv_) 0
    $self dbg "MFT Entry Timeout! $src_:$grp_:$recv_"
}

Class SkelTimer/MFTentry/Destroy -superclass SkelTimer/MFTentry

SkelTimer/MFTentry/Destroy instproc timeout {} {
    $self instvar node_ src_ grp_ recv_

    $self dbg "MFT Entry Destroy! $src_:$grp_:$recv_"
    $node_ hbh-destroy-mft-entry $src_ $grp_ $recv_

    # if there is only one entry in MFT, destroy MFT and put this entry
    # in the MCT - exception: source node...
    #if { ([$node_ hbh-mft-size? $src_ $grp_] == 1) && ($src_ != $node_) } {
#	$self dbg "MFT Entry Destroy, size = 1"
#	array set resp [$node_ hbh-read-mft $src_ $grp_]
#	set dst_node [array names resp]
#	$self dbg "kos: $dst_node"
#	$node_ hbh-create-mct $src_ $grp_ $dst_node
#	$node_ hbh-destroy-mft-entry $src_ $grp_ $dst_node
#	$node_ hbh-destroy-mft $src_ $grp_
#    }

}
    
Class SkelTimer/MCT -superclass SkelTimer

SkelTimer/MCT instproc init { ns node src grp } {
    $self instvar node_ src_ grp_
    $self next $ns $node $src $grp

    $self dbg "MCT Table Timer created"
}

SkelTimer/MCT instproc timeout {} {
    $self instvar node_ src_ grp_

    $self dbg "MCT Timeout!"
    $node_ set mct_fg_($src_:$grp_) 0
}

Class SkelTimer/MCT/Destroy -superclass SkelTimer/MCT

SkelTimer/MCT/Destroy instproc timeout {} {
    $self instvar node_ src_ grp_

    $self dbg "MCT Destroy! $src_:$grp_"
    $node_ hbh-destroy-mct $src_ $grp_
}

Class SkelTimer/Mark -superclass SkelTimer

SkelTimer/Mark instproc init { ns node src grp recv } {
    $self instvar node_ src_ grp_ recv_
    $self next $ns $node $src $grp

    set recv_ $recv
}

SkelTimer/Mark instproc timeout {} {
    $self instvar node_ src_ grp_ recv_

    $self dbg "Mark Timeout! ($src_:$grp_:$recv_)"
    $node_ hbh-unmark-mft-entry $src_ $grp_ $recv_
}

Class SkelTimer/MCT/Destroy -superclass SkelTimer/MCT

SkelTimer/MCT/Destroy instproc timeout {} {
    $self instvar node_ src_ grp_

    $self dbg "MCT Destroy! $src_:$grp_"
    $node_ hbh-destroy-mct $src_ $grp_
}


Node instproc join-group { agent group { src "" } } {
    $self instvar replicator_ Agents_ mrtObject_

    set group [expr $group] ;# use expr to convert to decimal
    $mrtObject_ join-group $group $src
}

Node instproc leave-group { agent group { src "" } } {
    $self instvar replicator_ Agents_ mrtObject_

    $mrtObject_ leave-group $group $src
}
