#
# tcl/lip6/REU.tcl
#

set timeout1_  1
set timeout2_  1

Class REU -superclass McastProtocol

REU set advertInterval 0.9

set rtglibRNG [new RNG]
$rtglibRNG seed 1

REU set mid_ 0

REU instproc init { sim node } {
    $self instvar mctrl_ oiflist_
    $self instvar reu_agent

    set reu_agent [new Agent/REU $self]
    $node attach $reu_agent

    set mctrl_ [new Agent/Mcast/Control $self]
    $node attach $mctrl_
    $self next $sim $node
}

REU instproc join-group  { grp src } {
    $self instvar node_ ns_ joined_

    $self dbg "Sending REU join-group < $src ([$src id]), $grp, [$node_ id] > "

    # We create a MFT with one entry (itself) at the endnodes.
    # This entry forwards the packets to the agent that joined at this node.

    $node_ reu-create-mft $src $grp [$node_ id]
    $node_ reu-add-mft-entry $src $grp [$node_ id]

    # At the end-nodes we don't need timers to destroy the MFT entry 
    # (which is unique). We destroy this entry when the member leaves 
    # the group.

    [$node_ set mft_entry_timer_($src:$grp:[$node_ id])] cancel
    [$node_ set mft_timer_($src:$grp)] cancel

    set joined_($src:$grp) 1
    #
    # Send a periodic join concerning the agents joined locally
    #
    #set join_updateTime [$ns_ now]
    #$ns_ at $join_updateTime "$self send-periodic-join $src $grp"
    $self send-periodic-join $src $grp
}

REU instproc leave-group { grp src } {
    $self instvar joined_ node_ ns_
    $self dbg "testando leave-group"
    set joined_($src:$grp) 0

    $node_ reu-destroy-mft-entry $src $grp [$node_ id]
    $node_ reu-destroy-mft $src $grp
}

REU instproc join { src grp join_node } {
    $self instvar node_

    #send internal join msg

    $self send-ctrl "join" $src $grp $join_node [$src id]
    $self dbg "sent join: < $src $grp $join_node [$src id]>"
}

REU instproc recv-join { src grp from_node to_node } {
    $self instvar node_ ns_ 
    
    #join msg processing
    #called like this: $proto recv-$type $src $grp $from_node
    
    $self dbg "received join <$src $grp $from_node $to_node> at [$ns_ now]"

    # If I am the source...
    ###########################################################################

    if { $node_ == $src } {
	$self dbg "I am the source"
	# Create the MFT if it does not exist
	#######################################################################
	if { [$node_ reu-exist-mft $src $grp] == 0 } {
	    $node_ reu-create-mft $src $grp $to_node  ;# to_node == src
	    $node_ reu-add-mft-entry $src $grp $from_node
	    $self dbg "creating mft at the source"
	    #
	    # Send a periodic tree to live branches (in MFT)
	    # Start sending tree msgs
	    ###################################################################
	    #set tree_updateTime [$ns_ now]
	    #$ns_ at $tree_updateTime "$self send-periodic-tree $src $grp"
	    $self send-periodic-tree $src $grp

	} else {

	    # Add the node to MFT if it is not there, refresh it otherwise
	    ###################################################################
	    if { [$node_ reu-exist-mft-entry $src $grp $from_node] == 0 } {
		$node_ reu-add-mft-entry $src $grp $from_node
		$node_ reu-refresh-mft $src $grp
	    } else {
		$node_ reu-refresh-mft-entry $src $grp $from_node
		$node_ reu-refresh-mft $src $grp
	    }
	}

    } else {

	# If I am not the source, ...
	# Does MFT<S> exist?
	#######################################################################
	if [$node_ reu-exist-mft $src $grp] {
	    #
	    # Yes, MFT<S> exists
	    #
	    # R = MFT<S>.dst?
	    ###################################################################
	    if { $from_node == [$node_ set mft_dst_($src:$grp)] } {
		# Yes, R = MFT<S>.dst
		# forward join message
		$self send-ctrl "join" $src $grp $from_node $to_node
		$self dbg "join saida1"
	    } else {
		# No, R != MFT<S>.dst
		# Is MFT<S> stale?
		###############################################################
		if { [$node_ reu-mft-fresh? $src $grp] == 0 } {
		    # Yes, MFT<S> stale
		    #
		    # forward join message
		    $self send-ctrl "join" $src $grp $from_node $to_node
		    $self dbg "join saida2"
		} else {
		    # No, MFT<S> is not stale
		    #
		    # Is R in MFT<S>.rcv_list?
		    ###########################################################
		    if [$node_ reu-exist-mft-entry $src $grp $from_node] {
			# Yes, MFT<S>.R exists
			# Refresh MFT[S].rcv_list[R]
			#######################################################
			$node_ reu-refresh-mft-entry $src $grp $from_node
			##$node_ reu-refresh-mft $src $grp  ;# luis should I?
			$self dbg "join saida3"
		    } else {
			# No, R is not in MFT<S>
			#
			# Insert R in MFT[S].rcv_list
			#######################################################
			$node_ reu-add-mft-entry $src $grp $from_node
			#$node_ reu-refresh-mft $src $grp  ;# luis should I?
			$self dbg "join saida4"
		    }
		}
	    }
	} else {
	    # No MFT<S>
	    #
	    # Does MCT<S> exist?
	    ###################################################################
	    if { [$node_ reu-exist-mct $src $grp] } {
		# Yes, MCT<S> exists
		#
		# R = MCT<S>.dst?
		if { $from_node == [$node_ reu-read-mct $src $grp] } {
		    # Yes, R = MCT<S>.dst
		    #
		    # Forward join message
		    ###########################################################
		    $self send-ctrl "join" $src $grp $from_node $to_node
		    $self dbg "join saida5"
		} else {
		    # No, R != MCT<S>.dst
		    #
		    # Insert S, MCT<S>.dst in MFT
		    # Insert R in MFT<S>.rcv_list
		    ###########################################################
		    set dst_node [$node_ reu-read-mct $src $grp]
		    ##########$node_ reu-destroy-mct $src $grp
		    $node_ reu-create-mft $src $grp $dst_node
		    $node_ reu-add-mft-entry $src $grp $from_node
		    $self dbg "join saida6"
		}
	    } else {
		# No MCT<S>
		#
		# Forward join(src,from_node) unchanged
		###############################################################
		$self send-ctrl "join" $src $grp $from_node $to_node
		$self dbg "join saida7"
	    }
	}
    }
}

REU instproc tree { src grp to_node } {
    $self instvar node_ reu_agent
    # send tree msg
    $self dbg "sent tree <$src $grp [$node_ id] $to_node>"
    $self send-ctrl "tree" $src $grp [$node_ id] $to_node
}

REU instproc send-all-mft-tree { src grp } {
    $self instvar node_

    array set reureplicator_ [$node_ array get reureplicator_]
    array set mft_entry_fg_ [$node_ array get mft_entry_fg_]
    array set mft_fg_ [$node_ array get mft_fg_]
    
    if { ![info exists reureplicator_($src:$grp)] } {
	return
    }
    
    foreach recv [$reureplicator_($src:$grp) slots] {
	if { $mft_entry_fg_($src:$grp:$recv) } { 
	    set stale 0 
	} else { 
	    puts stderr "entry stale"
	    set stale 1 
	}
	# luis: se a mft esta stale....
	if { !$mft_fg_($src:$grp) } { 
	    $self dbg "MFT stale"
	    set stale 1 
	}   
	if { [$node_ id] != $recv } {
	    $self send-ctrl "tree" $src $grp [$src id] $recv $stale
	}
    }

    # luis: se a mft esta stale....
    if { !$mft_fg_($src:$grp) } { 
	$self dbg "MFT stale"
	set stale 1 
    } else {
	set stale 0
    }

    if { $node_ != $src } {
	$self send-ctrl "tree" $src $grp [$src id] [$node_ set mft_dst_($src:$grp)] $stale
    }
}

REU instproc recv-tree { src grp from_node to_node stale } {
    $self instvar node_
    # tree msg processing
    $self dbg "received tree <$src $grp $from_node $to_node>"

    # Am I R? (That is, I am an end-node receiver)
    ###########################################################################
    if { [$node_ id] == $to_node } {
	# Yes, I'm R
	$self dbg "I am R in recv-tree"
	if [$node_ reu-exist-mft $src $grp] {
	    # nothing to do, I am a end-node receiver
	} else {
	    $self err "I am R in recv-tree but I don't have a MFT"
	}

    } else {

	# No, I'm not R
	# Does MFT<S> exist?
	#######################################################################
	if [$node_ reu-exist-mft $src $grp] {	    
	    # Yes, MFT<S> exists
	    #
	    # TREE msg stale?
	    ###################################################################
	    if { $stale } {
		# Yes, msg stale
		#
		# R = MFT<S>.dst?
		###############################################################
		if { $to_node == [$node_ set mft_dst_($src:$grp)] } {
		    # Yes, R = MFT<S>.dst
		    #
		    # MFT<S>.stale = 1
		    # Set timer T02
		    # Forward tree msg
		    ###########################################################
		    [$node_ set mft_timer_($src:$grp)] expire
		    $self send-ctrl "tree" $src $grp $from_node $to_node $stale
		    $self dbg "tree saida1"
		} else {
		    # No, R != MFT<S>.dst
		    #
		    # Forward tree msg
		    ###########################################################
		    if [$node_ reu-exist-mft-entry $src $grp $to_node] {
			#$node_ set mft_entry_fg_($src:$grp:$to_node) 0
			set fr [$node_ reu-mft-entry-fresh? $src $grp $to_node]
			$self dbg "UGA UGA fr: $fr"
			#$self send-all-mft-tree $src $grp
		    } 
		    $self send-ctrl "tree" $src $grp $from_node $to_node $stale
		    $self dbg "tree saida2"
		}
	    } else {
		# No, msg is not stale
		# R = MFT<S>.dst?
		###############################################################
		if { $to_node == [$node_ set mft_dst_($src:$grp)] } {
		    # Yes, R = MFT<S>.dst 
		    #
		    # MFT<S>.stale = 0
		    # set timer T01
		    # forward tree msg
		    ###########################################################
		    $node_ reu-refresh-mft $src $grp     ;# also sets timer t01
		    $self send-all-mft-tree $src $grp
		    $self dbg "tree saida3"
		} else {
		    # No, R != MFT<S>.dst 
		    # MFT<S> stale?
		    ###########################################################
		    if { [$node_ reu-mft-fresh? $src $grp] == 0 } {
			# Yes, MFT<S> stale  (1)
			# 
			# MFT<S>.dst = R
			# MFT<S>.stale = 0
			# Set timer T01
			# Remove MFT<S>.rcv_list[R]
			# Forward tree msg
			#######################################################
			#$node_ set mft_dst_($src:$grp) $to_node
			$node_ reu-write-mft-dst $src $grp $to_node
			$node_ reu-refresh-mft $src $grp ;# also set timer t01
			# luis
			if [$node_ reu-exist-mft-entry $src $grp $to_node] {
			    $node_ reu-destroy-mft-entry $src $grp $to_node
			}
			# testing luis
			if { [$node_ reu-mft-size? $src $grp] == 0 } {
			    #set dst_node [$node_ set mft_dst_($src:$grp)]
			    $node_ reu-insert-mct-entry $src $grp $to_node
			    $node_ reu-destroy-mft $src $grp
			    $self dbg "tree saida4.0"
			}
			$self send-ctrl "tree" $src $grp $from_node $to_node $stale
		    } else {
			# No, MFT<S> is not stale
			#
			# R in MFT<S>.rcv_list && !MFT<S>.rcv_list[R].alive?
			#######################################################
			if { [$node_ reu-exist-mft-entry $src $grp $to_node] \
				&& ![$node_ reu-mft-entry-fresh? $src $grp $to_node] } {
			    # Yes
			    #
			    # Remove MFT<S>.rcv_list[R]
			    # Forward tree msg
			    ###################################################
			    $node_ reu-destroy-mft-entry $src $grp $to_node

			    # testing luis
			    if { [$node_ reu-mft-size? $src $grp] == 0 } {
				set dst_node [$node_ set mft_dst_($src:$grp)]
				$node_ reu-insert-mct-entry $src $grp $to_node
				$node_ reu-destroy-mft $src $grp
				$self send-ctrl "tree" $src $grp $from_node $to_node $stale
				$self dbg "tree saida5.0"
			    } else {
				$self send-all-mft-tree $src $grp
				$self send-ctrl "tree" $src $grp $from_node $to_node $stale
				$self dbg "tree saida5.1"
			    }
			} else {
			    # No
			    #
			    # Forward tree msg
			    ##################################################
			    $self send-ctrl "tree" $src $grp $from_node $to_node $stale
			    $self dbg "tree saida6"
			}
		    }
		}
	    }
	} else {
	    # No, MFT<S> does not exist
	    #
	    # TREE msg stale?
	    ###################################################################
	    if { $stale } {
		# Yes, msg stale
		#
		# Remove MCT<S>
		# Forward tree msg
		###############################################################
		if { [$node_ reu-read-mct $src $grp] == $to_node } {
		    $node_ reu-destroy-mct $src $grp
		} else {
		    if [$node_ reu-exist-mct-entry $src $grp $to_node] {
			$node_ reu-destroy-mct-entry $src $grp $to_node
		    }
		}
		$self send-ctrl "tree" $src $grp $from_node $to_node $stale
		$self dbg "tree saida7"
	    } else {
		# No, msg is not stale
		#
		# Insert S,R in MCT<S>
		# Forward tree msg
		###############################################################
		$node_ reu-insert-mct-entry $src $grp $to_node
		$self send-ctrl "tree" $src $grp $from_node $to_node $stale
		$self dbg "tree saida8"
	    }
	}
    }
}

REU instproc drop { replicator src dst iface} {
    $self instvar node_ ns_
    puts stderr "testando drop"
}

REU instproc send-ctrl { which src grp from_node to_node {fid ""} } {
    $self instvar mctrl_ ns_ node_
    $self instvar reu_agent
    
    set message(0) $which
    set message(1) $src
    set message(2) $grp
    set message(3) $from_node
    set message(4) $to_node
    # luis: trocar embaixo por um switch
    if { $which == "tree" } {
	set message(5) $fid
    } else { set message(5) 0 }
    set id [$class get-next-mid]
    $class set msg_($id) [array get message]
    $reu_agent send-control $to_node 0 $id [array size message]]
}

################ Helpers

REU instproc dbg arg {
	$self instvar ns_ node_
	puts [format "At %.4f : Node([$node_ id]) $arg" [$ns_ now]]
}

REU instproc err arg {
    $self instvar ns_ node_
    puts [format "Error at %.4f : Node([$node_ id]) $arg" [$ns_ now]]
}

Node instproc dbg arg {
    $self instvar ns_
    puts [format "At %.4f : Node([$self id]) $arg" [$ns_ now]]
}

Node instproc err arg {
    $self instvar ns_
    puts [format "Error at %.4f : Node([$self id]) $arg" [$ns_ now]]
}

Class SkelTimer -superclass Timer

SkelTimer instproc init { ns node src grp } {
    $self instvar node_ src_ grp_
    $self next $ns

    set ns_ $ns
    set node_ $node
    set src_ $src
    set grp_ $grp
}

SkelTimer instproc dbg arg {
    $self instvar ns_ node_
    puts [format "At %.4f : Node([$node_ id]) $arg" [$ns_ now]]
}

########
# Table manipulation
########

Classifier/REU instproc init args {
    $self next
    $self instvar nrep_
    set nrep_ 1  
    # luish: 1 because slot 0 is reserved to the unicast classifer
}

Classifier/REU instproc add-rep { rep src group } {
    $self instvar nrep_
    $self set-hash [$src id] $group $nrep_
    $self install $nrep_ $rep
    puts stderr "reureplicator installed: $nrep_ $rep"
    incr nrep_
}

Node instproc reu-create-mft { src grp dst_node } {

    # multiclassifier_ is a reference to the Multicast/REU Classifier
    # reureplicator_ is an array indexed by ($src:$grp)

    $self instvar ns_ reuclassifier_ reureplicator_ Agents_ 
    $self instvar mft_fg_ mft_dst_ mft_timer_ mft_destroy_timer_
    $self instvar mft_entry_fg_ mft_entry_timer_ mft_entry_destroy_timer_
    global timeout1_ timeout2_

    $self dbg "reu-create-mft: $src $grp $dst_node"

    if [info exists reureplicator_($src:$grp)] {
	set r $reureplicator_($src:$grp)
    } else {
	set r [new Classifier/Replicator/REU $ns_ $self $src $grp]
	$r set srcID_ $src
	$r set grp_ $grp
	set reureplicator_($src:$grp) $r
	$r set node_ $self
	#
	# Install the replicator.  
	#
	# luish - instala-lo em reuclassifier_ 
	$reuclassifier_ add-rep $r $src $grp
    }
    
    set mft_timer_($src:$grp) [new SkelTimer/MFT $ns_ $self $src $grp]
    $mft_timer_($src:$grp) sched $timeout1_

    $self reu-write-mft-dst $src $grp $dst_node

    set mft_fg_($src:$grp) 1
}

Node instproc reu-write-mft-dst { src grp dst_node } {
    $self instvar reuclassifier_ mft_dst_

    set mft_dst_($src:$grp) $dst_node
    $reuclassifier_ installDst $dst_node
}

Node instproc reu-exist-mft { src grp } {
    # multiclassifier_ is a reference to the Multicast/REU Classifier
    # reureplicator_ is an array indexed by ($src:$grp)

    $self instvar multiclassifier_ reureplicator_

    $self dbg "reu-exist-mft: $src $grp"

    return [info exists reureplicator_($src:$grp)]
}

Node instproc reu-destroy-mft { src grp } {
    $self instvar reureplicator_ reuclassifier_ mft_dst_ mft_fg_
    $self instvar mft_timer_ mft_destroy_timer_

    $self dbg "reu-destroy-mft $src:$grp"

    if [ $self reu-exist-mft $src $grp ] {
	# Destroy all entries
	array set mft [$self reu-read-mft $src $grp]
	$self dbg "destroy-mft entries: [array names mft]"
	foreach recv [array names mft] {
	    $self reu-destroy-mft-entry $src $grp $recv
	}

	if { [info exists mft_timer_($src:$grp)] } {
	    $self dbg "existe luis 1"
	    $mft_timer_($src:$grp) destroy
	    delete $mft_timer_($src:$grp)
	    unset mft_timer_($src:$grp)
	}
	if { [info exists mft_destroy_timer_($src:$grp)] } {
	    $self dbg "existe luis 2"
	    $mft_destroy_timer_($src:$grp) destroy
	    delete $mft_destroy_timer_($src:$grp)
	    unset mft_destroy_timer_($src:$grp)
	}

	unset mft_dst_($src:$grp)
	unset mft_fg_($src:$grp)

	#luish:  falta tirar a referencia a rep de reuclassifier...
	$reuclassifier_ delete-node $reureplicator_($src:$grp)

	delete $reureplicator_($src:$grp)
	unset reureplicator_($src:$grp)
    }
}

Node instproc reu-read-mft { src grp } {
    $self instvar reureplicator_
    array set resp [$reureplicator_($src:$grp) slot-pairs]
    return [array get resp]
}

Node instproc reu-refresh-mft { src grp } {
    $self instvar mft_fg_ mft_timer_ mft_destroy_timer_
    global timeout1_

    $self dbg "reu-refresh-mft $src $grp"

    set mft_fg_($src:$grp) 1
    $mft_timer_($src:$grp) sched $timeout1_

    # luis: if there is a destroy timer...
    if { [info exists mft_destroy_timer_($src:$grp)] } {
	$self dbg "luis passei aqui2 $src:$grp"
	$mft_destroy_timer_($src:$grp) destroy
	delete $mft_destroy_timer_($src:$grp)
	unset mft_destroy_timer_($src:$grp)
    }
}

Node instproc reu-mft-fresh? { src grp } {
    $self instvar mft_fg_
    $self dbg "reu-mft-fresh?"
    return $mft_fg_($src:$grp)
}

Node instproc reu-mft-entry-fresh? { src grp node } {
    $self instvar mft_entry_fg_
    $self dbg "reu-mft-entry-fresh?"
    return $mft_entry_fg_($src:$grp:$node)
}

Node instproc reu-mft-size? { src grp } {
    $self instvar reureplicator_
    array set lixo [$reureplicator_($src:$grp) slot-pairs]
    return [array size lixo]
}

Node instproc reu-add-mft-entry {src grp node} {
    $self instvar ns_ classifier_ reureplicator_ mft_entry_fg_ 
    $self instvar mft_entry_timer_
    global timeout1_

    if [info exists reureplicator_($src:$grp)] {
	$reureplicator_($src:$grp) install $node $classifier_
	set mft_entry_fg_($src:$grp:$node) 1
	#
	# Stale Timer creation
	#
	set mft_entry_timer_($src:$grp:$node) \
		[new SkelTimer/MFTentry $ns_ $self $src $grp $node]
	$mft_entry_timer_($src:$grp:$node) sched $timeout1_
    } else {
	$self err "reu-add-mft-entry error: no replicator for $src:$grp"
    }
    $self dbg "reu-add-mft-entry: $src $grp $node"
}

Node instproc reu-exist-mft-entry {src grp node} {
    $self instvar reureplicator_ ns_

    $self dbg "Node [$self id] reu-exist-mft-entry: $src $grp $node"

    if [info exists reureplicator_($src:$grp)] {
	array set x [$reureplicator_($src:$grp) slot-pairs]
	foreach i [array names x] {
	    if { $i == $node } { return 1 }
	}
	return 0
    } else {
	$self err "reu-exist-mft-entry error: no replicator for $src:$grp"
    }
}

Node instproc reu-refresh-mft-entry {src grp node} {
    $self instvar mft_entry_fg_ mft_ mft_entry_timer_ mft_entry_destroy_timer_
    global timeout1_ timeout2_

    $self dbg "reu-refresh-mft-entry $src $grp $node"

    $mft_entry_timer_($src:$grp:$node) sched $timeout1_
    set mft_entry_fg_($src:$grp:$node) 1

    ###$self reu-refresh-mft $src $grp   ;# ??????  luis

    # luis: if there is a destroy timer...
    if { [info exists mft_entry_destroy_timer_($src:$grp:$node)] } {
	$mft_entry_destroy_timer_($src:$grp:$node) destroy
	delete $mft_entry_destroy_timer_($src:$grp:$node)
	unset mft_entry_destroy_timer_($src:$grp:$node)
    }
}

Node instproc reu-destroy-mft-entry {src grp node} {
    $self instvar ns_ reureplicator_ mft_entry_fg_
    $self instvar mft_entry_timer_ mft_entry_destroy_timer_
    
    $self dbg "reu-destroy-mft-entry $src:$grp:$node" 

    if [info exists reureplicator_($src:$grp)] {
	$reureplicator_($src:$grp) clear $node
	$mft_entry_timer_($src:$grp:$node) cancel
	delete $mft_entry_timer_($src:$grp:$node)
	unset mft_entry_timer_($src:$grp:$node)
	if { [info exists mft_entry_destroy_timer_($src:$grp:$node)] } {
	    $self dbg "luis passei aqui $src:$grp:$node"
	    $mft_entry_destroy_timer_($src:$grp:$node) cancel
	    delete $mft_entry_destroy_timer_($src:$grp:$node)
	    unset mft_entry_destroy_timer_($src:$grp:$node)
	}
	unset mft_entry_fg_($src:$grp:$node)
	array set lixo [$reureplicator_($src:$grp) slot-pairs]
	$self dbg "mft size: [array size lixo]"

    } else {
	$self err "reu-destroy-mft-entry error: no replicator for $src:$grp"
    }
}    

Node instproc reu-create-mft-t02 { src grp } {
    $self instvar ns_ mft_destroy_timer_
    global timeout2_

    $self dbg "reu-create-mft-t02 $src $grp"

    set mft_destroy_timer_($src:$grp) [new SkelTimer/MFT/Destroy $ns_ $self $src $grp]
    $mft_destroy_timer_($src:$grp) sched $timeout2_
}

Node instproc reu-create-mft-entry-t02 { src grp node } {
    $self instvar ns_ mft_entry_destroy_timer_
    global timeout2_

    $self dbg "reu-create-mft-entry-t02 $src $grp $node"

    set mft_entry_destroy_timer_($src:$grp:$node) \
	    [new SkelTimer/MFTentry/Destroy $ns_ $self $src $grp $node]
    $mft_entry_destroy_timer_($src:$grp:$node) sched $timeout2_
}

#
# MCT
#
Node instproc reu-create-mct { src grp node } {
    $self instvar ns_ mct_ mct_timer_
    global timeout1_

    $self dbg "reu-create-mct $src $grp $node"

    set mct_($src:$grp) $node                             ;# this is mct<S>.dst
    set mct_timer_($src:$grp) [new SkelTimer/MCT $ns_ $self $src $grp]
    $mct_timer_($src:$grp) sched $timeout1_
}

Node instproc reu-insert-mct-entry { src grp node } {
    $self instvar ns_ mct_ mct_timer_ mct_entry_ mct_entry_timer_
    global timeout1_

    $self dbg "reu-insert-mct $src $grp $node"

    if [$self reu-exist-mct $src $grp] {
	if { [$self reu-read-mct $src $grp] == $node } {
	    $self reu-refresh-mct $src $grp
	} else {
	    if [info exists mct_entry_($src:$grp:$node)] {
		$mct_entry_timer_($src:$grp:$node) sched $timeout1_
	    } else {
		set mct_entry_($src:$grp:$node) $node
		set mct_entry_timer_($src:$grp:$node) \
			[new SkelTimer/MCTentry $ns_ $self $src $grp $node]
		$mct_entry_timer_($src:$grp:$node) sched $timeout1_
	    }
	}
    } else {
	$self reu-create-mct $src $grp $node
    }
}

Node instproc reu-exist-mct { src grp } {
    $self instvar mct_
    if [info exists mct_($src:$grp)] {
	return 1
    } else { return 0 }
}

Node instproc reu-exist-mct-entry { src grp node } {
    $self instvar mct_entry_
    if [info exists mct_entry_($src:$grp:$node)] {
	return 1
    } else { return 0 }
}

Node instproc reu-read-mct { src grp } {
    $self instvar mct_
    if [info exists mct_($src:$grp)] {
	return $mct_($src:$grp)
    } else { 
	$self err "reu-read-mct error: no mct_($src:$grp)" 
    }
}

Node instproc reu-write-mct { src grp node } {
    $self instvar mct_
    $self dbg "reu-write-mct $src $grp $node"
    set mct_($src:$grp) $node        ;# this is mct<S>.dst
}

Node instproc reu-refresh-mct { src grp } {
    $self instvar mct_ mct_timer_
    global timeout1_
   
    $self dbg "reu-refresh-mct $src $grp"

    $mct_timer_($src:$grp) sched $timeout1_
}

Node instproc reu-destroy-mct { src grp } {
    $self instvar mct_ mct_timer_ mct_entry_ mct_entry_timer_

    unset mct_($src:$grp)
    $mct_timer_($src:$grp) destroy
    delete $mct_timer_($src:$grp)
    unset mct_timer_($src:$grp)

    # destroy all MCT entries
    foreach recv [array names mct_entry_] {
	unset mct_entry_($recv)
	$mct_entry_timer_($recv) destroy
	delete $mct_entry_timer_($recv)
	unset mct_entry_timer_($recv)
    }

}

Node instproc reu-destroy-mct-entry { src grp node } {
    $self instvar mct_entry_ mct_entry_timer_

    unset mct_entry_($src:$grp:$node)
    $mct_entry_timer_($src:$grp:$node) destroy
    delete $mct_entry_timer_($src:$grp:$node)
    unset mct_entry_timer_($src:$grp:$node)
}

Node instproc reu-dump-tables { file } {
    $self instvar ns_ node_ mct_ reureplicator_
    $self instvar mft_dst_ mft_fg_ mft_entry_fg_

    if {$ns_ != ""} {
	set time [$ns_ now]
    } else {
	set time 0.0
    }

    puts $file [concat "Node: $self ([$self id])\tat t =" [format "%4.2f" $time]]
    puts $file "---MCT---"
    puts $file "<source,group>\trecv"
    foreach mcti [array names mct_] {
	puts $file "<$mcti>\t$mct_($mcti)"
    }

    puts $file "---MFT---"
    puts $file "<source,group>\tFlag\trecv(F-M)"
    foreach mfti [array names reureplicator_] {
	puts -nonewline $file "<$mfti>\t $mft_dst_($mfti) $mft_fg_($mfti)\t"
	foreach recv [$reureplicator_($mfti) slots] {
	    puts -nonewline $file "  $recv\
		    $mft_entry_fg_($mfti:$recv)"
	}
    }
    
    puts $file ""
    puts $file "--------------------------------------------------------------"
}

# Classifier/Replicator/REU

Classifier/Replicator/REU instproc init { ns node src grp } {
    $self instvar table_timer_
    $self next

    puts stderr "Classifier/Replicator/REU init"
}    

Agent/REU instproc init { proto } {
    $self instvar proto_
    $self set proto_ $proto
    puts stderr "Agent REU init: $proto"
    $self next $proto
}

Agent/REU instproc recv-control { from id } {
    $self instvar node_ proto_

    #puts stderr "Node [$node_ id] testando recv-control - msgid: $id"

    set msg [REU retrieve-msg $id]
    array set message $msg
    #puts stderr "mensagem: $msg"
    set type      $message(0)
    set src       $message(1)
    set grp       $message(2)
    set from_node $message(3)
    set to_node   $message(4)
    #puts stderr "message: [array names message] type: $type src: $src grp: $grp fid: $message(5)"
    if { $type == "tree" } {
	$proto_ recv-$type $src $grp $from_node $to_node $message(5)
    } else {
	$proto_ recv-$type $src $grp $from_node $to_node
    }
}

REU proc get-next-mid {} {
    set ret [REU set mid_]
    REU set mid_ [expr $ret + 1]
    set ret
}

REU proc retrieve-msg id {
    set ret [REU set msg_($id)]
    REU unset msg_($id)
    set ret
}

REU proc read-msg id {
    set ret [REU set msg_($id)]
#    REU unset msg_($id)
    set ret
}

REU instproc send-periodic-join { src grp } {
    global rtglibRNG
    $self instvar node_ joined_
    set ns_ [Simulator instance]

    $self dbg "send-periodic-join entrada, instante: [$ns_ now]"

    if { $joined_($src:$grp) } {
	$self join $src $grp [$node_ id]
	set join_updateTime [expr [$ns_ now] + \
		([$class set advertInterval] * [$rtglibRNG uniform 0.9 1.1])]
#	set join_updateTime [expr [$ns_ now] + [$class set advertInterval]]
	$ns_ at $join_updateTime "$self send-periodic-join $src $grp"
    }
}

REU instproc send-periodic-tree { src grp } {
    global rtglibRNG
    $self instvar proto_

    set ns_ [Simulator instance]

    $self dbg "send-periodic-tree entrada, instante: [$ns_ now]"

    #
    # Send a tree message to all non-stale receivers in MFT
    #
    $self send-all-mft-tree $src $grp

    set tree_updateTime [expr [$ns_ now] + \
	    ( [$class set advertInterval] * [$rtglibRNG uniform 0.9 1.1])]
    $ns_ at $tree_updateTime "$self send-periodic-tree $src $grp"
}

# Timers...

# Timer T01 - MFT group entry
#
Class SkelTimer/MFT -superclass SkelTimer

SkelTimer/MFT instproc init { ns node src grp } {
    $self instvar node_ src_ grp_
    
    $self next $ns $node $src $grp

    $self dbg "MFT Table Timer created"
}

SkelTimer/MFT instproc timeout {} {
    $self instvar ns_ node_ src_ grp_

    $self dbg "MFT Timeout! $self"

    $node_ set mft_fg_($src_:$grp_) 0                   ;# MFT is marked stale

    array set mft [$node_ reu-read-mft $src_ $grp_]
    $self dbg "MFT Timeout mft entries: [array names mft]"
    foreach recv [array names mft] {
	$node_ set mft_entry_fg_($src_:$grp_:$recv) 0   ;# All recvs are stale
	###[$node_ set mft_entry_timer_($src_:$grp_:$recv)] expire
    }

    $node_ reu-create-mft-t02 $src_ $grp_
}

###
# Timer T02 - MFT group entry
#
Class SkelTimer/MFT/Destroy -superclass SkelTimer/MFT

SkelTimer/MFT/Destroy instproc init { ns node src grp } {
    $self instvar node_ src_ grp_
    
    $self next $ns $node $src $grp

    $self dbg "MFT Destroy Timer created"
}

SkelTimer/MFT/Destroy instproc timeout {} {
    $self instvar node_ src_ grp_

    $self dbg "MFT Destroy! $src_:$grp_"

    $node_ reu-destroy-mft $src_ $grp_
}

###
# Timer T01 - MFT recv entry
#
Class SkelTimer/MFTentry -superclass SkelTimer

SkelTimer/MFTentry instproc init { ns node src grp recv } {
    $self instvar node_ src_ grp_ recv_
    $self next $ns $node $src $grp

    $self dbg "MFT Entry Timer created"
    set recv_ $recv
}

SkelTimer/MFTentry instproc timeout {} {
    $self instvar node_ src_ grp_ recv_

    $self dbg "MFT Entry Timeout! $src_:$grp_:$recv_"

    $node_ set mft_entry_fg_($src_:$grp_:$recv_) 0   ;# recv entry is not alive
    $node_ reu-create-mft-entry-t02 $src_ $grp_ $recv_
#    if { [$node_ reu-mft-size? $src_ $grp_] == 1 } {
#	$node_ set mft_fg_($src_:$grp_) 0
#    }
}

# Timer T02 - MFT recv entry
#
Class SkelTimer/MFTentry/Destroy -superclass SkelTimer/MFTentry

SkelTimer/MFTentry/Destroy instproc timeout {} {
    $self instvar node_ src_ grp_ recv_

    $self dbg "MFT Entry Destroy Timeout! $src_:$grp_:$recv_"

    $node_ reu-destroy-mft-entry $src_ $grp_ $recv_
    if { [$node_ reu-mft-size? $src_ $grp_] == 0 } {
	set dst_node [$node_ set mft_dst_($src_:$grp_)]
	$node_ reu-insert-mct-entry $src_ $grp_ $dst_node
	$node_ reu-destroy-mft $src_ $grp_
    }
}

Node instproc join-group { agent group { src "" } } {
    $self instvar replicator_ Agents_ mrtObject_
    set group [expr $group] ;# use expr to convert to decimal

    $self dbg "I used my join group"
    $mrtObject_ join-group $group $src
}

Node instproc leave-group { agent group { src "" } } {
    $self instvar replicator_ Agents_ mrtObject_

    $self dbg "I used my leave group"    
    $mrtObject_ leave-group $group $src
}

# Timer T01 - MCT
#
Class SkelTimer/MCT -superclass SkelTimer

SkelTimer/MCT instproc init { ns node src grp } {
    $self instvar node_ src_ grp_
    
    $self next $ns $node $src $grp
    $self dbg "MCT Timer created"
}

SkelTimer/MCT instproc timeout {} {
    $self instvar ns_ node_ src_ grp_

    $self dbg "MCT Timeout! $self"
    $node_ reu-destroy-mct $src_ $grp_
}

# Timer T01 - MCT entry
#
Class SkelTimer/MCTentry -superclass SkelTimer

SkelTimer/MCTentry instproc init { ns node src grp recv } {
    $self instvar node_ src_ grp_ recv_
    
    $self next $ns $node $src $grp
    $self dbg "MCT Entry Timer created"
    set recv_ $recv
}

SkelTimer/MCTentry instproc timeout {} {
    $self instvar ns_ node_ src_ grp_ recv_

    $self dbg "MCT Entry Timeout! $self"
    $node_ reu-destroy-mct-entry $src_ $grp_ $recv_
}
