#
# Maintainer: <Luis.Costa@lip6.fr>.
# This file only contains the methods for dynamic SMM_LS routing
#

###############################################################################
# Segunda versao do Roteamento com QoS - SPF algorithm                        #
# Metrica Unica Combinada                                                     #
###############################################################################

#Esse trecho estava em tcl/lib/ns-default.tcl

Agent/rtProto/SMM_LS set preference_     90 ;# pref maior que direct(num menor)
Agent/rtProto/SMM_LS set INFINITY        [Agent set ttl_]
Agent/rtProto/SMM_LS set advertInterval  2

# unreach_ com valor -1 como original nao fica muito legal...
rtObject set unreach_ 32

# Rotina que escolhe e instala as melhores rotas entre todos os protocolos,
# de acordo com sua preferencia (preference_).

rtObject instproc SMM_LS-compute-routes {} {
    $self instvar ns_ node_ rtProtos_ nullAgent_
    $self instvar nextHop_ rtpref_ metric_ rtVia_
    
    set UNREACH_      [$class set unreach_]

    set protos ""
    set changes 0

    foreach p [array names rtProtos_] {
	if [$rtProtos_($p) set rtsChanged_] {
	    incr changes
	    $rtProtos_($p) set rtsChanged_ 0
	}
	lappend protos $rtProtos_($p)
    }
    if !$changes return

    set changes 0
    
    foreach dst [$ns_ all-nodes-list] {
	if {$dst == $node_} continue   ;# no'=destino, passe para o proximo
	set nh ""
	set pf [$class set maxpref_]
	set mt  $UNREACH_
	set rv ""
	# para cada protocolo
	foreach p $protos {                    
	    set pnh [$p set nextHop_($dst)]   
	    # nao havia rota, prox. prot.
	    if { $pnh == "" } continue
	    #se ha' rota, pegamos os valores de prot, pref e metricas
	    set ppf  [$p set rtpref_($dst)]  ;# preferencia do prot
	    set pmt  [$p set metric_($dst)]  ;# hops
	    
	    # se rota aceitavel...
	    if {$ppf < $pf || ($ppf == $pf && $pmt < $mt) \
		    || $mt >= $UNREACH_ } {
		set nh  $pnh                   ;# proximo hop
		set pf  $ppf                   ;# pref da rota
		set mt   $pmt
		set rv  $p                     ;# protocolo
	    }
	} 
	# foreach protos
	
	# se concluimos rota inexistente... apague as rotas existentes
	if { $nh == "" } {
	    if { $nextHop_($dst) != "" } {        
		$node_ delete-routes [$dst id] $nextHop_($dst) $nullAgent_
		set nextHop_($dst)  $nh
		set rtpref_($dst)   $pf
		set metric_($dst)   $mt
		set rtVia_($dst)    $rv
		incr changes
	    }
	    #se nao, existe uma rota. Vamos analisa-la
	} else {
	    #primeiro caso, o prot e' o mesmo que ja' tinha a melhor rota
	    if { $rv == $rtVia_($dst) } {                         
		# Current protocol still has best route.
		# See if changed
		# caso em que a rota mudou (prox salto)
		if { $nh != $nextHop_($dst) } {                      
		    $node_ delete-routes [$dst id] $nextHop_($dst) $nullAgent_
		    set nextHop_($dst) $nh
		    $node_ add-routes [$dst id]  $nextHop_($dst)
		    incr changes
		}
		
		# se houve alguma mudanca nas metricas da rota...	    
		if { ($metric_($dst) != $mt) } {
		    set metric_($dst) $mt
		    incr changes
		}
		if { $pf != $rtpref_($dst) } {   
		    set rtpref_($dst) $pf                             
		}
		#segundo caso, o prot nao e' o mesmo que tinha melhor rota
	    } else {
		if { $rtVia_($dst) != "" } {
		    set nextHop_($dst) [$rtVia_($dst) set nextHop_($dst)]
		    set rtpref_($dst) [$rtVia_($dst) set rtpref_($dst)]
		    set metric_($dst) [$rtVia_($dst) set metric_($dst)]
		}
		
		if {$rtpref_($dst) != $pf || $metric_($dst) != $mt } {
		    # Then new prefs must be better, or
		    # new prefs are equal, and new metrics are lower
		    $node_ delete-routes [$dst id] $nextHop_($dst) $nullAgent_
		    set nextHop_($dst) $nh
		    set rtpref_($dst)  $pf
		    set metric_($dst)   $mt
		    set rtVia_($dst)   $rv
		    $node_ add-routes [$dst id] $nextHop_($dst)
		    incr changes
		}
	    }
	}
    }
    # foreach dest
    
    foreach proto [array names rtProtos_] {
	$rtProtos_($proto) send-updates $changes
    }
    #
    # XXX
    # detailed multicast routing hooks must come here.
    # My idea for the hook will be something like:
    # set mrtObject [$node_ mrtObject?]
    # if {$mrtObject != ""} { ;#
    #    $mrtObject recompute-mroutes $changes
    # }
    # $changes == 0	if only interfaces changed state.  Look at how
    #			Agent/rtProto/DV handles ifsUp_
    # $changes > 0	if new unicast routes were installed.
    #
    
    $self flag-multicast $changes
}

###############################################################################
#                                                                             #
# Shortest Path First Route Computation - SMM_LS approach                     #
#                                                                             #
###############################################################################

Agent/rtProto/SMM_LS set mid_		  0

Agent/rtProto/SMM_LS set UNREACH_	[rtObject set unreach_]
Agent/rtProto/SMM_LS set WST_DELAY	[rtObject set wst_delay]
Agent/rtProto/SMM_LS set WST_BW	        [rtObject set wst_bw]
Agent/rtProto/SMM_LS set WST_LP  	[rtObject set wst_lp]
Agent/rtProto/SMM_LS set WST_UNIC	[rtObject set wst_unic]

# Initialization of all nodes
Agent/rtProto/SMM_LS proc init-all args {

    if { [llength $args] == 0 } {
	set nodeslist [[Simulator instance] all-nodes-list]
    } else {
	#eval "set nodeslist $args"                          ;# luish - patched
	set nodeslist $args
    }
    Agent set-maxttl Agent/rtProto/SMM_LS INFINITY
    eval rtObject init-all $nodeslist
    foreach node $nodeslist {
	set proto($node) [[$node rtObject?] add-proto SMM_LS $node]
    }
    foreach node $nodeslist {
	foreach nbr [$node neighbors] {
	    set rtobj [$nbr rtObject?]
	    if { $rtobj != "" } {
		set rtproto [$rtobj rtProto? SMM_LS]
		if { $rtproto != "" } {
		    #$proto($node) add-peer $nbr [$rtproto set agent_addr_]  
		    $proto($node) add-peer $nbr [$rtproto set agent_addr_] [$rtproto set agent_port_]
		}
	    }
	}
    }
}

# Initialization of each individual node
Agent/rtProto/SMM_LS instproc init node {
    global rtglibRNG

    $self next $node
    $self instvar ns_ rtObject_ ifsUp_ ifs_
    $self instvar preference_ rtpref_ nextHop_ nextHopPeer_ multiPath_
    $self instvar metric_

    # topology representation
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_ 

    #set UNREACH_      [$class set UNREACH_]
    #set WST_DELAY [$class set WST_DELAY]
    #set WST_BW    [$class set WST_BW]
    #set WST_LP    [$class set WST_LP]
    #set WST_UNIC  [$class set WST_UNIC]

    foreach dest [$ns_ all-nodes-list] {
	set rtpref_($dest) $preference_
	set nextHop_($dest) ""
	set nextHopPeer_($dest) ""
	set metric_($dest) [$class set UNREACH_]
    }
    
    # topology initialization - we know nothing by now...
    foreach node_i [$ns_ all-nodes-list] {
	foreach node_j [$ns_ all-nodes-list] {
	    set link_mH_($node_i$node_j)  [$class set UNREACH_]        ;# luish
	    set link_m0_($node_i$node_j)  [$class set WST_DELAY]       ;# luish
	    set link_m1_($node_i$node_j)  [$class set WST_BW]          ;# luish
	    set link_m2_($node_i$node_j)  [$class set WST_LP]          ;# luish
	    set link_mU_($node_i$node_j)  [$class set WST_UNIC]        ;# luish
	}
    }
    # but we do know about our links... (from HERE to our neighbors)
    foreach nbr [$node neighbors] {
	#puts stderr "vizinho: $nbr $node"
	set link_mH_($node$nbr) [$ifs_($nbr) cost?]
	set link_m0_($node$nbr) [$ifs_($nbr) cost0?]
	set link_m1_($node$nbr) [$ifs_($nbr) cost1?]
	set link_m2_($node$nbr) [$ifs_($nbr) cost2?]
	set link_mU_($node$nbr) [unica $link_m0_($node$nbr) $link_m2_($node$nbr)]
    }
    
    set ifsUp_ ""
    set multiPath_ [[$rtObject_ set node_] set multiPath_]
    set updateTime [$rtglibRNG uniform 0.0 0.5]
    #mudei o tempo inicial de 0.5 para 0.05 para aumentar a capacidade 
    #de nos com poucos vizinhos calcularem as rotas
    $ns_ at $updateTime "$self send-periodic-update"
}

Agent/rtProto/SMM_LS instproc add-peer {nbr agentAddr agentPort} {
    $self instvar peers_
    $self set peers_($nbr) [new rtPeer $agentAddr $agentPort $class]
}

Agent/rtProto/SMM_LS instproc send-periodic-update {} {
    global rtglibRNG

    $self instvar ns_
    #puts stderr "send-periodic-update entrada, instante: [$ns_ now]"  ;# luish
    $self send-updates 1	;# Anything but 0
    set updateTime [expr [$ns_ now] + \
	    ([$class set advertInterval] * [$rtglibRNG uniform 0.9 1.1])]
    $ns_ at $updateTime "$self send-periodic-update"
}

# this procedure stays as a "best-effort" route computation
Agent/rtProto/SMM_LS instproc compute-routes {} {
    $self instvar ns_ ifs_ rtpref_ nextHop_ nextHopPeer_
    $self instvar peers_ rtsChanged_ multiPath_
    $self instvar node_ metric_ 
    $self instvar rtObject_					       ;# luish

    #puts stderr "entrada de SMM_LS compute-routes, no: [$node_ id]"  ;# luish

    set MAXPREF          [rtObject set maxpref_]
    set UNREACH_	 [rtObject set unreach_]
    set rtsChanged_ 0
    
    foreach dst [$ns_ all-nodes-list] {
	
	catch "unset nh"

	set p [lindex $nextHopPeer_($dst) 0]
	if {$p != ""} {
	    set metric_($dst)  [$p metric? $dst]
	    set rtpref_($dst)  [$p preference? $dst]
	}
	
	set pf  $MAXPREF
	set mt  $UNREACH_
	set nh(0) 0
	
	foreach nbr [array names peers_] {
	    set pmt  [$peers_($nbr) metric? $dst]
	    set ppf  [$peers_($nbr) preference? $dst]
	    
	    # if peer metric not valid	continue
	    # if peer pref higher		continue
	    # if peer pref lower		set to latest values
	    # else peer pref equal
	    #	if peer metric higher	continue
	    #	if peer metric lower	set to latest values
	    # 	else peer metrics equal	append latest values
	    
	    if { $pmt < 0 || $pmt >= $UNREACH_ || $ppf > $pf || $pmt > $mt} {
		continue
	    }
	    if { $ppf < $pf || $pmt < $mt } { 
		# metrica do peer melhor
		set pf $ppf
		set mt $pmt
		unset nh	   ;# because we must compute *new* next hops
	    }                   
	    # senao, metrica igual
	    set nh($ifs_($nbr)) $peers_($nbr)
	}    
	#foreach nbr
	
	catch "unset nh(0)"               ;# adiantamos este unset - luish

 	if { $pf == $MAXPREF && $mt == $UNREACH_ } continue
 	if { $pf > $rtpref_($dst) ||		\
 		($metric_($dst) >= 0 && $mt > $metric_($dst) ) } \
 		continue
	if {$mt >= $UNREACH_ } {
	    set mt  $UNREACH_
	}
	
	incr rtsChanged_
	if {$pf < $rtpref_($dst) || $mt < $metric_($dst) } {
	    set rtpref_($dst) $pf
	    set metric_($dst)  $mt
	    set nextHop_($dst) ""
	    set nextHopPeer_($dst) ""
	    foreach n [array names nh] {
		lappend nextHop_($dst) $n
		lappend nextHopPeer_($dst) $nh($n)
		if !$multiPath_ break;
	    }
	    continue
	}
	
	set rtpref_($dst) $pf
	set metric_($dst) $mt
	set newNextHop ""
	set newNextHopPeer ""
	foreach rt $nextHop_($dst) {
	    if [info exists nh($rt)] {
		lappend newNextHop $rt
		lappend newNextHopPeer $nh($rt)
		unset nh($rt)
	    }
	}
	
	set nextHop_($dst) $newNextHop
	set nextHopPeer_($dst) $newNextHopPeer
	if { $multiPath_ || $nextHop_($dst) == "" } {
	    foreach rt [array names nh] {
		lappend nextHop_($dst) $rt
		lappend nextHopPeer_($dst) $nh($rt)
		if !$multiPath_ break
	    }
	}
    }
    # foreach dst    
    set rtsChanged_
}

###############################################################################
# get-QoS-route:                                                              #
#     interface to be used by flows requesting an explicit QoS route          #
# Parameters: Required Delay, Bandwidth, and Loss Probability + Destination   #
###############################################################################

rtObject instproc get-QoS-route { algorithm r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     

    set no_route 1

    #switch $algorithm {
	#0 { $self get-QoS-route-0 $r_dl $r_bw $r_lp $dest }
	#1 { $self get-QoS-route-1 $r_dl $r_bw $r_lp $dest }
	#2 { $self get-QoS-route-2 $r_dl $r_bw $r_lp $dest }
	#3 { $self get-QoS-route-3 $r_dl $r_bw $r_lp $dest }
	#4 { $self get-QoS-route-4 $r_dl $r_bw $r_lp $dest }
	#5 { $self get-QoS-route-5 $r_dl $r_bw $r_lp $dest }
    #}

	
    $self get-QoS-route-$algorithm $r_dl $r_bw $r_lp $dest

    # We found a route, let's prepare the result
    if { $no_route == 0 } {
	set j 0
	set atraso 0
	set banda 1000
	set probab 0.0
	set searched $dest
	while { $searched != [$node_ id] } {
	    foreach i [array names L] {
		if { [lindex $L($i) 0] == $searched } {
		    set rota($j) $searched
		    set j [expr $j + 1]
		    set newsearched [lindex $L($i) 1]
		    set no1 [$ns_ get-node-by-id $searched]
		    set no2 [$ns_ get-node-by-id $newsearched]
		    set atraso [expr $atraso + $link_m0_($no2$no1)]
		    set banda  [minimo $banda $link_m1_($no2$no1)]
		    set probab [prob $probab $link_m2_($no2$no1)]
		    set searched $newsearched
		}
	    }
	}
    } else {
	set route [$node_ id]
	#puts stderr "No route found by the algorithm"
	return $route
    }
    # Are the QoS requirements respected?
    if { $atraso > $r_dl || $banda < $r_bw || $probab > $r_lp } {
	set route [$node_ id]
	#if { $atraso > $r_dl } { puts stderr "Not enough delay" }
	#if { $banda < $r_bw }  { puts stderr "Not enough bandwidth" }
	#if { $probab > $r_lp } { puts stderr "Not enough loss_prob" }
	#puts stderr "The route found does not have enough QoS"
	return $route
    }
    
    # So let's return the route...
    set route [$node_ id]    
    set size [array size rota]    
    while { $size != 0 } {
	set size [expr $size - 1]
	set route [concat $route $rota($size)]
    }
    return $route
}

###############################################################################
# Version "0": Crowcroft version (bandwidth pruning + delay dijkstra search)  #
###############################################################################

rtObject instproc get-QoS-route-0 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     

    set WST_DELAY        [rtObject set wst_delay]

    catch "unset L"
    set no_route 1

    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    set d([$i id][$j id]) $link_m0_($i$j)
	}
    }

    #Step 1: set dij = inf if bij < B   - bandwidth pruning
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    if { $link_m1_($i$j) < $r_bw } {
		set d([$i id][$j id]) $WST_DELAY
	    }
	}
    }
    
    #Step 2: set L = {(1,1)}
    set L(0) [concat [$node_ id] [$node_ id]]
    set n 1
    
    # Di = d1i for all i != 1
    foreach i [$ns_ all-nodes-list] {       
	if { $i != $node_ } {
	    set D([$i id]) $d([$node_ id][$i id])
	    set ph([$i id]) "[$node_ id]"
	}
    }

    #Step 3:     find k not in L() | Dk = min Di
    while { 1 } {
	set min $WST_DELAY                
	foreach i [array names D] {
	    if { $D($i) < $min } {
		set min $D($i)
		set k $i
	    }
	}
	#if Dk > r_dl no possible path, finish
	if { $min > $r_dl } {
	    break
	}
	
	# L = L U (k,ph)
	set L($n) [concat $k $ph($k)]
	set n [expr $n + 1]

	#if L(,) contains dest, we got a path, finish
	if { $k == $dest } {
	    set no_route 0
	    break
	}
	
	unset D($k)
	#Step 4: forall i not in L, set Di = min(Di,Dk+dki)
	foreach i [array names D] {
	    if { $D($i) > [expr $min + $d($k$i)] } {
		set D($i) [expr $min + $d($k$i)]
		    set ph($i) $k
	    }
	}
    }
}

###############################################################################
# Version 1: Bw pruning + mixed metric dijkstra search(without remainder appr)#
###############################################################################

rtObject instproc get-QoS-route-1 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     
    
    #set WST_BW           [rtObject set wst_bw]   
    set WST_UNIC         [rtObject set wst_unic] 

    catch "unset L"
    set no_route 1
    set r_unic [unica $r_dl $r_lp]
    
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    set u([$i id][$j id]) $link_mU_($i$j)
	}
    }

    #Step 1: set uij = inf if bij < B   - bandwidth pruning
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    if { $link_m1_($i$j) < $r_bw } {
		set u([$i id][$j id]) $WST_UNIC
	    }
	}
    }

    #Step 2: set L = {(1,1)}
    set L(0) [concat [$node_ id] [$node_ id]]
    set n 1

    # Ui = u1i for all i != 1
    foreach i [$ns_ all-nodes-list] {       
	if { $i != $node_ } {
	    set U([$i id]) $u([$node_ id][$i id])
	    set ph([$i id]) "[$node_ id]"
	}
    }

    #Step 3:     find k not in L() | Uk = min Ui
    while { 1 } {
	set min $WST_UNIC
	foreach i [array names U] {
	    if { $U($i) < $min } {
		set min $U($i)
		set k $i
	    }
	}
	#if Uk > r_unic no possible path, finish  
	if { $min > $r_unic } {
	    break
	}

	# L = L U (k,ph)
	set L($n) [concat $k $ph($k)]
	set n [expr $n + 1]

	#if L(,) contains dest, we got a path, finish
	if { $k == $dest } {
	    set no_route 0
	    break
	}

	unset U($k)
	#Step 4: forall i not in L, set Ui = min(Ui,Uk+uki)
	foreach i [array names U] {
	    if { $U($i) > [expr $min + $u($k$i)] } {
		set U($i) [expr $min + $u($k$i)]
		set ph($i) $k
	    }
	}
    }
}

#
# OLD OLD OLD OLD OLD
#
###############################################################################
# Version 2: SMM-LS approach: bw pruning, mixed metric + remainder heuristic  #
###############################################################################

rtObject instproc get-QoS-route-7 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     
    
    set WST_UNIC         [rtObject set wst_unic] 

    catch "unset L"
    set no_route 1
    # r_unic = r_delay + r_slog
    set r_unic [unica $r_dl $r_lp]

    set SlogM [expr -log10(1-$r_lp)]
    
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    set u([$i id][$j id]) $link_mU_($i$j)
	}
    }

    #Step 1: set uij = inf if bij < B   - bandwidth pruning
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    if { $link_m1_($i$j) < $r_bw } {
		set u([$i id][$j id]) $WST_UNIC
	    }
	}
    }

    foreach C { 2 3 4 5 6 7 8 9 10 11 12 } {
	catch "unset L"
	catch "unset U"
	catch "unset ph"
	catch "unset SS"
	catch "unset SP"

	# Step 2: set L = {(1,1)}
	#         SS1 = C ; SSi = C-1 ; SPi = SlogM
	#         Ui = u1i for all i != 1

	set L(0) [concat [$node_ id] [$node_ id]]
	set n 1
	set SS([$node_ id]) $C
	set SP([$node_ id]) $SlogM

	foreach i [$ns_ all-nodes-list] {       
	    if { $i != $node_ } {
		set U([$i id]) $u([$node_ id][$i id])
		set ph([$i id]) "[$node_ id]"
		if { $U([$i id]) == $WST_UNIC } {
		    set SS([$i id]) $C
		    set SP([$i id]) $SlogM
		} else {
		    set SS([$i id]) [expr $C - 1]
		    set SP([$i id]) [expr $SlogM - \
		      ($u([$node_ id][$i id]) - floor($u([$node_ id][$i id])))]
		}
	    }
	}
	

	# Step 3: 
	# find k not in L() | Uk = min Ui AND (u_ki - floor(u_ki)) < SPi/SSi
	set k [$node_ id]
	while { 1 } {
	    set min $WST_UNIC                 
	    foreach i [array names U] {
		if { ($U($i) < $min)  } {
		 #  if { ($u($k$i) - floor($u($k$i))) < ($SP($i)/$SS($i)) } 
			set min $U($i)
			set k $i
		   }
		
	    }
	    #if Uk > r_unic no possible path, finish
	    if { $min > $r_unic } break

	    #set k $k_l
	    # L = L U (k,ph)
	    set L($n) [concat $k $ph($k)]
	    set n [expr $n + 1]

	    #if L(,) contains dest, we got a path, finish
	    if { $k == $dest } {
		set no_route 0
		break
	    }
	   
	    # SS = SS - 1; SP = SP - (Uk - floor(Uk))
	    unset U($k)
	    #Step 4: forall i not in L, set Ui = min(Ui,Uk+uki)
	    foreach i [array names U] {
		set newSS [expr $SS($k) - 1]
		set newSP [expr $SP($k) - ( $u($k$i) - floor ($u($k$i)))]
		set sim 0
		if { ( [expr $min + $u($k$i)] < $U($i) ) } {
		    if { $SS($i) > 0 } {
		    if { ($u($k$i) - floor($u($k$i))) < ($SP($i)/$SS($i)) } {
		    set sim 1
		    }
		}
		}
		#if { ( $SS($i) > 0 ) && ( $newSS > 0 ) } {
		#    if { ( ( [expr $min + $u($k$i)] < $r_unic ) && 
		#    ( [expr $newSP/$newSS] < [expr $SP($i)/$SS($i)] ) )} {
		#	set sim 1
		#    }
		#}
		if { $sim } {
		    set U($i) [expr $min + $u($k$i)]
		    set ph($i) $k
		    set SS($i) [expr $SS($k) - 1]
		    set SP($i) [expr $SP($k) - ( $u($k$i)-floor($u($k$i)) )]
		}
	    }
	}
	if { $no_route == 0 } {
	    puts stderr "Ce: $C"
	    break
	}
    }
}

###############################################################################
# Version "3": Overall Cost version (f(c) = d*lp / b) (dijkstra search)       #
###############################################################################

rtObject instproc get-QoS-route-3 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     
    
    set WST_DELAY        [rtObject set wst_delay]
    #set WST_BW           [rtObject set wst_bw]   
    #set WST_LP           [rtObject set wst_lp]

    catch "unset L"
    set no_route 1
    set r_fc [expr ($r_dl*$r_lp) / $r_bw]
    
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    if { $link_m1_($i$j) != 0 } {
		set f([$i id][$j id]) [expr ($link_m0_($i$j)*$link_m2_($i$j)) / $link_m1_($i$j)]
	    } else {
		set f([$i id][$j id]) $WST_DELAY
	    }
	}
    }

    #Step 2: set L = {(1,1)}
    set L(0) [concat [$node_ id] [$node_ id]]
    set n 1

    # Fi = f1i for all i != 1
    foreach i [$ns_ all-nodes-list] {       
	if { $i != $node_ } {
	    set F([$i id]) $f([$node_ id][$i id])
	    set ph([$i id]) "[$node_ id]"
	}
    }

    #Step 3:     find k not in L() | Fk = min Fi
    while { 1 } {
	set min $WST_DELAY                
	foreach i [array names F] {
	    if { $F($i) < $min } {
		set min $F($i)
		set k $i
	    }
	}
	#if Fk > r_fc no possible path, finish
	if { $min > $r_fc } {
	    break
	}

	# L = L U (k,ph)
	set L($n) [concat $k $ph($k)]
	set n [expr $n + 1]

	#if L(,) contains dest, we got a path, finish
	if { $k == $dest } {
	    set no_route 0
	    break
	}

	unset F($k)
	#Step 4: forall i not in L, set Fi = min(Fi,Fk+fki)
	foreach i [array names F] {
	    if { $F($i) > [expr $min + $f($k$i)] } {
		set F($i) [expr $min + $f($k$i)]
		set ph($i) $k
	    }
	}
    }
}

###############################################################################
# Version "4": Overall Cost version (f(c) = d*l / b) (dijkstra search)        #
#              with scaled metrics                                            #
###############################################################################

rtObject instproc get-QoS-route-4 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     
    
    set WST_DELAY        [rtObject set wst_delay]
    #set WST_BW           [rtObject set wst_bw]   
    set WST_LP           [rtObject set wst_lp]
    set WST_F            [expr ($WST_DELAY * ($WST_LP*1e3)) * 1e4]

    catch "unset L"
    set no_route 1
    set r_fc [expr ( $r_dl * ($r_lp*1e3) ) / ($r_bw/1e3)]
    
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    if { $link_m1_($i$j) != 0 } {
		set f([$i id][$j id]) [expr ( $link_m0_($i$j) * \
			($link_m2_($i$j)*1e3) ) / ($link_m1_($i$j)/1e3) ]
	    } else {
		set f([$i id][$j id]) $WST_F
	    }
	}
    }

    #Step 2: set L = {(1,1)}
    set L(0) [concat [$node_ id] [$node_ id]]
    set n 1

    # Fi = f1i for all i != 1
    foreach i [$ns_ all-nodes-list] {
	if { $i != $node_ } {
	    set F([$i id]) $f([$node_ id][$i id])
	    set ph([$i id]) "[$node_ id]"
	}
    }

    #Step 3:     find k not in L() | Fk = min Fi
    while { 1 } {
	set min $WST_F
	foreach i [array names F] {
	    if { $F($i) < $min } {
		set min $F($i)
		set k $i
	    }
	}
	#if Fk > r_fc no possible path, finish
	if { $min > $r_fc } {
	    break
	}

	# L = L U (k,ph)
	set L($n) [concat $k $ph($k)]
	set n [expr $n + 1]

	#if L(,) contains dest, we got a path, finish
	if { $k == $dest } {
	    set no_route 0
	    break
	}

	unset F($k)
	#Step 4: forall i not in L, set Fi = min(Fi,Fk+fki)
	foreach i [array names F] {
	    if { $F($i) > [expr $min + $f($k$i)] } {
		set F($i) [expr $min + $f($k$i)]
		set ph($i) $k
	    }
	}
    }
}

###############################################################################
# Version "5": Simple minimum hop search based on the Dijkstra algorithm      #
###############################################################################

rtObject instproc get-QoS-route-5 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     

    set UNREACH_      [rtObject set unreach_]
   
    catch "unset L"
    set no_route 1
    
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    set d([$i id][$j id]) $link_mH_($i$j)
	}
    }

    #Step 2: set L = {(1,1)}
    set L(0) [concat [$node_ id] [$node_ id]]
    set n 1

    # Di = d1i for all i != 1
    foreach i [$ns_ all-nodes-list] {       
	if { $i != $node_ } {
	    set D([$i id]) $d([$node_ id][$i id])
	    set ph([$i id]) "[$node_ id]"
	}
    }

    #Step 3:     find k not in L() | Dk = min Di
    while { 1 } {
	set min $UNREACH_                
	foreach i [array names D] {
	    if { $D($i) < $min } {
		set min $D($i)
		set k $i
	    }
	}
	#if Dk > r_dl no possible path, finish
	if { $min >= $UNREACH_ } {
	    break
	}

	# L = L U (k,ph)
	set L($n) [concat $k $ph($k)]
	set n [expr $n + 1]

	#if L(,) contains dest, we got a path, finish
	if { $k == $dest } {
	    set no_route 0
	    break
	}

	unset D($k)
	#Step 4: forall i not in L, set Di = min(Di,Dk+dki)
	foreach i [array names D] {
	    if { $D($i) > [expr $min + $d($k$i)] } {
		set D($i) [expr $min + $d($k$i)]
		set ph($i) $k
	    }
	}
    }
}

###############################################################################
# Version "6": Rajagopalan's Heuristic + bandwith pruning                     #
###############################################################################

rtObject instproc get-QoS-route-6 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     

    set WST_DELAY        [rtObject set wst_delay]
    set WST_SLOG         [rtObject set wst_unic]  ;# slog has to be minimized

    set r_slog  [expr -log10(1 - $r_lp)]

    catch "unset L"
    set no_route 1
    
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    set d([$i id][$j id]) $link_m0_($i$j)
	    set slog([$i id][$j id]) [expr -log10(1 - $link_m2_($i$j)) ]
	}
    }

    #Step 1: set dij = inf if bij < B   - bandwidth pruning
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    if { $link_m1_($i$j) < $r_bw } {
		set d([$i id][$j id]) $WST_DELAY
		set slog([$i id][$j id]) $WST_SLOG
	    }
	}
    }

    # Dijkstra 1, delay-based
    
    #Step 2: set L = {(1,1)}
    set L(0) [concat [$node_ id] [$node_ id]]
    set n 1
    
    # Di = d1i for all i != 1
    foreach i [$ns_ all-nodes-list] {       
	if { $i != $node_ } {
	    set D([$i id])  $d([$node_ id][$i id])
	    set SL([$i id]) $slog([$node_ id][$i id])
	    set ph([$i id]) "[$node_ id]"
	}
    }
    
    #Step 3:     find k not in L() | Dk = min Di
    while { 1 } {
	set min $WST_DELAY                
	foreach i [array names D] {
	    if { $D($i) < $min } {
		set min  $D($i)
		set sl_d $SL($i)
		set k $i
	    }
	}
	#if Dk > r_dl no possible path, finish
	if { $min > $r_dl } {
	    break
	}
	
	# L = L U (k,ph)
	set L($n) [concat $k $ph($k)]
	set n [expr $n + 1]

	#if L(,) contains dest, we got a path, finish
	if { $k == $dest } {
	    set no_route 0
	    break
	}
	
	unset D($k)
	unset SL($k)
	#Step 4: forall i not in L, set Di = min(Di,Dk+dki)
	foreach i [array names D] {
	    if { $D($i) > [expr $min + $d($k$i)] } {
		set D($i) [expr $min + $d($k$i)]
		set SL($i) [expr $sl_d + $slog($k$i)]
		set ph($i) $k
	    }
	}
    }
    ;# End Dijkstra 1

    # if there is no delay-feasible route, no need to do slog search...
    if { $no_route == 1 } return

    # if we found a delay-feasible route...
    if { $no_route == 0 } {
	# if slog is good, ok, we have a feasible route
	if { $sl_d <= $r_slog } return
    }

    set no_route 1
    # if slog is not good, we try a slog-based search
    unset L
    unset min
    # Dijkstra 2, slog-based
    
    #Step 2: set L = {(1,1)}
    set L(0) [concat [$node_ id] [$node_ id]]
    set n 1
    
    # Di = d1i for all i != 1
    foreach i [$ns_ all-nodes-list] {       
	if { $i != $node_ } {
	    set D([$i id])  $d([$node_ id][$i id])
	    set SL([$i id]) $slog([$node_ id][$i id])
	    set ph([$i id]) "[$node_ id]"
	}
    }
    
    #Step 3:     find k not in L() | Dk = min Di
    while { 1 } {
	set min $WST_SLOG                
	foreach i [array names SL] {
	    if { $SL($i) < $min } {
		set min  $SL($i)
		set d_sl $D($i)
		set k $i
	    }
	}
	#if SLk > r_slog no possible path, finish
	if { $min > $r_slog } {
	    break
	}
	
	# L = L U (k,ph)
	set L($n) [concat $k $ph($k)]
	set n [expr $n + 1]

	#if L(,) contains dest, we got a path, finish
	if { $k == $dest } {
	    set no_route 0
	    break
	}
	
	unset SL($k)
	unset D($k)
	#Step 4: forall i not in L, set SLi = min(SLi,SLk+slog_ki)
	foreach i [array names SL] {
	    if { $SL($i) > [expr $min + $slog($k$i)] } {
		set SL($i) [expr $min + $slog($k$i)]
		set D($i) [expr $d_sl + $d($k$i)]
		set ph($i) $k
	    }
	}
    }
    ;# End Dijkstra 2
    
    # if I found a slog-feasible route...
    if { $no_route == 0 } {
	# is it delay-feasible?
	if { $d_sl <= $r_dl } return
    }
    # if delay is not good, no route...
    set no_route 1
}


#
# NEW NEW NEW NEW NEW NEW
#
###############################################################################
# Version 2: SMM-LS approach: bw pruning, mixed metric + remainder heuristic  #
###############################################################################

rtObject instproc get-QoS-route-2 { r_dl r_bw r_lp dest } {
    $self instvar ns_ node_ no_route L
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_     
    
    set WST_UNIC         [rtObject set wst_unic] 

    catch "unset L"
    set no_route 1
    # r_unic = r_delay + r_slog
    set r_unic [unica $r_dl $r_lp]

    set SlogM [expr -log10(1-$r_lp)]
    
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    set u([$i id][$j id]) $link_mU_($i$j)
	}
    }

    #Step 1: set uij = inf if bij < B   - bandwidth pruning
    foreach i [$ns_ all-nodes-list] {
	foreach j [$ns_ all-nodes-list] {
	    if { $link_m1_($i$j) < $r_bw } {
		set u([$i id][$j id]) $WST_UNIC
	    }
	}
    }

    foreach C { 2 3 4 5 6 7 8 9 10 11 12 } {
	catch "unset L"
	catch "unset U"
	catch "unset ph"
	catch "unset SS"
	catch "unset SP"

	# Step 2: set L = {(1,1)}
	#         SS1 = C ; SSi = C-1 ; SPi = SlogM
	#         Ui = u1i for all i != 1

	set L(0) [concat [$node_ id] [$node_ id]]
	set n 1
	set SS([$node_ id]) $C
	set SP([$node_ id]) $SlogM

	foreach i [$ns_ all-nodes-list] {       
	    if { $i != $node_ } {
		set U([$i id]) $u([$node_ id][$i id])
		set ph([$i id]) "[$node_ id]"
		if { $U([$i id]) == $WST_UNIC } {
		    set SS([$i id]) $C
		    set SP([$i id]) $SlogM
		} else {
		    set SS([$i id]) [expr $C - 1]
		    set SP([$i id]) [expr $SlogM - \
		      ($u([$node_ id][$i id]) - floor($u([$node_ id][$i id])))]
		}
	    }
	}
	

	# Step 3: 
	# find k not in L() | Uk = min Ui AND (u_ki - floor(u_ki)) < SPi/SSi
	set k [$node_ id]
	while { 1 } {
	    set min $WST_UNIC                 
	    foreach i [array names U] {
		if { ($U($i) < $min) && (($U($i) - floor($U($i))) < $SlogM) } {
		 #  if { ($u($k$i) - floor($u($k$i))) < ($SP($i)/$SS($i)) } 
			set min $U($i)
			set k $i
		   }
		
	    }
	    #if Uk > r_unic no possible path, finish
	    if { $min > $r_unic } break

	    #set k $k_l
	    # L = L U (k,ph)
	    set L($n) [concat $k $ph($k)]
	    set n [expr $n + 1]

	    #if L(,) contains dest, we got a path, finish
	    if { $k == $dest } {
		set no_route 0
		break
	    }
	   
	    # SS = SS - 1; SP = SP - (Uk - floor(Uk))
	    unset U($k)
	    #Step 4: forall i not in L, set Ui = min(Ui,Uk+uki)
	    foreach i [array names U] {
		set newSS [expr $SS($k) - 1]
		set newSP [expr $SP($k) - ( $u($k$i) - floor ($u($k$i)))]
		set sim 0
		if { ( [expr $min + $u($k$i)] < $U($i) && $SS($i) > 0 ) } {
		    if { ($u($k$i) - floor($u($k$i))) < ($SP($i)/$SS($i)) } {
			set sim 1
		    }
		}
		
		# or if ...

		if {( [expr $min + $u($k$i)] < $r_unic ) && $SS($i) > 0 \
			&& $newSS > 0 } {
		    if { ($newSP/$newSS) < ($SP($i)/$SS($i)) } {
			set sim 1
		    }
		}
	 
		#if { ( $SS($i) > 0 ) && ( $newSS > 0 ) } {
		#    if { ( ( [expr $min + $u($k$i)] < $r_unic ) && 
		#    ( [expr $newSP/$newSS] < [expr $SP($i)/$SS($i)] ) )} {
		#	set sim 1
		#    }
		#}
		if { $sim } {
		    set U($i) [expr $min + $u($k$i)]
		    set ph($i) $k
		    set SS($i) [expr $SS($k) - 1]
		    set SP($i) [expr $SP($k) - ( $u($k$i)-floor($u($k$i)) )]
		}
	    }
	}
	if { $no_route == 0 } {
	    puts stderr "Ce: $C"
	    break
	}
    }
}





Agent/rtProto/SMM_LS instproc intf-changed {} {
    $self instvar ns_ peers_ ifs_ ifstat_ ifsUp_ nextHop_ nextHopPeer_ metric_
    $self instvar node_

    set UNREACH_      [$class set UNREACH_]
    set ifsUp_ ""
    puts stderr "Entrei em intf-changed, no: [$node_ id], instante: [$ns_ now]"
    foreach nbr [array names peers_] {
	set state [$ifs_($nbr) up?]
	if {$state != $ifstat_($nbr)} {
	    set ifstat_($nbr) $state
	    if {$state != "up"} {
		if ![info exists all-nodes] {
		    set all-nodes [$ns_ all-nodes-list]
		}
		foreach dest ${all-nodes} {
		    $peers_($nbr) metric  $dest $UNREACH_
		}
	    } else {
		lappend ifsUp_ $nbr
	    }
	}
    }
}

Agent/rtProto/SMM_LS proc get-next-mid {} {
    set ret [Agent/rtProto/SMM_LS set mid_]
    Agent/rtProto/SMM_LS set mid_ [expr $ret + 1]
    set ret
}

Agent/rtProto/SMM_LS proc retrieve-msg id {
    set ret [Agent/rtProto/SMM_LS set msg_($id)]
    Agent/rtProto/SMM_LS unset msg_($id)
    set ret
}

Agent/rtProto/SMM_LS instproc send-updates changes {
    $self instvar peers_ ifs_ ifsUp_
    $self instvar ns_ node_                                   ;# luish!!!!!!!
    
    #puts stderr "SMM_LS-send-updates, node: [$node_ id], instante: [$ns_ now]"
    if $changes {
	set to-send-to [array names peers_]
    } else {
	set to-send-to $ifsUp_
    }
    set ifsUp_ ""
    foreach nbr ${to-send-to} {
	if { [$ifs_($nbr) up?] == "up" } {
	    $self send-to-peer $nbr
	}
    }

    #here we send the link-state updates...
    set nodeslist [[Simulator instance] all-nodes-list]
    foreach node $nodeslist {
	$self send-link-state $node
    }   
}

Agent/rtProto/SMM_LS instproc send-to-peer nbr {
    $self instvar ns_ rtObject_ ifs_ peers_
    $self instvar node_                          ;# luish!!!!!!!!!!

    set UNREACH_      [$class set UNREACH_]

    #puts stderr "SMM_LS-send-to-peer, node: [$node_ id], instante: [$ns_ now]"

    set update(0)  0    ;#type of message - protocol update message - 0

    foreach dest [$ns_ all-nodes-list] {
	set metric  [$rtObject_ metric?  $dest]
	
	if {$metric < 0 || $metric >= $UNREACH_ } {
	    set update($dest) $UNREACH_
	} else {
	    set update($dest) [$rtObject_ metric? $dest]
	    foreach nh [$rtObject_ nextHop? $dest] {
		if {$nh == $ifs_($nbr)} {
		    set update($dest) $UNREACH_
		}
	    }
	}
    }

    ### modifed by Liang Guo, 11/11/99, what if there's no peer on that end?
    ### needed when only part of the network nodes are using DV routing
    if { $peers_($nbr) == "" } {
        return
    }
    ##################### End ##########                                            
    set id [$class get-next-mid]
    $class set msg_($id) [array get update]
    
    # XXX Note the singularity below...
    $self send-update [$peers_($nbr) addr?] [$peers_($nbr) port?] $id [array size update]
    #puts stderr "Saidasend-to-peer"       ;# luish
}

Agent/rtProto/SMM_LS instproc send-link-state dest {
    $self instvar ns_ rtObject_ ifs_ peers_
    $self instvar node_ link_m1_

    #set UNREACH_      [$class set UNREACH_]
    #set WST_DELAY [$class set WST_DELAY]
    #set WST_BW    [$class set WST_BW]
    #set WST_LP    [$class set WST_LP]
    #set WST_UNIC  [$class set WST_UNIC]

    #puts stderr "SMM_LS-send-link-state, node: [$node_ id], inst: [$ns_ now]"

    set update(0) 1  ;# type of protocol message: link-state - 1

    set this $node_
    foreach nbr [array names ifs_] {
	if { [$ifs_($nbr) up?] == "up" } {
	    set update([concat $this$nbr 0]) [$ifs_($nbr) cost?]
	    set update([concat $this$nbr 1]) [$ifs_($nbr) cost0?]
	    set update([concat $this$nbr 2]) [$ifs_($nbr) cost1?]
	    set update([concat $this$nbr 3]) [$ifs_($nbr) cost2?]
	    set update([concat $this$nbr 4]) [unica [$ifs_($nbr) cost0?] \
		    [$ifs_($nbr) cost2?]]
	    if { [$ifs_($nbr) cost1?] != $link_m1_($this$nbr) } {
		set link_m1_($this$nbr) [$ifs_($nbr) cost1?]
	    }

	} elseif {$nextHop_($nbr) != "" && [$ifs_($nbr) up?] != "up"} {
	    puts stderr "LINK DOWN"
	    set update([concat $this$nbr 0])  [$class set UNREACH_]   ;# hops
	    set update([concat $this$nbr 1])  [$class set WST_DELAY]  ;# atraso
	    set update([concat $this$nbr 2])  [$class set WST_BW]     ;# banda 
	    set update([concat $this$nbr 3])  [$class set WST_LP]     ;# prob 
	    set update([concat $this$nbr 4])  [$class set WST_UNIC]   ;# unica
	}
    }

    set id [$class get-next-mid]
    $class set msg_($id) [array get update]
    
    # XXX Note the singularity below...
    set rtobj [$dest rtObject?]
    if { $rtobj != "" } {
	set rtproto [$rtobj rtProto? SMM_LS]
	if { $rtproto != "" } {
	    #$proto($node) add-peer $nbr [$rtproto set agent_addr_]
	    $self send-update [$rtproto set agent_addr_] [$rtproto set agent_port_] $id [array size update]
	}
    }
}

Agent/rtProto/SMM_LS instproc recv-update {peerAddr id} {
    $self instvar peers_ ifs_ nextHopPeer_
    $self instvar metric_
    $self instvar rtsChanged_ rtObject_
    $self instvar ns_ node_
    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_
        
    set UNREACH_      [$class set UNREACH_]

    set msg [$class retrieve-msg $id]
    array set metrics $msg

    if { $metrics(0) == 0 } {
	foreach nbr [array names peers_] {
	    if {[$peers_($nbr) addr?] == $peerAddr} {
		set peer $peers_($nbr)
		
		set metricsChanged 0
		unset metrics(0)
		foreach dest [array names metrics] {
		    set metric [soma $metrics($dest) [$ifs_($nbr) cost?]]
		    if {$metric >= $UNREACH_} {
			set metric $UNREACH_
		    }
		    if {$metric != [$peer metric? $dest ]} {
			$peer metric $dest $metric
			incr metricsChanged
		    }
		}
		if $metricsChanged {
		    $self compute-routes
		    incr rtsChanged_ $metricsChanged
		    $rtObject_ SMM_LS-compute-routes
		} else {
		    # dynamicDM multicast hack.
		    # If we get a message from a neighbour, then something
		    # at that neighbour has changed.  While this may not
		    # cause any unicast changes on our end, dynamicDM
		    # looks at neighbour's routing tables to compute
		    # parent-child relationships, and has to do them
		    # again.
		    $rtObject_ flag-multicast -1
		}
		return
	    }
	}
	#error "$class::$proc update $peerAddr:$msg:$count from unknown peer"
    } elseif { $metrics(0) == 1 } {
	unset metrics(0)
	foreach pair [array names metrics] {
	    set pairlh [lindex $pair 0]
	    switch [lindex $pair 1] {
		0 { $rtObject_ set link_mH_($pairlh) $metrics($pair) }
		1 { $rtObject_ set link_m0_($pairlh) $metrics($pair) }
		2 { $rtObject_ set link_m1_($pairlh) $metrics($pair) }
		3 { $rtObject_ set link_m2_($pairlh) $metrics($pair) }
		4 { $rtObject_ set link_mU_($pairlh) $metrics($pair) }
	    }
	}
    }
}

Agent/rtProto/SMM_LS proc compute-all {} {
    # Because proc methods are not inherited from the parent class.
}
