#
# Maintainer: <Luis.Costa@lip6.fr>.
#

# This file only contains the methods for dynamic SMM routing
# Check ../lib/ns-route.tcl for the Simulator routing support
#

#
# Primeira versao do Roteamento com QoS
# Metrica Unica Combinada             
#

#Esse trecho estava em tcl/lib/ns-default.tcl          luish

Agent/rtProto/SMM set preference_ 90 ;# preferencia maior que direct(num menor)
Agent/rtProto/SMM set INFINITY            [Agent set ttl_]
Agent/rtProto/SMM set advertInterval       2

# Rotina que escolhe e instala as melhores rotas entre todos os protocolos,
# de acordo com sua preferencia (preference_).

rtObject instproc SMM-compute-routes {} {
    # choose the best route to each destination from all protocols
    $self instvar ns_ node_ rtProtos_ nullAgent_
    $self instvar nextHop_ rtpref_ metric_ rtVia_
    
    $self instvar metric0_ metric1_ metric2_ metricU_                  ;# luish
    $self instvar contador_

    set UNREACH_DELAY [$class set unreach_delay]
    set UNREACH_BW    [$class set unreach_bw]
    set UNREACH_LP    [$class set unreach_lp]
    set UNREACH_UNIC  [$class set unreach_unic]

    set protos ""
    set changes 0
    #puts stderr "entrada de rtObject SMM-compute-routes, no: [$node_ id]"
    #puts stderr "momento: [$ns_ now]   valor do contador: $contador_"
    foreach p [array names rtProtos_] {
	if [$rtProtos_($p) set rtsChanged_] {
	    incr changes
	    $rtProtos_($p) set rtsChanged_ 0
	}
	lappend protos $rtProtos_($p)
    }
    if !$changes return

    set changes 0

    set t0 [$ns_ now]                          ;# aqui produzimos o atraso na 
    set tempo [expr $t0 + (0.002 * $contador_)] ;# atualizacao das rotas
    #puts stderr "tempo inicial: $t0 tempo final: $tempo"
    puts stderr "no: [$node_ id] contador: $contador_"
    
    foreach dst [$ns_ all-nodes-list] {
	foreach CoS {0 1 2} {
	    if {$dst == $node_} continue   ;# no'=destino, passe para o proximo
	    set nh ""
	    set pf [$class set maxpref_]
	    set mt $UNREACH_DELAY
	    set mt1 $UNREACH_BW
	    set mt2 $UNREACH_LP
	    set mtU $UNREACH_UNIC
	    set rv ""
	    set dst_CoS $dst$CoS          ;# destino = destino + CoS
	    # para cada protocolo
	    foreach p $protos {                    
		set pnh [$p set nextHop_($dst_CoS)]   ;# 
		# nao havia rota, prox. prot.
		if { $pnh == "" } continue
		#se ha' rota, pegamos os valores de prot, pref e metricas
		set ppf [$p set rtpref_($dst_CoS)]  ;# preferencia do prot
		set pmt [$p set metric0_($dst_CoS)]  ;# primeira metrica do prot
		set pmt1 [$p set metric1_($dst_CoS)] ;# luish - metric1 do prot
		set pmt2 [$p set metric2_($dst_CoS)] ;# luish - metric2 do prot
		set pmtU [$p set metricU_($dst_CoS)] ;# luish - metricU do prot

		# se rota aceitavel...
		# Classe de Servico 0: privilegia o atraso
		if {$CoS==0} {            
		    if {$ppf < $pf || ($ppf == $pf && $pmt < $mt) \
			    || $mt >= $UNREACH_DELAY } {
			set nh  $pnh                   ;# proximo hop
			set pf  $ppf                   ;# pref da rota
			set mt  $pmt                   ;# metrica da rota
			set mt1  $pmt1                 ;# metric1 da rota
			set mt2  $pmt2                 ;# metric2 da rota
			set mtU  $pmtU                 ;# metricU da rota
			set rv  $p                     ;# protocolo
		    }
		}
	
		# Classe de Servico 1: privilegia a banda passante
		if {$CoS==1} {             
		    if {$ppf < $pf || ($ppf == $pf && $pmt1 > $mt1) \
			    || $mt1 <= $UNREACH_BW } {   
			set nh  $pnh                   ;# proximo hop
			set pf  $ppf                   ;# pref da rota
			set mt  $pmt                   ;# metrica da rota
			set mt1  $pmt1                 ;# metric1 da rota
			set mt2  $pmt2                 ;# metric2 da rota
			set mtU  $pmtU                 ;# metricU da rota
			set rv  $p                     ;# protocolo
		    }
		}

		# Classe de Servico 2: prob. de perda
		if {$CoS==2} {              
		    if {$ppf < $pf || ($ppf == $pf && $pmt2 < $mt2) \
			    || $mt2 >= $UNREACH_LP} {    
			set nh  $pnh                   ;# proximo hop
			set pf  $ppf                   ;# pref da rota
			set mt  $pmt                   ;# metrica da rota
			set mt1  $pmt1                 ;# metric1 da rota
			set mt2  $pmt2                 ;# metric2 da rota
			set mtU  $pmtU                 ;# metric2 da rota
			set rv  $p                     ;# protocolo
		    }
		}
	    } 
	    # foreach protos
	    
	    # se concluimos rota inexistente... apague as rotas existentes
	    if { $nh == "" } {
		if { $nextHop_($dst_CoS) != "" } {        
		    $node_ delete-qos-routes [$dst id] \
			    $nextHop_($dst_CoS) $nullAgent_ $CoS ;# luish
		    set nextHop_($dst_CoS)  $nh                        ;# luish
		    set rtpref_($dst_CoS)   $pf                        ;# luish
		    set metric0_($dst_CoS)   $mt                        ;# luish
		    set metric1_($dst_CoS)  $mt1                       ;# luish
		    set metric2_($dst_CoS)  $mt2                       ;# luish
		    set metricU_($dst_CoS)  $mtU                       ;# luish
		    set rtVia_($dst_CoS)    $rv                        ;# luish
		    incr changes
		}
		#se nao, existe uma rota. Vamos analisa-la
	    } else {
		#primeiro caso, o prot e' o mesmo que ja' tinha a melhor rota
		if { $rv == $rtVia_($dst_CoS) } {                         
		    # Current protocol still has best route.
		    # See if changed
		    # caso em que a rota mudou (prox salto)
		    if { $nh != $nextHop_($dst_CoS) } {                      
			$node_ delete-qos-routes [$dst id] \
				$nextHop_($dst_CoS) $nullAgent_ $CoS
			set nextHop_($dst_CoS) $nh                     ;# luish
			#$ns_ at $tempo 
				$node_ add-qos-routes [$dst id]  $nextHop_($dst_CoS) $CoS ;#luis
			incr changes
		    }
		    
		    # se houve alguma mudanca nas metricas da rota...	    
		    if { ($metric0_($dst_CoS) != $mt) || \
			    ($metric1_($dst_CoS) != $mt1) \
			    || ($metric2_($dst_CoS) != $mt2) } {          
			set metric0_($dst_CoS) $mt                      ;# luish
			set metric1_($dst_CoS) $mt1                    ;# luish
			set metric2_($dst_CoS) $mt2                    ;# luish
			set metricU_($dst_CoS) $mtU                    ;# luish
			incr changes
		    }
		    if { $pf != $rtpref_($dst_CoS) } {   
			set rtpref_($dst_CoS) $pf                             
		    }
		    #segundo caso, o prot nao e' o mesmo que tinha melhor rota
		} else {
		    if { $rtVia_($dst_CoS) != "" } {
			set nextHop_($dst_CoS) \
				[$rtVia_($dst_CoS) set nextHop_($dst_CoS)];#lui
			set rtpref_($dst_CoS)  \
				[$rtVia_($dst_CoS) set rtpref_($dst_CoS)] ;#lui
			set metric0_($dst_CoS)  \
				[$rtVia_($dst_CoS) set metric0_($dst_CoS)] ;#lui
			set metric1_($dst_CoS)  \
				[$rtVia_($dst_CoS) set metric1_($dst_CoS)];#lui
			set metric2_($dst_CoS)  \
				[$rtVia_($dst_CoS) set metric2_($dst_CoS)];#lui
			set metricU_($dst_CoS)  \
				[$rtVia_($dst_CoS) set metricU_($dst_CoS)];#lui
		    }
		    
		    if {$rtpref_($dst_CoS) != $pf || $metric0_($dst_CoS) != $mt || \
			    $metric1_($dst_CoS) != $mt1 || $metric2_($dst_CoS) != $mt2 } {
			# Then new prefs must be better, or
			# new prefs are equal, and new metrics are lower
			$node_ delete-qos-routes [$dst id] \
				$nextHop_($dst_CoS) $nullAgent_ $CoS  ;# luish
			set nextHop_($dst_CoS) $nh
			set rtpref_($dst_CoS)  $pf
			set metric0_($dst_CoS)  $mt
			set metric1_($dst_CoS)  $mt1
			set metric2_($dst_CoS)  $mt2
			set metricU_($dst_CoS)  $mtU
			set rtVia_($dst_CoS)   $rv
			#$ns_ at $tempo 
				$node_ add-qos-routes [$dst id] $nextHop_($dst_CoS) $CoS ;#luis
			incr changes
		    }
		}
	    }
	    #puts stderr "CoS: $CoS Dest: $dst rota: $nextHop_($dst$CoS)"
	    #puts stderr "Metricas: $metric0_($dst$CoS) $metric1_($dst$CoS) $metric2_($dst$CoS)"
	    
	}
	# foreach CoS
    }
    # foreach dest

    foreach proto [array names rtProtos_] {
	$rtProtos_($proto) send-updates $changes
    }
    #
    # XXX
    # detailed multicast routing hooks must come here.
    # My idea for the hook will be something like:
    # set mrtObject [$node_ mrtObject?]
    # if {$mrtObject != ""} { ;#
    #    $mrtObject recompute-mroutes $changes
    # }
    # $changes == 0	if only interfaces changed state.  Look at how
    #			Agent/rtProto/DV handles ifsUp_
    # $changes > 0	if new unicast routes were installed.
    #
    
    $self flag-multicast $changes
    #puts stderr "saida de rtObject SMM-compute-routes"                 ;# luish
}

###############################################################################
#
# Distance Vector Route Computation - SMM approach
#
###############################################################################

# Class Agent/rtProto/DV -superclass Agent/rtProto

#Agent/rtProto/DV set UNREACHABLE	[rtObject set unreach_]
Agent/rtProto/SMM set mid_		  0

Agent/rtProto/SMM set UNREACH_DELAY	[rtObject set unreach_delay]
Agent/rtProto/SMM set UNREACH_BW	[rtObject set unreach_bw]
Agent/rtProto/SMM set UNREACH_LP  	[rtObject set unreach_lp]
Agent/rtProto/SMM set UNREACH_UNIC	[rtObject set unreach_unic]

Agent/rtProto/SMM proc init-all args {

    if { [llength $args] == 0 } {
	set nodeslist [[Simulator instance] all-nodes-list]
    } else {
	#eval "set nodeslist $args"                          ;# luish - patched
	set nodeslist $args
    }
    Agent set-maxttl Agent/rtProto/SMM INFINITY
    eval rtObject init-all $nodeslist
    foreach node $nodeslist {
	set proto($node) [[$node rtObject?] add-proto SMM $node]
    }
    foreach node $nodeslist {
	foreach nbr [$node neighbors] {
	    set rtobj [$nbr rtObject?]
	    if { $rtobj != "" } {
		set rtproto [$rtobj rtProto? SMM]
		if { $rtproto != "" } {
		    $proto($node) add-peer $nbr [$rtproto set agent_addr_]
		}
	    }
	}
	# foreach nbr
    }
    # foreach node
}

Agent/rtProto/SMM instproc init node {
    global rtglibRNG

    $self next $node
    $self instvar ns_ rtObject_ ifsUp_
    $self instvar preference_ rtpref_ nextHop_ nextHopPeer_ metric_ multiPath_
    $self instvar metric0_ metric1_ metric2_ metricU_                           ;# luish

    set UNREACH_DELAY [$class set UNREACH_DELAY]
    set UNREACH_BW    [$class set UNREACH_BW]
    set UNREACH_LP    [$class set UNREACH_LP]
    set UNREACH_UNIC  [$class set UNREACH_UNIC]

    foreach dest [$ns_ all-nodes-list] {
	foreach CoS {0 1 2} {
	    set dest_CoS $dest$CoS
	    set rtpref_($dest_CoS) $preference_                        ;# luish
	    set nextHop_($dest_CoS) ""                                 ;# luish
	    set nextHopPeer_($dest_CoS) ""                             ;# luish
	    set metric0_($dest_CoS)   $UNREACH_DELAY                    ;# luish
	    set metric1_($dest_CoS)  $UNREACH_BW                       ;# luish
	    set metric2_($dest_CoS)  $UNREACH_LP                       ;# luish
	    set metricU_($dest_CoS)  $UNREACH_UNIC                     ;# luish
	}
    }
    
    set ifsUp_ ""
    set multiPath_ [[$rtObject_ set node_] set multiPath_]
    set updateTime [$rtglibRNG uniform 0.0 0.005]
    #mudei o tempo inicial de 0.5 para 0.05 para aumentar a capacidade 
    #de nos com poucos vizinhos calcularem as rotas
    $ns_ at $updateTime "$self send-periodic-update"
}

Agent/rtProto/SMM instproc add-peer {nbr agentAddr} {
    $self instvar peers_
    $self set peers_($nbr) [new rtPeer $agentAddr $class]
}

Agent/rtProto/SMM instproc send-periodic-update {} {
    global rtglibRNG

    $self instvar ns_
    #puts stderr "send-periodic-update entrada, instante: [$ns_ now]"  ;# luish
    $self send-updates 1	;# Anything but 0
    set updateTime [expr [$ns_ now] + \
	    ([$class set advertInterval] * [$rtglibRNG uniform 0.9 1.1])]
    $ns_ at $updateTime "$self send-periodic-update"
}

Agent/rtProto/SMM instproc compute-routes {} {
    $self instvar ns_ ifs_ rtpref_ metric_ nextHop_ nextHopPeer_
    $self instvar peers_ rtsChanged_ multiPath_
    $self instvar metric0_ metric1_ metric2_ metricU_ node_            ;# luish
    $self instvar rtObject_					       ;# luish

    #puts stderr "entrada de SMM compute-routes, no: [$node_ id]"      ;# luish

    set MAX_DELAY	 [rtObject set max_delay]                  ;# luish
    set MIN_BW	         [rtObject set min_bw]                     ;# luish
    set MAX_LP	         [rtObject set max_lp]                     ;# luish
    set MAX_UNIC         [rtObject set max_unic]                   ;# luish

    set MAXPREF  [rtObject set maxpref_]
    
    set UNREACH_DELAY	 [rtObject set unreach_delay]                  ;# luish
    set UNREACH_BW	 [rtObject set unreach_bw]                     ;# luish
    set UNREACH_LP	 [rtObject set unreach_lp]                     ;# luish
    set UNREACH_UNIC	 [rtObject set unreach_unic]                   ;# luish
    set rtsChanged_ 0

    set counter 1                ;# variavel que vai conter o numero de passos 
				 ;# gasto na descoberta das rotas - luish

    foreach dst [$ns_ all-nodes-list] {
	#
	#   Classe de Servico 0 - Privilegia o Atraso primeiramente
	#
	set CoS 0
	if {$CoS==0} {
	    catch "unset nh"
	    #puts stderr "Calculo SMM CoS 0, no: $node_ ([$node_ id]) \
		    dst: $dst ([$dst id]) $CoS"
	    #puts stderr "Momento: [$ns_ now]"
	    #puts stderr "peers: [array names peers_]"
	    set p [lindex $nextHopPeer_($dst$CoS) 0]      ;# luish
	    if {$p != ""} {
		set metric0_($dst$CoS) [$p metric0? $dst$CoS]
		set metric1_($dst$CoS) [$p metric1? $dst$CoS]
		set metric2_($dst$CoS) [$p metric2? $dst$CoS]
		set metricU_($dst$CoS) [$p metricU? $dst$CoS]
		set rtpref_($dst$CoS) [$p preference? $dst$CoS]
	    }
	    
	    set pf  $MAXPREF
	    set mt  $UNREACH_DELAY
	    set mt1 $UNREACH_BW                                        ;# luish
	    set mt2 $UNREACH_LP                                        ;# luish
	    set mtU $UNREACH_UNIC                                      ;# luish
	    set nh(0) 0

	    foreach nbr [array names peers_] {
		set pmt  [$peers_($nbr) metric0? $dst$CoS]
		set pmt1 [$peers_($nbr) metric1? $dst$CoS]         ;# luish
		set pmt2 [$peers_($nbr) metric2? $dst$CoS]         ;# luish
		set pmtU [$peers_($nbr) metricU? $dst$CoS]         ;# luish
		set ppf [$peers_($nbr) preference? $dst$CoS]

		# if peer metric not valid	continue
		# if peer pref higher		continue
		# if peer pref lower		set to latest values
		# else peer pref equal
		#	if peer metric higher	continue
		#	if peer metric lower	set to latest values
		# 	else peer metrics equal	append latest values
	    
		set pmtdummy $pmt
		set mtdummy  $mt
		#puts stderr "pmt: $pmt mt: $mt mt1: $mt1 pmt1: $pmt1 \
			mt2: $mt2 pmt2: $pmt2"
		#puts stderr "pmtdummy: $pmtdummy, mtdummy: $mtdummy"
		if { $pmtU < 0 || $pmtU >= $UNREACH_UNIC || $pmt1 < $MIN_BW ||\
			$ppf > $pf || $pmtU > $mtU || $pmt2 > $MAX_LP } {
	            #puts stderr "caso1"
		    continue
		}
		if { $ppf < $pf || $pmtU < $mtU } { 
		    # metrica do peer melhor
		    set pf $ppf
		    set mtdummy $pmtdummy
		    set mt $pmt
		    set mt1 $pmt1
		    set mt2 $pmt2
		    set mtU $pmtU
		    unset nh	   ;# because we must compute *new* next hops
		}                   
		# senao, metrica igual
		set nh($ifs_($nbr)) $peers_($nbr)
		set counter [expr $counter + 1]  ;# incremento do num de passos
	    }    
	    #foreach nbr

	    #puts stderr "CoS: $CoS nh inicial: [array names nh]"
	    #puts stderr "mt: $mt mt1: $mt1 mt2: $mt2"

	    #puts stderr "contador depois das buscas: $counter"
	    $rtObject_ set contador_ $counter

	    set metric_dummy $metric0_($dst$CoS)

	    catch "unset nh(0)"               ;# adiantamos este unset - luish

	    if { $pf == $MAXPREF && $mtdummy == $UNREACH_DELAY } continue
	    if { $pf > $rtpref_($dst$CoS) ||		\
		    ($metric_dummy >= 0 && $mtdummy >         \
		    $metric_dummy ) }                     \
		    continue
	    if {$mtdummy >= $MAX_DELAY || $mt1 <= $UNREACH_BW || $mt2 >= $UNREACH_LP} {
		set mtdummy $UNREACH_DELAY
		set mt $UNREACH_DELAY
		set mt1 $UNREACH_BW
		set mt2 $UNREACH_LP
		set mtU $UNREACH_UNIC
	    }

	    incr rtsChanged_
	    if {$pf < $rtpref_($dst$CoS) ||               \
		    $mtdummy < $metric0_($dst$CoS) } {
		set rtpref_($dst$CoS) $pf                        ;# luish
		set metric0_($dst$CoS) $mt                        ;# luish
		set metric1_($dst$CoS) $mt1                      ;# luish
		set metric2_($dst$CoS) $mt2                      ;# luish
		set metricU_($dst$CoS) $mtU                      ;# luish
		set nextHop_($dst$CoS) ""                        ;# luish
		set nextHopPeer_($dst$CoS) ""                    ;# luish
		foreach n [array names nh] {
		    lappend nextHop_($dst$CoS) $n
		    lappend nextHopPeer_($dst$CoS) $nh($n)       ;# luish
		    if !$multiPath_ break;
		}
		continue
	    }

	    set rtpref_($dst$CoS) $pf
	    set metric0_($dst$CoS) $mt
	    set metric1_($dst$CoS) $mt1
	    set metric2_($dst$CoS) $mt2
	    set metricU_($dst$CoS) $mtU
	    set newNextHop ""
	    set newNextHopPeer ""
	    foreach rt $nextHop_($dst$CoS) {
		if [info exists nh($rt)] {
		    lappend newNextHop $rt
		    lappend newNextHopPeer $nh($rt)
		    unset nh($rt)
		}
	    }
      
	    set nextHop_($dst$CoS) $newNextHop                   ;# luish
	    set nextHopPeer_($dst$CoS) $newNextHopPeer           ;# luish
	    if { $multiPath_ || $nextHop_($dst$CoS) == "" } {
		foreach rt [array names nh] {
		    lappend nextHop_($dst$CoS) $rt                   ;# luish
		    lappend nextHopPeer_($dst$CoS) $nh($rt)          ;# luish
		    if !$multiPath_ break
		}
	    }
	}
	# fim de CoS 0
	
	#puts stderr "Metricas no fim de CoS 0: $metric0_($dst$CoS), \
		$metric1_($dst$CoS), $metric2_($dst$CoS)"
	
	#
	#   Classe de Servico 1 - Privilegia a banda passante
	#
	set CoS 1
	if {$CoS==1} {
	    catch "unset nh"
	    #puts stderr "Calculo SMM CoS 1, no: $node_ ([$node_ id]) \
		    dst: $dst ([$dst id]) $CoS"
	    #puts stderr "Momento: [$ns_ now]"
	    #puts stderr "peers: [array names peers_]"
	    set p [lindex $nextHopPeer_($dst$CoS) 0]      ;# luish
	    if {$p != ""} {
		set metric0_($dst$CoS) [$p metric0? $dst$CoS]
		set metric1_($dst$CoS) [$p metric1? $dst$CoS]
		set metric2_($dst$CoS) [$p metric2? $dst$CoS]
		set metricU_($dst$CoS) [$p metricU? $dst$CoS]
		set rtpref_($dst$CoS) [$p preference? $dst$CoS]
	    }

	    set pf $MAXPREF
	    set mt  $UNREACH_DELAY
	    set mt1 $UNREACH_BW                                    ;# luish
	    set mt2 $UNREACH_LP                                    ;# luish
	    set mtU $UNREACH_UNIC                                  ;# luish
	    set nh(0) 0
	    
	    foreach nbr [array names peers_] {
		set pmt  [$peers_($nbr) metric0? $dst$CoS]
		set pmt1 [$peers_($nbr) metric1? $dst$CoS]         ;# luish
		set pmt2 [$peers_($nbr) metric2? $dst$CoS]         ;# luish
		set pmtU [$peers_($nbr) metricU? $dst$CoS]         ;# luish
		set ppf [$peers_($nbr) preference? $dst$CoS]

		# if peer metric not valid	continue
		# if peer pref higher		continue
		# if peer pref lower		set to latest values
		# else peer pref equal
		#	if peer metric higher	continue
		#	if peer metric lower	set to latest values
		# 	else peer metrics equal	append latest values

		set pmtdummy $pmt1
		set mtdummy  $mt1
		#puts stderr "pmt: $pmt mt: $mt mt1: $mt1 pmt1: $pmt1 \
			mt2: $mt2 pmt2: $pmt2"
		#puts stderr "pmtdummy: $pmtdummy, mtdummy: $mtdummy"
		
		if { $pmtdummy == 0 || $pmtdummy <= $MIN_BW || \
			$ppf > $pf || $pmtdummy < $mtdummy } \
			continue
		if { $ppf < $pf || $pmtdummy > $mtdummy } {
		    #quanto mais banda melhor
		    set pf $ppf
		    set mtdummy $pmtdummy
		    set mt $pmt
		    set mt1 $pmt1
		    set mt2 $pmt2
		    set mtU $pmtU
		    unset nh     ;# because we must compute *new* next hops
		}
		set nh($ifs_($nbr)) $peers_($nbr)
		set counter [expr $counter + 1] ;# incr do num de passos
	    }    
	    #foreach nbr
	    
	    #puts stderr "CoS: $CoS nh inicial: [array names nh]"
	    #puts stderr "mt: $mt mt1: $mt1 mt2: $mt2"
	    
	    # Buscamos agora pela segunda metrica - unica combinada
	    catch "unset nh(0)"
	    catch "unset nh_iguais"
	    set met_min $UNREACH_UNIC
	    set nh_iguais(0) 0
	    foreach viz [array names nh] {
		set met_atual [$nh($viz) metricU? $dst$CoS]            ;# unica
		if { $met_atual > $met_min } {
		    catch "unset nh($viz)"
		    #puts stderr "menor, nh_i: [array names nh_iguais]"
		}
		if { $met_atual == $met_min } {
		    set nh_iguais($viz) 81
		    set mt2 [$nh($viz) metric2? $dst$CoS]  ;# prob.
		    set mt  [$nh($viz) metric0?  $dst$CoS]  ;# atraso
		    #puts stderr "igual, nh_i: [array names nh_iguais]"
		}
		if { $met_atual < $met_min } {          
		    # quanto menor, melhor
		    set met_min $met_atual
		    set mt2 [$nh($viz) metric2? $dst$CoS] ;# prob.
		    set mt  [$nh($viz) metric0? $dst$CoS]  ;# atraso
		    foreach lui [array names nh_iguais] {
			catch "unset nh($lui)"
		    }
		    set nh_iguais($viz) 82
		    #puts stderr "maior, nh_i: [array names nh_iguais]"
		}
		set nbr_ant $viz
		set counter [expr $counter + 1] ;# incr do num de passos
	    }
	    #puts stderr "nh apos primeira busca: [array names nh]"
	    set mtU $met_min
	    
	    # Buscamos agora pela terceira metrica - perda
	    #  	  catch "unset nh_iguais"
	    # 	  set met_min 1
	    # 	  set nh_iguais(0) 0
	    # 	  foreach viz [array names nh] {
	    # 	   set met_atual [$nh($viz) metric2? $dst$CoS] ;# perda
	    #  	      if { $met_atual > $met_min } {
	    #  		  catch "unset nh($viz)"
	    #  	puts stderr "menor, nh_i: [array names nh_iguais]"
	    #  	      }
	    # 	      if { $met_atual == $met_min } {
	    # 		  set nh_iguais($viz) 83
	    # 		set mtU [$nh($viz) metricU? $dst$CoS]  ;# unica
	    #       puts stderr "igual, nh_i: [array names nh_iguais]"
	    # 	      }
	    # 	      if { $met_atual < $met_min } {         
	    # quanto menor, melhor
	    # 		  set met_min $met_atual
	    # 	  set mtU [$nh($viz) metricU? $dst$CoS]  ;# unica
	    # 	          foreach lui [array names nh_iguais] {
	    # 		      catch "unset nh($lui)"
	    # 		  }
	    # 		  set nh_iguais($viz)  84
	    #  	puts stderr "maior, nh_i: [array names nh_iguais]"
	    #  	      }
	    #   set counter [expr $counter + 1] 
	    # incremento do numero de passos
	    # 	  }
	    # 	  puts stderr "nh apos segunda busca: [array names nh]"
	    # 	  set mt2 $met_min

	    #puts stderr "contador depois das buscas: $counter"
	    $rtObject_ set contador_ $counter
	    
	    set metric_dummy $metric1_($dst$CoS)
	    
	    catch "unset nh(0)"
	    
	    if { $pf == $MAXPREF && $mtdummy == $UNREACH_BW } continue
	    if { $pf > $rtpref_($dst$CoS) ||		\
		    ($metric_dummy > $MIN_BW && $mtdummy < \
		    $metric_dummy ) }                     \
		    continue                                       ;# rever
	    if {$mtdummy <= $MIN_BW || $mt >= $UNREACH_DELAY || $mt2 >= $UNREACH_LP} {
		set mtdummy $UNREACH_BW
		set mt  $UNREACH_DELAY
		set mt1 $UNREACH_BW
		set mt2 $UNREACH_LP
		set mtU $UNREACH_UNIC
	    }
	    
	    incr rtsChanged_
	    if {$pf < $rtpref_($dst$CoS) || \
		    $mtdummy > $metric1_($dst$CoS) } {
		set rtpref_($dst$CoS) $pf                          ;# luish
		set metric0_($dst$CoS) $mt                          ;# luish
		set metric1_($dst$CoS) $mt1                        ;# luish
		set metric2_($dst$CoS) $mt2                        ;# luish
		set metricU_($dst$CoS) $mtU                        ;# luish
		set nextHop_($dst$CoS) ""                          ;# luish
		set nextHopPeer_($dst$CoS) ""                      ;# luish
		foreach n [array names nh] {
		    lappend nextHop_($dst$CoS) $n
		    lappend nextHopPeer_($dst$CoS) $nh($n)         ;# luish
		    if !$multiPath_ break;
		}
		continue
	    }
	    
	    set rtpref_($dst$CoS) $pf
	    set metric0_($dst$CoS) $mt
	    set metric1_($dst$CoS) $mt1
	    set metric2_($dst$CoS) $mt2
	    set metricU_($dst$CoS) $mtU
	    set newNextHop ""
	    set newNextHopPeer ""
	    foreach rt $nextHop_($dst$CoS) {
		if [info exists nh($rt)] {
		    lappend newNextHop $rt
		    lappend newNextHopPeer $nh($rt)
		    unset nh($rt)
		}
	    }
	    
	    set nextHop_($dst$CoS) $newNextHop                   ;# luish
	    set nextHopPeer_($dst$CoS) $newNextHopPeer           ;# luish
	    if { $multiPath_ || $nextHop_($dst$CoS) == "" } {
		foreach rt [array names nh] {
		    lappend nextHop_($dst$CoS) $rt                 ;# luish
		    lappend nextHopPeer_($dst$CoS) $nh($rt)        ;# luish
		    if !$multiPath_ break
		}
	    }
	}      
	# fim de CoS 1
	
	#puts stderr "Metricas no fim de CoS 1: $metric0_($dst$CoS), \
		$metric1_($dst$CoS), $metric2_($dst$CoS)"
	
	#
	#   Classe de Servico 2
	#
	set CoS 2
	if {$CoS==2} {
	    catch "unset nh"
	    #puts stderr "Calculo SMM CoS 2, no: $node_ ([$node_ id]) \
		    dst: $dst ([$dst id]) $CoS"
	    #puts stderr "Momento: [$ns_ now]"
	    #puts stderr "peers: [array names peers_]"
	    set p [lindex $nextHopPeer_($dst$CoS) 0]      ;# luish
	    if {$p != ""} {
		set metric0_($dst$CoS) [$p metric0? $dst$CoS]
		set metric1_($dst$CoS) [$p metric1? $dst$CoS]
		set metric2_($dst$CoS) [$p metric2? $dst$CoS]
		set metricU_($dst$CoS) [$p metricU? $dst$CoS]
		set rtpref_($dst$CoS) [$p preference? $dst$CoS]
	    }
	    
	    set pf $MAXPREF
	    set mt  $UNREACH_DELAY
	    set mt1 $UNREACH_BW                                ;# luish
	    set mt2 $UNREACH_LP                                ;# luish
	    set mtU $UNREACH_UNIC                              ;# luish
	    set nh(0) 0
		    
	    foreach nbr [array names peers_] {
		set pmt  [$peers_($nbr) metric0? $dst$CoS]
		set pmt1 [$peers_($nbr) metric1? $dst$CoS]     ;# luish
		set pmt2 [$peers_($nbr) metric2? $dst$CoS]     ;# luish
		set pmtU [$peers_($nbr) metricU? $dst$CoS]     ;# luish
		set ppf [$peers_($nbr) preference? $dst$CoS]
		
		# if peer metric not valid	continue
		# if peer pref higher		continue
		# if peer pref lower		set to latest values
		# else peer pref equal
		#	if peer metric higher	continue
		#	if peer metric lower	set to latest values
		# 	else peer metrics equal	append latest values

		set pmtdummy $pmt2
		set mtdummy  $mt2
		#puts stderr "pmt: $pmt mt: $mt mt1: $mt1 pmt1: $pmt1 \
			mt2: $mt2 pmt2: $pmt2"
		#puts stderr "pmtdummy: $pmtdummy, mtdummy: $mtdummy"
		
		#if { $pmtdummy < 0 || $pmtdummy >= $MAX_LP || \
			$pmt1 < $MIN_BW || \
			$ppf > $pf || $pmtdummy > $mtdummy } 
		if { $pmtU < 0 || $pmtU >= $UNREACH_UNIC || \
			$pmt1 < $MIN_BW || $ppf > $pf || \
			$pmtU > $mtU || $pmt2 > $MAX_LP } \
			continue
		#	  if { $ppf < $pf || $pmtdummy < $mtdummy } 
		if { $ppf < $pf || $pmtU < $mtU } {
		    set pf $ppf
		    set mtdummy $pmtdummy
		    set mt $pmt
		    set mt1 $pmt1
		    set mt2 $pmt2
		    set mtU $pmtU
		    unset nh ;# because we must compute *new* next hops
		}
		set nh($ifs_($nbr)) $peers_($nbr)
		set counter [expr $counter + 1]  ;# incremento do num de passos
	    }    
	    #foreach nbr
      
	    #puts stderr "CoS: $CoS nh inicial: [array names nh]"
	    #puts stderr "mt: $mt mt1: $mt1 mt2: $mt2"

	    # Buscamos agora pela segunda metrica - atraso
	    #         catch "unset nh(0)"
	    #	  catch "unset nh_iguais"
	    #	  set met_min 32
	    #	  set nh_iguais(0) 0
	    #	  foreach viz [array names nh] {
	    #	      set met_atual [$nh($viz) metric0? $dst$CoS] ;# atraso
	    #	      if { $met_atual > $met_min } {
	    #	 	  catch "unset nh($viz)"
	    #		  puts stderr "menor, nh_i: [array names nh_iguais]"
	    #	      }	
	    #	      if { $met_atual == $met_min } {
	    #		  set nh_iguais($viz) 91
	    #		  set mt [$nh($viz) metric0? $dst$CoS] ;# atraso
	    #		  set mtU [$nh($viz) metricU? $dst$CoS] ;# unica
	    #		  puts stderr "igual, nh_i: [array names nh_iguais]"
	    #	      }	  
	    #	      if { $met_atual < $met_min } {        
	    # quanto menor, melhor
	    #		  set met_min $met_atual
	    #		  set mt [$nh($viz) metric0? $dst$CoS] ;# atraso
	    #		  set mtU [$nh($viz) metricU? $dst$CoS] ;# unica
	    #	            foreach lui [array names nh_iguais] {
	    #		        catch "unset nh($lui)"
	    #		     }
	    #		  set nh_iguais($viz) 92
	    #               puts stderr "maior, nh_i: [array names nh_iguais]"
	    #	      }
	    #	      set counter [expr $counter + 1] ;# incr do num de passos
	    #	  }
	    #	  puts stderr "nh apos primeira busca: [array names nh]"
	    #	  set mt $met_min
	    #
	    # Buscamos agora pela terceira metrica - banda
	    #	  catch "unset nh(0)"
	    #          catch "unset nh_iguais"
	    # 	  set met_max 0
	    #	  set nh_iguais(0) 0
	    # 	  foreach viz [array names nh] {
	    # 	      set met_atual [$nh($viz) metric1? $dst$CoS] ;# banda
	    #	      if { $met_atual < $met_max } {
	    #		  catch "unset nh($viz)"
	    #		  puts stderr "menor, nh_i: [array names nh_iguais]"
	    #	      }
	    #	      if { $met_atual == $met_max } {
	    #		  set nh_iguais($viz) 93
	    #		  set mtU [$nh($viz) metricU? $dst$CoS] ;# unica
	    #		  puts stderr "igual, nh_i: [array names nh_iguais]"
	    #	      }
	    #	      if { $met_atual > $met_max } {         
	    # quanto maior, melhor
	    #		  set met_max $met_atual
	    #		  set mtU [$nh($viz) metricU? $dst$CoS] ;# unica
	    #            foreach lui [array names nh_iguais] {
	    #		         catch "unset nh($lui)"
	    #		      }
	    #		  set nh_iguais($viz) 94
	    #                puts stderr "maior, nh_i: [array names nh_iguais]"
	    #	      }
	    #	      set counter [expr $counter + 1] ;# incr do num de passos
	    #	  }
	    #	  puts stderr "nh apos segunda busca: [array names nh]"
	    #	  set mt1 $met_max

	    #puts stderr "contador depois das buscas: $counter"
	    $rtObject_ set contador_ $counter

	    set metric_dummy $metric2_($dst$CoS)
	    
	    catch "unset nh(0)"       ;# adiantamos este unset - luish
	    
	    if { $pf == $MAXPREF && $mtdummy == $UNREACH_LP } continue
	    if { $pf > $rtpref_($dst$CoS) ||		\
		    ($metric_dummy < $UNREACH_LP && $mtdummy >         \
		    $metric_dummy ) }                     \
		    continue                                           ;# rever
	    if {$mtdummy >= $MAX_LP || $mt >= $UNREACH_DELAY || $mt1 <= $UNREACH_BW} {
		set mtdummy $UNREACH_LP
		set mt $UNREACH_DELAY
		set mt1 $UNREACH_BW
		set mt2 $UNREACH_LP
		set mtU $UNREACH_UNIC
	    }
	    
	    incr rtsChanged_
	    if {$pf < $rtpref_($dst$CoS) ||               \
		    $mtdummy < $metric2_($dst$CoS) } {
		set rtpref_($dst$CoS) $pf                          ;# luish
		set metric0_($dst$CoS) $mt                          ;# luish
		set metric1_($dst$CoS) $mt1                        ;# luish
		set metric2_($dst$CoS) $mt2                        ;# luish
		set metricU_($dst$CoS) $mtU                        ;# luish
		set nextHop_($dst$CoS) ""                          ;# luish
		set nextHopPeer_($dst$CoS) ""                      ;# luish
		foreach n [array names nh] {
		    lappend nextHop_($dst$CoS) $n
		    lappend nextHopPeer_($dst$CoS) $nh($n)           ;# luish
		    if !$multiPath_ break;
		}
		continue
	    }

	    set rtpref_($dst$CoS) $pf
	    set metric0_($dst$CoS) $mt
	    set metric1_($dst$CoS) $mt1
	    set metric2_($dst$CoS) $mt2
	    set metricU_($dst$CoS) $mtU
	    set newNextHop ""
	    set newNextHopPeer ""
	    foreach rt $nextHop_($dst$CoS) {
		if [info exists nh($rt)] {
		    lappend newNextHop $rt
		    lappend newNextHopPeer $nh($rt)
		    unset nh($rt)
		}
	    }
      
	    set nextHop_($dst$CoS) $newNextHop                   ;# luish
	    set nextHopPeer_($dst$CoS) $newNextHopPeer           ;# luish
	    if { $multiPath_ || $nextHop_($dst$CoS) == "" } {
		foreach rt [array names nh] {
		    lappend nextHop_($dst$CoS) $rt                   ;# luish
		    lappend nextHopPeer_($dst$CoS) $nh($rt)          ;# luish
		    if !$multiPath_ break
		}
	    }
	}      
	# fim de CoS 2
      
	#puts stderr "Metricas no fim de CoS 2: $metric0_($dst$CoS), \
		$metric1_($dst$CoS), $metric2_($dst$CoS)"
    }
    # foreach dst

    set rtsChanged_
    #puts stderr "saida de SMM compute-routes, no: [$node_ id]"    ;# luish
}

Agent/rtProto/SMM instproc intf-changed {} {
    $self instvar ns_ peers_ ifs_ ifstat_ ifsUp_ nextHop_ nextHopPeer_ metric_
    $self instvar metric0_ metric1_ metric2_ metricU_ node_

    set UNREACH_DELAY [$class set UNREACH_DELAY]
    set UNREACH_BW [$class set UNREACH_BW]
    set UNREACH_LP [$class set UNREACH_LP]
    set UNREACH_UNIC [$class set UNREACH_UNIC]

    set zero 0
    set um 1
    set dois 2

    set ifsUp_ ""
    #puts stderr "Entrei em intf-changed, no: [$node_ id], instante: [$ns_ now]"
    foreach nbr [array names peers_] {
      set state [$ifs_($nbr) up?]
      if {$state != $ifstat_($nbr)} {
	set ifstat_($nbr) $state
	if {$state != "up"} {
	  if ![info exists all-nodes] {
	    set all-nodes [$ns_ all-nodes-list]
	  }
	  foreach dest ${all-nodes} {
            $peers_($nbr) metric0  $dest$zero $UNREACH_DELAY
            $peers_($nbr) metric0  $dest$um   $UNREACH_DELAY ;# ???
            $peers_($nbr) metric0  $dest$dois $UNREACH_DELAY ;# ???
	    $peers_($nbr) metric1 $dest$zero $UNREACH_BW       ;#luish
	    $peers_($nbr) metric1 $dest$um   $UNREACH_BW       ;#luish
	    $peers_($nbr) metric1 $dest$dois $UNREACH_BW     ;#luish
	    $peers_($nbr) metric2 $dest$zero $UNREACH_LP     ;#luish
	    $peers_($nbr) metric2 $dest$um   $UNREACH_LP     ;#luish
	    $peers_($nbr) metric2 $dest$dois $UNREACH_LP     ;#luish
	    $peers_($nbr) metricU $dest$zero $UNREACH_UNIC   ;#luish
	    $peers_($nbr) metricU $dest$um   $UNREACH_UNIC   ;#luish
	    $peers_($nbr) metricU $dest$dois $UNREACH_UNIC   ;#luish
	  }
        } else {
	    lappend ifsUp_ $nbr
	}
      }
    }
}

Agent/rtProto/SMM proc get-next-mid {} {
    set ret [Agent/rtProto/SMM set mid_]
    Agent/rtProto/SMM set mid_ [expr $ret + 1]
    set ret
}

Agent/rtProto/SMM proc retrieve-msg id {
    set ret [Agent/rtProto/SMM set msg_($id)]
    Agent/rtProto/SMM unset msg_($id)
    set ret
}

Agent/rtProto/SMM instproc send-updates changes {
    $self instvar peers_ ifs_ ifsUp_
    $self instvar ns_ node_                                   ;# luish!!!!!!!!!
    
    #puts stderr "Em SMM-send-updates, node: [$node_ id], instante: [$ns_ now]"
    if $changes {
	set to-send-to [array names peers_]
    } else {
	set to-send-to $ifsUp_
    }
    set ifsUp_ ""
    foreach nbr ${to-send-to} {
	if { [$ifs_($nbr) up?] == "up" } {
	    $self send-to-peer $nbr
	}
    }
}

Agent/rtProto/SMM instproc send-to-peer nbr {
    $self instvar ns_ rtObject_ ifs_ peers_
    $self instvar node_            ;# luish!!!!!!!!!!

    set UNREACH_DELAY [$class set UNREACH_DELAY]
    set UNREACH_BW [$class set UNREACH_BW]
    set UNREACH_LP [$class set UNREACH_LP]
    set UNREACH_UNIC [$class set UNREACH_UNIC]

    #puts stderr "Em SMM-send-to-peer, node: [$node_ id], instante: [$ns_ now]"
    foreach dest [$ns_ all-nodes-list] {
	foreach CoS {0 1 2} {
	    set controle 0
	    set controle1 0
	    set controle2 0
	    set controle3 0
	    set metric0  [$rtObject_ metric0? $dest$CoS]
	    set metric1 [$rtObject_ metric1? $dest$CoS]
	    set metric2 [$rtObject_ metric2? $dest$CoS]
	    set metricU [$rtObject_ metricU? $dest$CoS]
	    if {$metric0 < 0 || $metric0 >= $UNREACH_DELAY } {
		set update([concat $dest$CoS 0]) $UNREACH_DELAY
	    } else {
		set update([concat $dest$CoS 0]) \
			[$rtObject_ metric0? $dest$CoS]
		set controle 1
	    }
	    if {$metric1 < 0 || $metric1 <= $UNREACH_BW } {
		set update([concat $dest$CoS 1]) $UNREACH_BW
	    } else {
		set update([concat $dest$CoS 1]) \
			[$rtObject_ metric1? $dest$CoS]
		set controle1 1
	    }
	    if {$metric2 < 0.0 || $metric2 >= $UNREACH_LP } {
		set update([concat $dest$CoS 2]) $UNREACH_LP
	    } else {
		set update([concat $dest$CoS 2]) \
			[$rtObject_ metric2? $dest$CoS]
		set controle2 1
	    }
	    if {$metricU < 0.0 || $metricU >= $UNREACH_UNIC } {
		set update([concat $dest$CoS 3]) $UNREACH_UNIC
	    } else {
		set update([concat $dest$CoS 3]) \
			[$rtObject_ metricU? $dest$CoS]
		set controle3 1
	    }
	    if {$controle==1} {
		foreach nh [$rtObject_ nextHop? $dest$CoS] {
		    if {$nh == $ifs_($nbr)} {
			set update([concat $dest$CoS 0]) $UNREACH_DELAY
		    }
		}
	    }
	    if {$controle1==1} {
		foreach nh [$rtObject_ nextHop? $dest$CoS] {
		    if {$nh == $ifs_($nbr)} {
			set update([concat $dest$CoS 1]) $UNREACH_BW
		    }
		}
	    }
	    if {$controle2==1} {
		foreach nh [$rtObject_ nextHop? $dest$CoS] {
		    if {$nh == $ifs_($nbr)} {
			set update([concat $dest$CoS 2]) $UNREACH_LP
		    }
		}
	    }
	    if {$controle3==1} {
		foreach nh [$rtObject_ nextHop? $dest$CoS] {
		    if {$nh == $ifs_($nbr)} {
			set update([concat $dest$CoS 3]) $UNREACH_UNIC
		    }
		}
	    }
	}
    }
    
    set id [$class get-next-mid]
    $class set msg_($id) [array get update]
    
    # XXX Note the singularity below...
    $self send-update [$peers_($nbr) addr?] $id [array size update]
    #puts stderr "Saidasend-to-peer"       ;# luish
}

Agent/rtProto/SMM instproc recv-update {peerAddr id} {
    $self instvar peers_ ifs_ nextHopPeer_ metric_
    $self instvar metric0_ metric1_ metric2_ metricU_
    $self instvar rtsChanged_ rtObject_
    $self instvar ns_                                         ;# luish !!!!!!!!
    
    set routelogic [$ns_ get-routelogic]                             ;#luish
    
    set UNREACH_DELAY [$class set UNREACH_DELAY]
    set UNREACH_BW [$class set UNREACH_BW]
    set UNREACH_LP [$class set UNREACH_LP]
    set UNREACH_UNIC [$class set UNREACH_UNIC]

    set msg [$class retrieve-msg $id]
    array set metrics $msg

    #   set n [$rtObject_ set node_];					
    #   puts stderr [concat [format "<<< %7.5f" [[Simulator instance] now]]\;#
    #	    "${n}([$n id]) recv update from peer $peerAddr id = $id"]

    foreach nbr [array names peers_] {
	if {[$peers_($nbr) addr?] == $peerAddr} {
	    set peer $peers_($nbr)

	    # Retirei o if a seguir, pois aumentei o tamanho das mensagens
	    #	if { [array size metrics] > [Node set nn_] } { ;#
	    #   error "$class::$proc update $peerAddr:$msg:$count is \;#
	    #larger than the simulation topology"
	    #	    }
	    
	    set metricsChanged 0
	    foreach dest [array names metrics] {
		set destlh [lindex $dest 0]
		if {[lindex $dest 1]==0} {            
		    # calculo a seguir do atraso
		    set metric_d [soma $metrics($dest) [$ifs_($nbr) cost0?]] 
		    #puts stderr "metric 0 (atraso): $metric_d"
		    if {$metric_d >= $UNREACH_DELAY} {
			set metric_d $UNREACH_DELAY
		    }
		    if {$metric_d != [$peer metric0? $destlh ]} {        
			$peer metric0 $destlh $metric_d                        
			incr metricsChanged
		    } 
		}	
		if {[lindex $dest 1]==1} {     
		    # calculo a seguir da banda passante
		    set metric_b [minimo $metrics($dest) [$ifs_($nbr) cost1?]]
		    #puts stderr "metric 1(banda passante): $metric_b"
		    if {$metric_b <= $UNREACH_BW} {
			set metric_b $UNREACH_BW
		    }
		    if {$metric_b != [$peer metric1? $destlh ]} {       
			$peer metric1 $destlh $metric_b                   
			incr metricsChanged
		    } 
		}	
		if {[lindex $dest 1]==2} {     
		    # calculo a seguir da prob. de perda
		    set metric_p [prob $metrics($dest) [$ifs_($nbr) cost2?]]
		    #puts stderr "metric 2(prob. perda): $metric_p"
		    if {$metric_p >= $UNREACH_LP} {
			set metric_p $UNREACH_LP
		    }
		    if {$metric_p != [$peer metric2? $destlh ]} {          
			$peer metric2 $destlh $metric_p            
			incr metricsChanged
		    } 
		}	
		if {[lindex $dest 1]==3} {            
		    # calculo a seguir da unica 
		    set metric_u [soma $metrics($dest) \
			   [ unica [$ifs_($nbr) cost0?] [$ifs_($nbr) cost2?] ] ]
		    #puts stderr "metric U (unica): $metric_u"
		    if {$metric_u >= $UNREACH_UNIC} {
			set metric_u $UNREACH_UNIC
		    }
		    if {$metric_u != [$peer metricU? $destlh ]} {            
			$peer metricU $destlh $metric_u                       
			incr metricsChanged
		    } 
		}	
	    }
	    
	    if $metricsChanged {
		$self compute-routes
		incr rtsChanged_ $metricsChanged
		#$rtObject_ compute-routes ;#neste caso, sabemos que prot SMM
		$rtObject_ SMM-compute-routes
	    } else {
		# dynamicDM multicast hack.
		# If we get a message from a neighbour, then something
		# at that neighbour has changed.  While this may not
		# cause any unicast changes on our end, dynamicDM
		# looks at neighbour's routing tables to compute
		# parent-child relationships, and has to do them
		# again.
		#
		$rtObject_ flag-multicast -1
	    }
	    return
        }
    }
    error "$class::$proc update $peerAddr:$msg:$count from unknown peer"
}

Agent/rtProto/SMM proc compute-all {} {
    # Because proc methods are not inherited from the parent class.
}
