#
# Maintainer: <Luis.Costa@lip6.fr>.
#

# This file contains the GENERIC and/or COMMON methods for SMM and 3M routing
# These are derived from route-proto.tcl
# Check ../lib/ns-route.tcl for the Simulator routing support
#

#luish: review this. Functions with the same name would be updated in the 
#original files but not here. We should rename these functions or make the 
#modifications in the original ones.

Agent set ttl_ 32 ; # arbitrary choice here

rtObject set maxpref_   255           ;# quanto menor, maior a preferencia 
                                      ;# de cada protocolo

#luish - limites aceitaveis para as Classes de Servico

rtObject set max_delay  40          ;# atraso maximo, em ms,       
				      ;# 10 vezes o atraso max 
                                      ;# de um enlace experimentalmente

rtObject set min_bw      1	      ;# banda passante minima, em mbps,
				      ;# 2 experimentalmente

rtObject set max_lp	0.04	      ;# prob. de perda maxima, sem un.,
				      ;# 0.3 experimentalmente (menos de 30%)
                                      ;# dez enlaces com perda max

rtObject set max_unic   40.017        ;# metrica unica maxima
                                      ;# 100 + 0.05 (corres. a perda de 0.1)

# Definicao dos valores das diferentes metricas para rota inalcancavel
# Importantes para DV routing, para evitar counting to infinity
# correspondem a 16 saltos

rtObject set unreach_delay	50 
rtObject set unreach_bw		0
rtObject set unreach_lp		0.05
rtObject set unreach_unic	50.022 ;#0.xxx - corresponde a 0.xx de perda

# Def. dos valores das diferentes metricas para rota local (valores otimais)

rtObject set opt_delay              0   ;# atraso para transmissao "local"
rtObject set opt_bw             10000   ;# banda passante "infinita"
rtObject set opt_lp               0.0   ;# prob de perda "local"
rtObject set opt_unic             0.0   ;# met unica para transmissao "local"

# Definition of the worst possible values for all metrics

rtObject set wst_delay           10000   ;# 
rtObject set wst_bw               0.0   ;# 
rtObject set wst_lp               1.0   ;# 
rtObject set wst_unic            10000.99   ;# 

rtObject set contador_ 		1     ;# variavel que armazena o numero de 
				      ;# passos gasto no calculo de rotas

rtObject proc init-all args {

    set ns [Simulator instance]                                     ;#luish
    set routelogic [$ns get-routelogic]                             ;#luish

    set prot_DV     [catch "$routelogic set rtprotos_(DV)"]
    set prot_3M     [catch "$routelogic set rtprotos_(3M)"]
    set prot_SMM    [catch "$routelogic set rtprotos_(SMM)"]
    set prot_3M_LS  [catch "$routelogic set rtprotos_(3M_LS)"]
    set prot_SMM_LS [catch "$routelogic set rtprotos_(SMM_LS)"]

    # criacao de um rtObject para cada no'
    foreach node $args {
	if { [$node rtObject?] == "" } {
	    set rtobj($node) [new rtObject $node]
	}                             
    }
    
    # calculo de rotas para todos os nos  
    foreach node $args {	     
	if { $prot_DV == 0 } {
	    $rtobj($node) compute-routes
	}
	if { $prot_3M == 0 } {
	    $rtobj($node) 3M-compute-routes
	}
	if { $prot_SMM == 0 } {
	    $rtobj($node) SMM-compute-routes
	}
	if { $prot_3M_LS == 0 } {
	    $rtobj($node) 3M_LS-compute-routes
	}
	if { $prot_SMM_LS == 0 } {
	    $rtobj($node) SMM_LS-compute-routes
	}
    }
}

# inicializacao de cada no'
rtObject instproc init node {         
    $self next
    $self instvar ns_ nullAgent_
    $self instvar nextHop_ rtpref_ metric_ node_ rtVia_ rtProtos_  
    
    $self instvar metric0_ metric1_ metric2_ metricU_  ;# luish: mais duas metricas
    $self instvar contador_                   ;# variavel que com o numero de 
                                              ;# passos na descoberta das rotas

    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_

    set ns_ [Simulator instance]
    set nullAgent_ [$ns_ set nullAgent_]
    
    set contador_ 1                   ;# inicializacao do contador_

    $node init-routing $self
    set node_ $node                      ;# inicializacao das rotas de cada no'

    #Se estamos utilisando QoS Routing...
    if [Simulator set EnableQoSRt_] {
	# para todos os outros nos, para as tres classes de servico...
	foreach dest [$ns_ all-nodes-list] {    
	    foreach CoS {0 1 2} {               
		set dest_CoS $dest$CoS
		set nextHop_($dest_CoS) ""        ;# aqui, criamos N rotas 
		# rotas locais   
		if {$node == $dest} {                 
		    set rtpref_($dest_CoS) 0        ;# metrica melhor possivel:
		    set metric_($dest_CoS) 0
		    set metric0_($dest_CoS)  [$class set opt_delay]   ;#atraso
		    set metric1_($dest_CoS) [$class set opt_bw]      ;# banda
		    set metric2_($dest_CoS) [$class set opt_lp]      ;#prob
		    set metricU_($dest_CoS) [$class set opt_unic]    ;# unica
		    set rtVia_($dest_CoS) "Agent/rtProto/Local"
		} else {
		    set rtpref_($dest_CoS) [$class set maxpref_]
		    set metric_($dest_CoS) [$class set unreach_]
		    set metric0_($dest_CoS) [$class set unreach_delay] ;# luish
		    set metric1_($dest_CoS) [$class set unreach_bw]    ;# luish
		    set metric2_($dest_CoS) [$class set unreach_lp]    ;# luish
		    set metricU_($dest_CoS) [$class set unreach_unic]  ;# luish
		    set rtVia_($dest_CoS)    ""
	
		    # Passando a mais a CoS, considerado por qos-add-route.
		    # A rotina classify leva em conta esta convencao.
		    $node add-qos-route [$dest id] $nullAgent_ $CoS 
		}
	    }
	}
    } else {
	#Se nao, estamos no caso normal (sem QoS routing)
	foreach dest [$ns_ all-nodes-list] {
	    set nextHop_($dest) ""
	    if {$node == $dest} {
		set rtpref_($dest) 0
		set metric_($dest) 0
		set rtVia_($dest) "Agent/rtProto/Local" ;# make dump happy
	    } else {
		set rtpref_($dest) [$class set maxpref_]
		set metric_($dest) [$class set unreach_]
		set rtVia_($dest)    ""
		$node add-route [$dest id] $nullAgent_
	    }
	}
    }

    foreach node_i [$ns_ all-nodes-list] {
	foreach node_j [$ns_ all-nodes-list] {
	    set link_ij $node_i$node_j
	    set link_mH_($link_ij)   [$class set unreach_]
	    set link_m0_($link_ij)  [$class set unreach_delay]    ;# luish
	    set link_m1_($link_ij)  [$class set unreach_bw]        ;# luish
	    set link_m2_($link_ij)  [$class set unreach_lp]       ;# luish
	    set link_mU_($link_ij)  [$class set unreach_unic]       ;# luish
	}
    }

    $self add-proto Direct $node
    $rtProtos_(Direct) compute-routes
    
    #    $self compute-routes
}

rtObject instproc intf-changed {} {
    $self instvar ns_ node_ rtProtos_ rtVia_ nextHop_ rtpref_ metric_ 
    $self instvar metric0_ metric1_ metric2_ metricU_                           ;# luish
    #puts stderr "Entrei em intf-changed de rtObject"                  ;# luish

    set routelogic [$ns_ get-routelogic]                             ;#luish

    set prot_DV     [catch "$routelogic set rtprotos_(DV)"]
    set prot_3M     [catch "$routelogic set rtprotos_(3M)"]
    set prot_SMM    [catch "$routelogic set rtprotos_(SMM)"]
    set prot_3M_LS  [catch "$routelogic set rtprotos_(3M_LS)"]
    set prot_SMM_LS [catch "$routelogic set rtprotos_(SMM_LS)"]
    

    foreach p [array names rtProtos_] {
	$rtProtos_($p) intf-changed
	$rtProtos_($p) compute-routes
    }

    # calculo de rotas para todos os nos
    if { $prot_DV == 0 } {
	$self compute-routes  
    }
    if { $prot_3M == 0 } {
	$self 3M-compute-routes
    }
    if { $prot_SMM == 0 } {
	$self SMM-compute-routes
    }
    if { $prot_3M_LS == 0 } {
	$self 3M_LS-compute-routes
    }
    if { $prot_SMM_LS == 0 } {
	$self SMM_LS-compute-routes
    }
}

rtObject instproc QoS_dump-routes chan {
    $self instvar ns_ node_ nextHop_ rtpref_ metric_ rtVia_
    $self instvar metric0_ metric1_ metric2_ metricU_                           ;# luish

    #luish: retirei o trecho a seguir, pois da' problema com tcl/tk 8.0
    #    if {[info proc SplitObjectCompare] == ""} {;#
    #        # XXX: in tcl8 we need to find this in the global namespace
    #        if {[info proc ::SplitObjectCompare] == {}} {;#
    #            puts stderr "$class::$proc failed.  Update your TclCL library"
    #            return
    #            puts stderr "$class::$proc failed.  Update your TclCL library"
    #            return
    #        }
    #    }
    
    if {$ns_ != ""} {
	set time [$ns_ now]
    } else {
	set time 0.0
    }
    puts $chan [concat "Node:\t${node_}([$node_ id])\tat t ="		\
	    [format "%4.2f" $time]]
    puts $chan "Dest\t\tnextHop\tPref\tMetric\tMetric1\tMetric2\tMetricU\tProto"
    puts $chan "  \t\t\t\t\tatraso\tbanda\tperda\tunica"
    foreach dest [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	foreach CoS {0 1 2} {
	    set dest_CoS $dest$CoS
	    if {[llength $nextHop_($dest_CoS)] > 1} { 
		set p [split [$rtVia_($dest_CoS) info class] /]             
		set proto [lindex $p [expr [llength $p] - 1]]
		foreach rt $nextHop_($dest_CoS) {   
		    puts $chan [format \
			    "%-5s(%d)\t%-5s(%d)\t%3d\t%4d\t%4d\t%4d\t%4d\t %s"\
			    $dest_CoS [$dest id] $rt [[$rt set toNode_] id] \ 
		    $rtpref_($dest_CoS) $metric0_($dest_CoS) \
			    $metric1_($dest_CoS) $metric2_($dest_CoS) \
			    $metricU_($dest_CoS) $proto]
		}
	    } elseif {$nextHop_($dest_CoS) != ""} {                       
		set p [split [$rtVia_($dest_CoS) info class] /]         
		set proto [lindex $p [expr [llength $p] - 1]]
		puts $chan [format \
			"%-5s(%d)\t%-5s(%d)\t%3d\t%4d\t%4f\t%4f\t%4f\t %s" \
			$dest_CoS [$dest id] $nextHop_($dest_CoS) \
			[[$nextHop_($dest_CoS) set toNode_] id] \
			$rtpref_($dest_CoS) $metric0_($dest_CoS) \
			$metric1_($dest_CoS) $metric2_($dest_CoS) \
			$metricU_($dest_CoS) $proto]         
	    } elseif {$dest == $node_} {
		puts $chan [format \
			"%-5s(%d)\t%-5s(%d)\t%03d\t%4d\t%4f\t%4f\t%4f\t %s" \
			$dest_CoS [$dest id] $dest [$dest id] \
			$rtpref_($dest_CoS) $metric0_($dest_CoS) \
			$metric1_($dest_CoS) $metric2_($dest_CoS) \
			$metricU_($dest_CoS) "Local"]
	    } else {
		puts $chan [format "%-5s(%d)\t%-5s(%s)\t%03d\t%4d\t %s"	\
			$dest_CoS [$dest id] "" "-" 255 32 "Unknown"]   ;#rever
	    }
	}
    }
}

rtObject instproc metric0? dest {                                      
    $self instvar metric0_                                             ;# luish
    $self set metric0_($dest)                                          ;# luish
}

rtObject instproc metric1? dest {                                      
    $self instvar metric1_                                             ;# luish
    $self set metric1_($dest)                                          ;# luish
}

rtObject instproc metric2? dest {                                      
    $self instvar metric2_                                             ;# luish
    $self set metric2_($dest)                                          ;# luish
}

rtObject instproc metricU? dest {                                      ;# luish
    $self instvar metricU_                                             ;# luish
    $self set metricU_($dest)                                          ;# luish
}

proc soma {m1 m2} {
    return [expr $m1 + $m2]
}

proc minimo {m1 m2} {
    if {$m1 <= $m2} {
	return $m1
    } else {
	return $m2
    }
}

proc prob {m1 m2} {
    return [expr 1 - ((1-$m1)*(1-$m2)) ]
}

proc unica {m1 m2} {
    set a [expr log10(1 - $m2)]       ;# log. da prob. de transm. c/ sucesso
    return [expr $m1 - $a]
}

#Class rtPeer

rtPeer instproc init {addr port cls} {
    $self next
    $self instvar addr_ port_ metric_ rtpref_
    $self instvar metric0_ metric1_ metric2_ metricU_                           ;# luish

    set addr_ $addr
    set port_ $port
    

    #luish
    set UNREACH_        [rtObject set unreach_]
    set UNREACH_DELAY	[rtObject set unreach_delay]
    set UNREACH_BW	[rtObject set unreach_bw]
    set UNREACH_LP	[rtObject set unreach_lp]
    set UNREACH_UNIC	[rtObject set unreach_unic]

    #Se estamos utilisando QoS Routing...
    if [Simulator set EnableQoSRt_] {
	foreach dest [[Simulator instance] all-nodes-list] {
	    foreach CoS {0 1 2} {                                            
		set metric_($dest$CoS)  $UNREACH_
		set metric0_($dest$CoS)  $UNREACH_DELAY     ;# atraso
		set metric1_($dest$CoS) $UNREACH_BW        ;# banda  
		set metric2_($dest$CoS) $UNREACH_LP        ;# prob perda
		set metricU_($dest$CoS) $UNREACH_UNIC      ;# metr unica
		set rtpref_($dest$CoS) [$cls set preference_]        
	    }
	}
	#Se nao, estamos no caso normal (sem QoS routing)
    } else {
	foreach dest [[Simulator instance] all-nodes-list] {
	    set metric_($dest) [$cls set INFINITY]
	    set rtpref_($dest) [$cls set preference_]
	}
    }
}

rtPeer instproc port? {} {
    $self instvar port_
    return $port_
}                                                                              

rtPeer instproc metric0 {dest val} {                                  
    $self instvar metric0_                                             ;# luish
    set metric0_($dest) $val                                           ;# luish
}

rtPeer instproc metric1 {dest val} {                                  
    $self instvar metric1_                                             ;# luish
    set metric1_($dest) $val                                           ;# luish
}

rtPeer instproc metric2 {dest val} {                                
    $self instvar metric2_                                             ;# luish
    set metric2_($dest) $val                                           ;# luish
}

rtPeer instproc metric0? dest {                                     
    $self instvar metric0_                                             ;# luish
    return $metric0_($dest)                                            ;# luish
}

rtPeer instproc metric1? dest {                                     
    $self instvar metric1_                                             ;# luish
    return $metric1_($dest)                                            ;# luish
}

rtPeer instproc metric2? dest {                                       
    $self instvar metric2_                                             ;# luish
    return $metric2_($dest)                                            ;# luish
}

rtPeer instproc metricU {dest val} {                                
    $self instvar metricU_                                             ;# luish
    set metricU_($dest) $val                                           ;# luish
}

rtPeer instproc metricU? dest {                                     
    $self instvar metricU_                                             ;# luish
    return $metricU_($dest)                                            ;# luish
}

# inicializacao das variaveis 
# ifs_: link's (interfaces) para os vizinhos do no', indexadas pelo no' vizinho
# ifstat_: estado de cada um dos link's, up ou down

#Agent/rtProto instproc init node {
#    $self next
#
#    $self instvar ns_ node_ rtObject_ preference_ ifs_ ifstat_
#
#    set ns_ [Simulator instance]
#
#    catch "set preference_ [[$self info class] set preference_]" ret
#    if { $ret == "" } {
#	set preference_ [$class set preference_]
#    }
#    foreach nbr [$node set neighbor_] {   ;# vizinhos do no'
#	set link [$ns_ link $node $nbr]   ;# link entre o no' e o vizinho
#	set ifs_($nbr) $link
#	set ifstat_($nbr) [$link up?]     ;# estado do link
#    }
#    set rtObject_ [$node rtObject?]
#}

###############################################################################
# Direct routing - for direct connected nodes
###############################################################################

#Class Agent/rtProto/Direct -superclass Agent/rtProto

#luish
Agent/rtProto/Direct set UNREACH_       [rtObject set unreach_]
Agent/rtProto/Direct set UNREACH_DELAY  [rtObject set unreach_delay]
Agent/rtProto/Direct set UNREACH_BW	[rtObject set unreach_bw]
Agent/rtProto/Direct set UNREACH_LP  	[rtObject set unreach_lp]
Agent/rtProto/Direct set UNREACH_UNIC	[rtObject set unreach_unic]

Agent/rtProto/Direct instproc init node {
    $self next $node
    $self instvar ns_ rtpref_ nextHop_ metric_ ifs_
    $self instvar metric0_ metric1_ metric2_ metricU_                           ;# luish

    #Se estamos utilisando QoS Routing...    
    if [Simulator set EnableQoSRt_] {
	foreach node [$ns_ all-nodes-list] {
	    foreach CoS {0 1 2} {
		set node_CoS $node$CoS
		set rtpref_($node_CoS) 255                             ;# luish
		set nextHop_($node_CoS) ""                             ;# luish
		set metric_($node_CoS)  [$class set UNREACH_]
		set metric0_($node_CoS)  [$class set UNREACH_DELAY]   ;# atraso
		set metric1_($node_CoS) [$class set UNREACH_BW]      ;# banda
		set metric2_($node_CoS) [$class set UNREACH_LP]      ;# perda
		set metricU_($node_CoS) [$class set UNREACH_UNIC]    ;# unica
	    }
	}
	foreach node [array names ifs_] {
	    foreach CoS {0 1 2} {	
		set rtpref_($node$CoS) [$class set preference_]    ;# luish
	    }
	}
	#Se nao, estamos no caso normal (sem QoS routing)
    } else {            
	foreach node [$ns_ all-nodes-list] {
	    set rtpref_($node) 255
	    set nextHop_($node) ""
	    set metric_($node) -1
	}
	foreach node [array names ifs_] {
	    set rtpref_($node) [$class set preference_]
	}
    }
}

Agent/rtProto/Direct instproc compute-routes {} {

    $self instvar ifs_ ifstat_ nextHop_ metric_ rtsChanged_
    $self instvar metric0_ metric1_ metric2_ metricU_ node_                     ;# luish

    #puts stderr "entrada de Direct compute-routes, no: [$node_ id]"    ;# luish

    set zero 0
    set um 1
    set dois 2

    set rtsChanged_ 0

    #Se estamos utilisando QoS Routing... 
    if [Simulator set EnableQoSRt_] {
	foreach nbr [array names ifs_] {
	    set nbr_0 $nbr$zero
	    set nbr_1 $nbr$um
	    set nbr_2 $nbr$dois
	
	    if {$nextHop_($nbr_0) == "" && [$ifs_($nbr) up?] == "up"} {
		set ifstat_($nbr) 1
		set nextHop_($nbr_0) $ifs_($nbr)
		set nextHop_($nbr_1) $ifs_($nbr)
		set nextHop_($nbr_2) $ifs_($nbr)

		set metric_($nbr_0) [$ifs_($nbr) cost?]
		set metric_($nbr_1) [$ifs_($nbr) cost?]
		set metric_($nbr_2) [$ifs_($nbr) cost?]
		set metric0_($nbr_0) [$ifs_($nbr) cost0?]
		set metric0_($nbr_1) [$ifs_($nbr) cost0?]
		set metric0_($nbr_2) [$ifs_($nbr) cost0?]
		set metric1_($nbr_0) [$ifs_($nbr) cost1?]
		set metric1_($nbr_1) [$ifs_($nbr) cost1?]
		set metric1_($nbr_2) [$ifs_($nbr) cost1?]
		set metric2_($nbr_0) [$ifs_($nbr) cost2?]
		set metric2_($nbr_1) [$ifs_($nbr) cost2?]
		set metric2_($nbr_2) [$ifs_($nbr) cost2?]
		set metricU_($nbr_0) [unica [$ifs_($nbr) cost0?] [$ifs_($nbr) cost2?]]
		set metricU_($nbr_1) [unica [$ifs_($nbr) cost0?] [$ifs_($nbr) cost2?]]
		set metricU_($nbr_2) [unica [$ifs_($nbr) cost0?] [$ifs_($nbr) cost2?]]
		incr rtsChanged_
	    } elseif {$nextHop_($nbr_0) != "" && [$ifs_($nbr) up?] != "up"} {
		set ifstat_($nbr) 0
		set nextHop_($nbr_0) ""
		set nextHop_($nbr_1) ""
		set nextHop_($nbr_2) ""
		
		set metric_($nbr_0)  [$class set UNREACH_]    ;# hops  
		set metric_($nbr_1)  [$class set UNREACH_]
		set metric_($nbr_2)  [$class set UNREACH_]
		set metric0_($nbr_0)  [$class set UNREACH_DELAY]    ;# atraso
		set metric0_($nbr_1)  [$class set UNREACH_DELAY]
		set metric0_($nbr_2)  [$class set UNREACH_DELAY]
		set metric1_($nbr_0) [$class set UNREACH_BW]       ;# banda 
		set metric1_($nbr_1) [$class set UNREACH_BW]
		set metric1_($nbr_2) [$class set UNREACH_BW]
		set metric2_($nbr_0) [$class set UNREACH_LP]       ;# prob
		set metric2_($nbr_1) [$class set UNREACH_LP]
		set metric2_($nbr_2) [$class set UNREACH_LP]
		set metricU_($nbr_0) [$class set UNREACH_UNIC]     ;# unica
		set metricU_($nbr_1) [$class set UNREACH_UNIC]
		set metricU_($nbr_2) [$class set UNREACH_UNIC]
		incr rtsChanged_
	    }
	}
	#Se nao, estamos no caso normal (sem QoS routing)
    } else {
	foreach nbr [array names ifs_] {
	    if {$nextHop_($nbr) == "" && [$ifs_($nbr) up?] == "up"} {
		set ifstat_($nbr) 1
		set nextHop_($nbr) $ifs_($nbr)
		set metric_($nbr) [$ifs_($nbr) cost?]
		incr rtsChanged_
	    } elseif {$nextHop_($nbr) != "" && [$ifs_($nbr) up?] != "up"} {
		set ifstat_($nbr) 0
		set nextHop_($nbr) ""
		set metric_($nbr) -1
		incr rtsChanged_
	    }
	}
    }
    #puts stderr "saida de Direct compute-routes"                       ;# luish
}

rtObject instproc QoS_LS_dump-topology chan {
    $self instvar ns_ node_ nextHop_ rtpref_ metric_ rtVia_
    $self instvar metric0_ metric1_ metric2_ metricU_                ;# luish

    $self instvar link_mH_ link_m0_ link_m1_ link_m2_ link_mU_

    if {$ns_ != ""} {
	set time [$ns_ now]
    } else {
	set time 0.0
    }

    puts $chan [concat "Node:\t${node_}([$node_ id])\tat t ="		\
	    [format "%4.2f" $time]]
    
    puts $chan ""
    puts $chan "Metric H: Number of Hops"    
    puts -nonewline $chan "\t  "
    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
    }
    puts $chan ""

    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
	foreach j [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	    if { $link_mH_($i$j) == [$class set unreach_] } {
		puts -nonewline $chan "XX----XX  "
	    } else {
		puts -nonewline $chan [format "%8.5f  " $link_mH_($i$j)]
	    }
	}
	puts $chan ""
    }

    puts $chan ""
    puts $chan "Metric 0: Propagation Delay"    
    puts -nonewline $chan "\t  "
    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
    }
    puts $chan ""

    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
	foreach j [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	    if { $link_m0_($i$j) == [$class set unreach_delay] } {
		puts -nonewline $chan "XX----XX  "
	    } else {
		puts -nonewline $chan [format "%8.5f  " $link_m0_($i$j)]
	    }
	}
	puts $chan ""
    }

    puts $chan ""
    puts $chan "Metric 1: Residual Bandwidth"
    puts -nonewline $chan "\t  "
    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
    }
    puts $chan ""

    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
	foreach j [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	    if { $link_m1_($i$j) == [$class set unreach_bw] } {
		puts -nonewline $chan "XX----XX  "
	    } else {
		puts -nonewline $chan [format "%8.5f  " $link_m1_($i$j)]
	    }
	}
	puts $chan ""
    }

    puts $chan ""
    puts $chan "Metric 2: Loss Probability"
    puts -nonewline $chan "\t  "
    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
    }
    puts $chan ""

    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
	foreach j [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	    if { $link_m2_($i$j) == [$class set unreach_lp]} {
		puts -nonewline $chan "XX----XX  "
	    } else {
		puts -nonewline $chan [format "%8.5f  " $link_m2_($i$j)]
	    }
	}
	puts $chan ""
    }

    puts $chan ""
    puts $chan "Metric U: Single Mixed Metric"
    
    puts -nonewline $chan "\t  "
    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
    }
    puts $chan ""
    
    foreach i [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	puts -nonewline $chan "[$i id] ($i)  "
	foreach j [lsort -command SplitObjectCompare [$ns_ all-nodes-list]] {
	    if { $link_mU_($i$j) == [$class set unreach_unic] } {
		puts -nonewline $chan "XX----XX  "
	    } else {
		puts -nonewline $chan [format "%8.5f  " $link_mU_($i$j)]
	    }
	}
	puts $chan ""
    }
puts $chan ""
}
